#********************************************************/
#* Gsh Copyright (C) 1999 Vaughn Cato                   */
#* See the file "LICENSE", included with this software, */
#* for license information.                             */
#********************************************************/
proc handle_debug_command {} {
  set command [.debugger.entry get]
  .debugger.entry delete 0 end
  global debug_event
  set debug_event $command
}

proc open_debugger {} {
  if {![widget_exists .debugger]} {
    let w [toplevel .debugger -class dialog] {
      text $w.output
      pack $w.output -side top -fill both -expand true
      let w [entry $w.entry -background pink] {
	pack $w -side top -fill both -expand true
	bind $w <Return> {handle_debug_command}
	bind $w <Any-Enter> { focus %W }
	focus $w
      }
    }
  }
}

proc debug_output {msg} {
  open_debugger
  .debugger.output insert end $msg
}


proc bt {} {
  set result "Stack trace:\n"
  set level 1
  set current_level [expr {[info level]}]
  for {} {$level<$current_level} {incr level} {
    set result "$result$level [info level $level]\n"
  }
  return $result
}

proc debug_wait {} {
  global debug_event
  while {1} {
    set debug_event {}
    puts "Waiting for debug_event"
    tkwait variable debug_event
    debug_output "% $debug_event\n"
    catch {uplevel $debug_event} errormsg
    debug_output "$errormsg\n"
  }
}

proc debug {} {
  open_debugger
  debug_wait
}

proc assert {expr} {
  if {![uplevel [list expr $expr]]} {
    debug_output "Assertion failed: $expr\n"
    uplevel debug_wait
  }
}

proc pause {prompt} {
  puts -nonewline stdout $prompt
  flush stdout
  gets stdin
}

proc tkpause {prompt} {
  update
  message_dialog . $prompt $prompt {"Continue"} 0
}
