#********************************************************/
#* Gsh Copyright (C) 1999 Vaughn Cato                   */
#* See the file "LICENSE", included with this software, */
#* for license information.                             */
#********************************************************/

#****************************************************************
#* Routines for handling a portion of a text widget as an entry *
#****************************************************************

proc fieldref {widget range} {
  assert {[$widget compare [lindex $range 0] <= [lindex $range 1]]}
  return [list $widget $range]
}

proc fieldref_widget {fieldref} {
  return [lindex $fieldref 0]
}

proc fieldref_range {fieldref} {
  return [lindex $fieldref 1]
}

proc fieldref_index {fieldref start motion} {
  set widget [fieldref_widget $fieldref]
  set index "$start $motion"
  set line_range [fieldref_line_range $fieldref $index]
  if {[llength $line_range]==0} {
    set entry_range [fieldref_range $fieldref]
    set entry_start [lindex $entry_range 0]
    set entry_end [lindex $entry_range 1]
    if {[$widget compare $start < $entry_start]} {
      return $entry_start
    }
    if {[$widget compare $start > $entry_end]} {
      return $entry_end
    }
    return $start
  }
  set linestart [lindex $line_range 0]
  set lineend [lindex $line_range 1]
  if {[$widget compare $index < $linestart]} {
    set index $linestart
  }
  if {[$widget compare $index > $lineend]} {
    set index $lineend
  }
  return $index
}

proc fieldref_line_range {fieldref index} {
  set widget [fieldref_widget $fieldref]
  set range [fieldref_range $fieldref]
  set start [lindex $range 0]
  set end [lindex $range 1]
  if {[$widget compare "$index lineend" < $start]} {
    return {}
  }
  if {[$widget compare "$index linestart" > $end]} {
    return {}
  }
  if {[$widget compare "$index linestart" >= $start]} {
    set start [$widget index "$index linestart"]
  }
  if {[$widget compare "$index lineend" <= $end]} {
    set end [$widget index "$index lineend"]
  }
  return [list $start $end]
}

proc fieldref_move_insert {fieldref movement} {
  set widget [fieldref_widget $fieldref]
  set range [fieldref_range $fieldref]
  set start [lindex $range 0]
  set end [lindex $range 1]
  $widget tag remove sel $start $end
  $widget mark set insert [fieldref_index $fieldref "insert" $movement]
}

proc fieldref_at_bottom {fieldref} {
  set entry_range [fieldref_range $fieldref]
  set entry_end [lindex $entry_range 1]
  set line_range [fieldref_line_range $fieldref insert]
  set line_end [lindex $line_range 1]
  set widget [fieldref_widget $fieldref]
  return [$widget compare $line_end == $entry_end]
}

proc fieldref_at_top {fieldref} {
  set entry_range [fieldref_range $fieldref]
  set entry_start [lindex $entry_range 0]
  set line_range [fieldref_line_range $fieldref insert]
  set line_start [lindex $line_range 0]
  set widget [fieldref_widget $fieldref]
  return [$widget compare $line_start == $entry_start]
}

proc fieldref_get {fieldref} {
  set range [fieldref_range $fieldref]
  assert {[llength $range]==2}
  set start [lindex $range 0]
  set end [lindex $range 1]
  return [[fieldref_widget $fieldref] get $start $end]
}

proc fieldref_insert {fieldref index str} {
  set widget [fieldref_widget $fieldref]
  set tags [$widget tag names $index]
  $widget insert $index $str $tags
}

proc fieldref_set {fieldref str} {
  set widget [fieldref_widget $fieldref]
  set range [fieldref_range $fieldref]
  set start [lindex $range 0]
  set end [lindex $range 1]
  set tags [$widget tag names $start]
  $widget delete $start $end
  $widget insert $start $str $tags
}

proc fieldref_insert_pos {fieldref} {
  set entry_range [fieldref_range $fieldref]
  set entry_start [lindex $entry_range 0]
  return [string length [[fieldref_widget $fieldref] get $entry_start insert]]
}

proc fieldref_set_insert_pos {fieldref pos} {
  set entry_range [fieldref_range $fieldref]
  set entry_start [lindex $entry_range 0]
  set widget [fieldref_widget $fieldref]
  $widget mark set insert "$entry_start +$pos c"
}
