#********************************************************/
#* Gsh Copyright (C) 1999 Vaughn Cato                   */
#* See the file "LICENSE", included with this software, */
#* for license information.                             */
#********************************************************/
proc optionmenu {w args} {
  eval [concat [list menubutton $w] $args [list -indicatoron 1 -menu $w.menu \
	-relief raised -bd 2 -highlightthickness 2 -anchor c -direction flush]]
  return $w
}

proc font_dialog_update_example {} {
  global font_dialog
  if {$font_dialog(selected_font)=="Custom"} {
    set font [list \
      -family $font_dialog(family) \
      -size $font_dialog(size) \
      -weight $font_dialog(weight) \
      -slant $font_dialog(slant)]
  } else {
    set font $font_dialog(selected_font)
  }
  $font_dialog(example_text) configure -font $font
  set example_font [$font_dialog(example_text) cget -font]
  if {[$font_dialog(example_text) cget -font]!=$font_dialog(font)} {
    set label "Cancel"
  } else {
    set label "Close"
  }
  $font_dialog(cancel_button) configure -text $label
}

proc font_dialog_apply {} {
  global font_dialog

  set font [$font_dialog(example_text) cget -font]
  eval $font_dialog(change_proc) [list $font]
  set font_dialog(font) $font
  font_dialog_update_example
}

proc font_dialog_cancel {} {
  destroy .font_dialog
}

proc font_dialog_save {} {
  font_dialog_apply
  global theme
  global font_dialog
  set theme(output_font) [$font_dialog(example_text) cget -font]
  set theme(cmdline_font) [$font_dialog(example_text) cget -font]
  save_theme
}

proc font_dialog_reload {} {
  read_theme [rc_dir]/theme temp_theme
  font_dialog_set_attributes $temp_theme(output_font)
  font_dialog_update_example
}

proc font_dialog_set_attr {attr value} {
  global font_dialog
  set font_dialog($attr) $value
  font_dialog_update_example
}

proc font_dialog_set_selected_font {value} {
  global font_dialog
  set font_dialog(selected_font) $value
  font_dialog_update_sensitivity
  font_dialog_update_example
}

proc font_dialog_font_path {font} {
  if {[regexp {^[0-9]+x[0-9]+} $font match]} {
    set rest [string range $font [string length $match] end]
    if {$rest!=""} {
      return [list "fixed" $match $font]
    } else {
      return [list "fixed" $font]
    }
  }
  set groups {}
  set index 0
  while {1} {
    set rest [string range $font $index end]
    set dashindex [string first "-" $rest]
    if {$dashindex==-1} {
      if {[regexp {^[0-9]+} $rest match]} {
	set nextindex [string length $match]
	set endindex $nextindex
	incr endindex -1
      } elseif {[regexp {^[a-zA-Z]+} $rest match]} {
	set nextindex [string length $match]
	set endindex $nextindex
	incr endindex -1
      } else {
	set nextindex [string length $rest]
	set endindex $nextindex
	incr endindex -1
      }
    } else {
      set nextindex $dashindex
      incr nextindex
      set endindex $dashindex
      incr endindex -1
    }
    if {$nextindex==[string length $rest]} {
      break
    }
    incr endindex $index
    lappend groups [string range $font 0 $endindex]
    incr nextindex $index
    set index $nextindex
  }
  lappend groups $font
  return $groups
}

proc font_dialog_font_options_from_hierarchy {hierarchyvar path} {
  upvar $hierarchyvar hierarchy

  set subpaths $hierarchy($path)
  foreach subpath $hierarchy($path) {
    set name [lindex $subpath end]
    if {$subpath==$path || ![info exists hierarchy($subpath)]} {
      lappend options [list $name {}]
    } else {
      set suboptions \
	[font_dialog_font_options_from_hierarchy hierarchy $subpath]
      if {[llength $suboptions]==1} {
	lappend options [lindex $suboptions 0]
      } else {
	lappend options [list $name $suboptions]
      }
    }
  }
  return [lsort -index 0 $options]
}

# Get a list of all font options
proc font_dialog_get_font_options {font_paths} {
  foreach font_path $font_paths {
    set i 0
    set len [llength $font_path]
    incr len
    set prev_part {}
    while {$i<$len} {
      set part [lrange $font_path 0 $i]
      if {![info exists hierarchy($prev_part)]} {
	set hierarchy($prev_part) [list $part]
      } else {
	if {[lsearch $hierarchy($prev_part) $part]==-1} {
	  lappend hierarchy($prev_part) $part
	}
      }
      set prev_part $part
      incr i
    }
  }

  return [font_dialog_font_options_from_hierarchy hierarchy {}]
}

proc font_dialog_fill_hier_menu {w options command} {
  # "options" is a list of options, each option contains a label and a list
  # of suboptions.

  set count 0
  set submenunum 0

  foreach option $options {
    if {$count==20} {
      set menu [menu $w.menu -tearoff 0]
      $w add cascade -label "(MORE)" -menu $menu
      set w $menu
      set count 0
    }
    foreach {label suboptions} $option { }
    if {[llength $suboptions]==0} {
      if {$label=="---"} {
	$w add separator
      } else {
	$w add command -label $label -command [concat $command [list $label]]
      }
    } else {
      incr submenunum
      set submenu [menu $w.menu$submenunum -tearoff 0]
      $w add cascade -label $label -menu $submenu
      font_dialog_fill_hier_menu $submenu $suboptions $command
    }
    incr count
  }
}

proc font_dialog_fill_menu {w options command} {
  $w delete 0 end
  if {[winfo exists $w.menu]} {
    destroy $w.menu
  }
  set hier_options {}
  foreach option $options {
    lappend hier_options [list $option {}]
  }
  font_dialog_fill_hier_menu $w $hier_options $command
}

proc font_dialog_fill_font_menu {w fonts command} {
  set index 0

  set font_paths {}
  foreach font $fonts {
    lappend font_paths [font_dialog_font_path $font]
  }
  set options [font_dialog_get_font_options $font_paths]
  set options [concat {{"Custom" {}} {"---" {}}} $options]
  font_dialog_fill_hier_menu $w $options $command
}

proc font_dialog_family_options {} {
  global font_dialog

  array set spacings $font_dialog(spacings)
  set families $font_dialog(families)
  set fixed_families {}
  set proportional_families {}
  foreach family $families {
    if {$spacings($family)=="p"} {
      lappend proportional_families $family
    } else {
      lappend fixed_families $family
    }
  }
  set fixed_families [lsort $fixed_families]
  set proportional_families [lsort $proportional_families]
  if {!$font_dialog(allow_proportional)} {
    set options $fixed_families
  } else {
    set options [concat $fixed_families {"---"} $proportional_families]
  }
  return $options
}

proc font_dialog_set_attributes {font} {
  global font_dialog

  foreach attr {family size weight slant} {
    set font_dialog($attr) [font actual $font -$attr]
  }
}

proc font_dialog_update_family_menu {} {
  global font_dialog

  set options [font_dialog_family_options]
  set command [list font_dialog_set_attr family]
  font_dialog_fill_menu $font_dialog(family_menu) $options $command
}

proc font_dialog_update_sensitivity {} {
  global font_dialog
  if {$font_dialog(selected_font)=="Custom"} {
    set sensitive 1
  } else {
    set sensitive 0
  }
  set_frame_sensitivity $font_dialog(custom_frame) $sensitive
}

proc open_font_dialog {font} {
  global font_dialog

  puts "Getting font list"
  set font_list [exec xlsfonts]
  set font_list [split $font_list "\n"]
  set font_families {}
  set non_xlfd_fonts {}
  foreach font_name $font_list {
    if {![regexp {^-([^-]*)-([^-]*)-([^-]*)-([^-]*)-([^-]*)-([^-]*)-([^-]*)-([^-]*)-([^-]*)-([^-]*)-([^-]*)-([^-]*)-([^-]*)-([^-]*)$} $font_name _match foundary family weight slant sWidth adstyl pxlsz ptsz resx rexy spc avgwidth registry encoding]} {
      lappend non_xlfd_fonts $font_name
    } else {
      set spacings($family) $spc
    }
  }
  set font_dialog(spacings) [array get spacings]
  foreach {family spc} $font_dialog(spacings) {
    lappend font_families $family
  }
  set font_dialog(families) $font_families
  let w .font_dialog {
    if {[winfo exists $w]} {
      return
    }
    toplevel $w -class Dialog
    wm title $w "Font Selection"
    let w [frame $w.selected_font_frame] {
      let w [label $w.label] {
	$w configure -text "Selected Font:"
	pack $w -side left
      }
      let w [optionmenu $w.font_option_menu] {
	$w configure -textvariable font_dialog(selected_font)
	menu $w.menu -tearoff 0
	font_dialog_fill_font_menu $w.menu $non_xlfd_fonts \
	  [list font_dialog_set_selected_font]
	pack $w -side left
      }
      pack $w -side top
    }
    let w [frame $w.custom_separator] {
      $w configure -height 2 -borderwidth 1 -relief sunken
      pack $w -side top -expand true -fill x
    }
    set font_dialog(allow_proportional) 0
    let w [frame $w.custom_frame] {
      set font_dialog(custom_frame) $w
      $w configure -borderwidth 2 -relief groove
      let w [label $w.label] {
	$w configure -text "Custom Font"
	pack $w -side top
      }
      let w [frame $w.separator] {
	$w configure -height 2 -borderwidth 1 -relief sunken
	pack $w -side top -expand true -fill x -padx 10
      }
      let w [checkbutton $w.allow_proportional_toggle] {
	$w configure -command "font_dialog_update_family_menu"
	$w configure -text "Allow proportional fonts: "
	$w configure -variable font_dialog(allow_proportional)
	pack $w -side top
      }
      let w [frame $w.options_frame] {
	set col 0
	set row 0
	foreach attr {family size weight slant} {
	  let w [label $w.${attr}_label] {
	    switch -- $attr {
	      family { set label "Family:" }
	      size { set label "Size:" }
	      weight { set label "Weight:" }
	      slant { set label "Slant:" }
	      default { assert {0} }
	    }
	    $w configure -text $label
	    grid $w -column $col -row $row -sticky w
	  }
	  font_dialog_set_attributes $font
	  let w [optionmenu $w.${attr}_option_menu] {
	    $w configure -textvariable font_dialog($attr)
	    switch -- $attr {
	      family { set options [font_dialog_family_options] }
	      size {
		set options {}
		for {set size 10} {$size<=30} {incr size} {
		  lappend options $size
		}
	      }
	      weight { set options {normal bold} }
	      slant { set options {roman italic} }
	    }
	    let w [menu $w.menu -tearoff 0] {
	      font_dialog_fill_menu $w $options [list font_dialog_set_attr $attr]
	      if {$attr=="family"} {
		set font_dialog(family_menu) $w
	      }
	    }
	    $w configure -menu $w.menu
	    grid $w -column [expr $col+1] -row $row -sticky w
	  }
	  incr row
	  if {$row>=2} {
	    set row 0
	    incr col 2
	  }
	}
	pack $w -side top -expand true -fill both
	grid columnconfigure $w 0 -weight 0
	grid columnconfigure $w 1 -weight 1
      }
      pack $w -side top -padx 10 -pady 10
    }
    let w [frame $w.example_frame] {
      let w [text $w.text] {
	set font_dialog(example_text) $w
	global theme
	$w configure -background $theme(output_background) 
	$w configure -foreground $theme(output_foreground)
	$w configure -font $font
	$w configure -width 50 -height 4
	$w insert end "prompt$ " prompt
	$w insert end "echo \"This is an inactive command.\"\n" inactive_command
	$w insert end "This is an inactive command.\n" output
	$w insert end "prompt$ " prompt
	$w insert end "echo \"This is a running command.\"\n" active_command
	$w insert end "This is a running command.\n" output
	$w tag configure prompt \
	  -foreground $theme(prompt_foreground) \
	  -background $theme(prompt_background)
	$w tag configure inactive_command \
	  -foreground $theme(inactive_command_foreground) \
	  -background $theme(inactive_command_background)
	$w tag configure active_command \
	  -foreground $theme(active_command_foreground) \
	  -background $theme(active_command_background)
	$w tag configure output \
	  -foreground $theme(output_foreground) \
	  -background $theme(output_background)
	pack $w
      }
      pack $w -side top -padx 10 -pady 10
    }
    let w [frame $w.separator] {
      $w configure -height 2 -borderwidth 1 -relief sunken
      pack $w -side top -expand true -fill x
    }
    let w [frame $w.buttons] {
      let w [button $w.apply] {
	$w configure -text "Apply" -command "font_dialog_apply"
	pack $w -side left -expand true
      }
      let w [button $w.save] {
	$w configure -text "Save" -command "font_dialog_save"
	pack $w -side left -expand true
      }
      let w [button $w.reload] {
	$w configure -text "Reload" -command "font_dialog_reload"
	pack $w -side left -expand true
      }
      let w [button $w.close] {
	set font_dialog(cancel_button) $w
	$w configure -text "Close" -command "font_dialog_cancel"
	pack $w -side left -expand true
      }
      pack $w -side top -expand true -fill x
    }
  }
  font_dialog_update_sensitivity
}

proc edit_font {window font change_proc} {
  global font_dialog

  set font_dialog(font) $font
  if {[llength $font]==1} {
    set font_dialog(selected_font) $font
  } else {
    set font_dialog(selected_font) "Custom"
  }
  set font_dialog(change_proc) $change_proc
  open_font_dialog $font
  center_window .font_dialog $window
}
