#********************************************************/
#* Gsh Copyright (C) 1999 Vaughn Cato                   */
#* See the file "LICENSE", included with this software, */
#* for license information.                             */
#********************************************************/
proc klnew {klname} {
  upvar $klname kl
  set kl { {} {} }
}

proc klfromarray {arrayname} {
  upvar $arrayname array
  set names [array names array]
  set values {}
  foreach name $names {
    lappend values $array($name)
  }
  return [list $names $values]
}

proc klkeys {klname} {
  upvar $klname kl
  return [lindex $kl 0]
}

proc klexists {klname member} {
  return [expr {[lsearch -exact [klkeys $klname] $member]>=0}]
}

proc klset {klname member value} {
  upvar $klname kl
  if {[info exists kl]} {
    set members [lindex $kl 0]
    set values [lindex $kl 1]
  } else {
    set members {}
    set values {}
  }
  set index [lsearch -exact $members $member]
  if {$index==-1} {
    lappend members $member
    lappend values $value
  } else {
    set values [lreplace $values $index $index $value]
  }
  set kl [list $members $values]
  return $value
}

proc klget {klname member args} {
  upvar $klname kl
  if {![info exists kl]} {
    error "$klname does not exist"
  }
  set members [lindex $kl 0]
  set values [lindex $kl 1]
  set index [lsearch -exact $members $member]
  if {[llength $args]==0} {
    if {$index==-1} {
      error "$member is not a member of $klname"
    }
    return [lindex $values $index]
  } elseif {[llength $args]==1} {
    upvar [lindex $args 0] var
    if {$index==-1} {
      return 0
    } else {
      set var [lindex $values $index]
      return 1
    }
  } else {
    error "Usage: klget <klname> <member> [<var>]"
  }
}

proc klappend {klname member value} {
  upvar $klname kl
  if {![info exists kl]} {
    set members {}
    set values {}
  } else {
    set members [lindex $kl 0]
    set values [lindex $kl 1]
  }
  set index [lsearch -exact $members $member]
  if {$index==-1} {
    lappend members $member
    lappend values [list $value]
  } else {
    set current_value [lindex $values $index]
    lappend current_value $value
    set values [lreplace $values $index $index $current_value]
  }
  set kl [list $members $values]
}
