#********************************************************/
#* Gsh Copyright (C) 1999 Vaughn Cato                   */
#* See the file "LICENSE", included with this software, */
#* for license information.                             */
#********************************************************/

# Profiling
if 0 {
  proc dump_profile {} {
    global _profile_times
    global _profile_counts

    set items {}
    foreach {name value} [array get _profile_times] {
      lappend items [list $name $value $_profile_counts($name)]
    }
    set items [lsort -index 1 -integer $items]
    set file [open "profile.out" "w"]
    foreach item $items {
      puts $file "[lindex $item 0]: [lindex $item 1] [lindex $item 2]"
    }
    close $file
    puts "Profile written to profile.out"
  }

  rename proc orig_proc

  orig_proc proc {name argv body} {
    global _profile_times _profile_counts
    set _profile_times($name) 0
    set _profile_counts($name) 0
    set body "
      set _profile_time \[time {
      set _profile_code \[catch [list $body] _profile_return]
      if {\$_profile_code==2} {
	set _profile_code 0
      }
      }
      ]
      global _profile_times _profile_counts
      incr _profile_times($name) \[lindex \$_profile_time 0]
      incr _profile_counts($name)
      return -code \$_profile_code \$_profile_return
    "
    orig_proc $name $argv $body
  }

  foreach proc [info procs] {
    if {$proc!="proc"} {
      proc $proc [info args $proc] [info body $proc]
    }
  }
}

set main_window_list {}

set clone_cmd "exec $restart_command &"

set gsh_lib [file dirname $argv0]
set hostname [info hostname]
set username [exec whoami]

set max_history 500
set theme(output_background) midnightblue
set theme(output_foreground) lightgrey
set theme(output_cursor_color) green
set theme(output_font) {fixed}
set theme(cmdline_font) {fixed}
set theme(prompt_background) midnightblue
set theme(prompt_foreground) white
set theme(active_command_foreground) green
set theme(active_command_background) midnightblue
set theme(inactive_command_foreground) cyan
set theme(inactive_command_background) midnightblue
set theme(selection_foreground) black
set theme(selection_background) lightgrey
set theme(focused_command_cursor_color) green
set theme(focused_command_foreground) yellow
set theme(focused_command_background) midnightblue

if {$env(SHELL)=="/bin/csh" || $env(SHELL)=="/bin/tcsh"} {
  set config(shell_command) "$env(SHELL) -f -c"
} else {
  set config(shell_command) "$env(SHELL) -c"
}
set config(grab_handles) 0
set options(cmdline_in_output) 0
set options(dir_menu_command) {cd $dir;ls}
set options(enter_blank_toggles_compact) 0
set options(cmdline_stays_at_top) 0
set options(start_compact) 0
set options(no_menus_with_top_cmdline) 0
set options(double_click_prompt) 1
set options(override_redirect) 1
set options(expand_only_for_output) 1
set options(initial_width) 80
set options(initial_height) 25
set options(fvwm_geometry_hack) 0

set aliases(history) {echo "$HISTORY"}
set aliases(size_window) {gshcmd size}
set aliases(new_window) {gshcmd new_window}

if {![info exists env(PS1)]} {
  set env(PS1) "$ "
}

set env(COLORTERM) gsh
if {[info exists env(WINDOWID)]} {
  # If gsh is run from an xterm, it has the WINDOWID set, but it
  # isn't correct.
  unset env(WINDOWID)
}

proc chdir {dir} {
  cd $dir
}

eval {
  set old_wd [pwd] 
  chdir $gsh_lib

  source word.tcl; #override normal word characters, these are more useful for 
    # shell stuff
  source let.tcl
  source util.tcl
  source term.tcl
  source debug.tcl
  source types.tcl
  source themedlg.tcl
  source fontdlg.tcl
  source menudlg.tcl
  source optiondlg.tcl
  source fieldref.tcl
  source arrow.tcl

  chdir $old_wd
  unset old_wd
}

proc bgerror {message} {
  global errorInfo
  debug_output "$message\n$errorInfo"
  debug_wait
}

proc arg_value {arg} {
  set value ""
  set index 0
  set char [string index $arg $index]
  set newarg ""
  while {$char!=""} {
    # Command should have already been through variable substitution which
    # changes double quoted strings to single quoted ones.
    #assert {$char!="\"" && $char!="$"}
    if {$char=="'"} {
      set end [string_end $arg "\[^']*" [incr index]]
      incr end
      assert {[string index $arg $end]=="'"}
      append newarg [string range $arg $index [expr {$end-1}]]
      set index $end
    } elseif {$char=="\\"} {
      set char [string index $arg [incr index]]
      switch -- $char {
	"\\" { append newarg "\\" }
	"\n" { }
	"" { error "Got '\\', but no character follows." }
	default { append newarg "\\$char" }
      }
    } else {
      set end [string_end $arg {[^'\\]*} $index]
      append newarg [string range $arg $index $end]
      set index $end
    }
    set char [string index $arg [incr index]]
  }
  return $newarg
}

proc varsub_command {termname cmd} {
  set index 0
  set newcmd ""
  set char [string index $cmd $index]
  while {$char!=""} {
    if {$char=="'"} {
      append newcmd $char
      set char [string index $cmd [incr index]]
      set end [getstring_index $cmd "'" $index]
      assert {$end!=-1}
      append newcmd [string range $cmd $index $end]
      set index $end
      set char [string index $cmd [incr index]]
    } elseif {$char=="\""} {
      append newcmd "'"
      set char [string index $cmd [incr index]]
      while {$char!=""} {
	if {$char=="\""} {
	  append newcmd "'"
	  set char [string index $cmd [incr index]]
	  break
	} elseif {$char=="'"} {
	  append newcmd "'\\''"
	  set char [string index $cmd [incr index]]
	} elseif {$char=="\\"} {
	  set char [string index $cmd [incr index]]
	  switch -- $char {
	    "\\" -
	    "\"" { append newcmd $char }
	    "'" { append newcmd "\\'\\''" }
	    default { append newcmd "\\$char" }
	  }
	  set char [string index $cmd [incr index]]
	} elseif {$char=="\$"} {
	  set end [varname_end $cmd [incr index]]
	  set varname [string range $cmd $index $end]
	  set index $end
	  set char [string index $cmd [incr index]]
	  set value [variable_value $termname $varname]
	  set subindex 0
	  set subchar [string index $value $subindex]
	  while {$subchar!=""} {
	    if {$subchar=="'"} {
	      append newcmd "'\"'\"'"
	    } elseif {$subchar=="\\"} {
	      append newcmd "\\$subchar"
	    } else {
	      append newcmd $subchar
	    }
	    set subchar [string index $value [incr subindex]]
	  }
	} else {
	  append newcmd $char
	  set char [string index $cmd [incr index]]
	}
      }
    } elseif {$char=="\$"} {
      set end [varname_end $cmd [incr index]]
      set varname [string range $cmd $index $end]
      set index $end
      set char [string index $cmd [incr index]]
      set value [variable_value $termname $varname]
      set newvalue ""
      set subindex 0
      set subchar [string index $value $subindex]
      while {$subchar!=""} {
	set subindex [string_end $value "\[ \n]*" $subindex]
	set subchar [string index $value [incr subindex]]
	if {$subchar!=""} {
	  if {$newvalue==""} {
	    append newvalue "'"
	  } else {
	    append newvalue " '"
	  }
	  while {$subchar!=" " && $subchar!="\n" && $subchar!=""} {
	    if {$subchar=="'"} {
	      append newvalue "'\\''"
	    } elseif {$subchar=="\\"} {
	      append newvalue "\\$subchar"
	    } else {
	      set end [string_end $value "\[^ \n'\\]*" $subindex]
	      append newvalue [string range $value $subindex $end]
	      set subindex $end
	    }
	    set subchar [string index $value [incr subindex]]
	    if {$subchar==""} {
	      break
	    }
	  }
	  append newvalue "'"
	}
      }
      append newcmd $newvalue
    } elseif {$char=="\\"} {
      append newcmd "\\\\"
      set char [string index $cmd [incr index]]
    } else {
      set end [string_end $cmd {[^"'$\\]*} $index]
      append newcmd [string range $cmd $index $end]
      set index $end
      set char [string index $cmd [incr index]]
    }
  }
  return $newcmd
}

proc expand_prompt {window prompt} {
  global username hostname
  upvar #0 main_windows($window,dir) dir

  set prompt [varsub_command "" $prompt]
  set prompt [arg_value $prompt]
  regsub {\\t} $prompt [clock format [clock seconds] -format "%H:%M:%S"] prompt
  regsub {\\d} $prompt [clock format [clock seconds] -format "%a %b %d"] prompt
  regsub {\\n} $prompt "\n" prompt
  regsub {\\s} $prompt "gsh" prompt
  regsub {\\w} $prompt $dir prompt
  regsub {\\W} $prompt [file tail $dir] prompt
  regsub {\\u} $prompt $username prompt
  regsub {\\h} $prompt $hostname prompt
  regsub {\\\$} $prompt "$" prompt
  return $prompt
}

proc create_dir {dir} {
  if [file exists $dir] { return }
  set parent_dir [file dirname $dir]
  create_dir $parent_dir
  file mkdir $dir
}

proc defaults_dir {} {
  set _profile_code [catch {
  global gsh_lib
  return $gsh_lib/defaults
  } _profile_message]
  if {$_profile_code==2} {
    set _profile_code 0
  }
  return -code $_profile_code $_profile_message
}

set defaults_dir [defaults_dir]
foreach filename {init types popup_commands cmd_menu functions} {
  let file [rc_dir]/$filename {
    if {![file exists $file]} {
      set default_file $defaults_dir/$filename
      if {![file exists $default_file]} {
	puts "neither $file nor $default_file exist"
      } else {
	create_dir [file dirname $file]
	if {[file isdirectory $default_file]} {
	  file mkdir $file
	  foreach sub_default_file [glob -nocomplain $default_file/*] {
	    set sub_file $file/[file tail $sub_default_file]
	    file copy $sub_default_file $sub_file
	    puts "Your $sub_file has been created with defaults."
	  }
	} else {
	  file copy $default_file $file
	  puts "Your $file has been created with defaults."
	}
      }
    }
  }
}

let init_file "[rc_dir]/init" {
  if [catch {source $init_file} error] {
    bgerror "Error processing $init_file: $error"
  }
}

set history {}
set variables(HISTORY) ""

proc widget_exists {widget} {
  return [expr [llength [info commands $widget]] != 0]
}

proc add_popup_command {window cmd desc verify} {
  global popup_menu_entries
  if {$desc==""} {
    set desc $cmd
  }
  lappend popup_menu_entries [list command $desc $cmd $verify]
}

proc files_matching {dir pattern} {
  chdir $dir
  return [glob -nocomplain -- $pattern]
}

proc add_popup_separator {window} {
  global popup_menu_entries
  $window.popup add separator
  lappend popup_menu_entries {}
}


proc nocase_re {str} {
  set upcase [string toupper $str]
  set lowcase [string tolower $str]
  set strlen [string length $str]
  set result ""
  for {set i 0} {$i < $strlen} {incr i} {
    set result "$result\[[string index $upcase $i][string index $lowcase $i]]"
  }
  return $result
}

proc quote_file {filename} {
  if {[regexp {[ "$><]} $filename]} {
    return "'$filename'"
  } else {
    return $filename
  }
}

proc quote_files {filenames} {
  set result ""
  foreach filename $filenames {
    if {$result!=""} {
      append result " [quote_file $filename]"
    } else {
      set result $filename
    }
  }
  return $result
}

# Return a list of the file names that are found in the given string.
proc file_list {dir names} {
  set files {}
  set all_exist 1
  set filename [file join $dir $names]
  if {[file exists $filename]} {
    set files [quote_file $filename]
  } else {
    foreach file $names {
      set filename [file join $dir $file]
      if {![file exists $filename]} {
	set all_exist 0
	return ""
      }
      set files "$files [quote_file $filename]"
    }
  }
  return $files
}

proc expand_command {_command _user_var_keylist} {
  foreach _key [klkeys _user_var_keylist]  {
    klget _user_var_keylist $_key $_key
  }
  regsub -all -- "\"" $_command "\\\"" _cmd2
  eval "return \"$_cmd2\""
}

proc fill_menu {window menu entries termname} {
  $menu delete 0 end

  bind Menu <ButtonRelease-2> { }
  bind $menu <2> [list menu_select $window $menu $entries $termname]

  set entry_num 0
  foreach entry $entries {
    set type [lindex $entry 0]
    switch -- $type {
      separator {
	$menu add separator
      }
      menu -
      history_menu {
	set label [lindex $entry 1]
	set submenu $menu.submenu$entry_num
	catch {destroy $submenu}
	if {$type=="history_menu"} {
	  menu $submenu \
	    -postcommand \
	      [list fill_history_menu $window $submenu [lindex $entry 2]]
	  bind $submenu <2> "menu_select_history $window $submenu"
	} else {
	  menu $submenu
	  bind $submenu <2> ""
	  fill_menu $window $submenu [lindex $entry 2] $termname
	}
	$menu add cascade -label $label -menu $submenu
      }
      command {
	set label [lindex $entry 1]
	$menu add command -label $label \
	  -command [list "command_selected" $window $termname $entry]
      }
    }
    incr entry_num
  }
}

proc popup_output_menu {window dir cmd x y termname} {
  global popup_menu_entries popup_parent_window
  global main_windows
  set window_dir $main_windows($window,dir)
  set popup_parent_window $window
  set popup_menu_entries {}
  $window.popup delete 0 end
  set selection ""
  catch { set selection [selection get] } message
  set data_types [data_types $selection $dir $cmd user_var_keylist]
  set data_value [lindex $data_types 1]
  set data_types [lindex $data_types 0]
  foreach data_type $data_types {
    global type_info
    if {[info exists type_info($data_type)]} {
      if {[klget type_info($data_type) commands commands]} {
	foreach command $commands {
	  set string [klget command string]
	  set expanded_command [expand_command $string $user_var_keylist]
	  set desc ""
	  klget command desc desc
	  set desc [expand_command $desc $user_var_keylist]
	  set verify 0
	  klget command verify verify
	  add_popup_command $window $expanded_command $desc $verify
	}
      }
    }
  }
  set files {}
  if {[lsearch -exact $data_types "file"]!=-1} {
    set files [list $data_value]
  } elseif {[lsearch -exact $data_types "multiple_files"]!=-1} {
    set files $data_value
  }
  if {[llength $files]>0} {
    set valid_list ""
    foreach file $files {
      set seldir [file join $window_dir $file]
      set seldir [file dirname $seldir]
      if {$seldir!=$dir && $file!=$dir} {
	set valid_list "$valid_list [quote_file $file]"
      }
    }
    if {$valid_list!=""} {
      add_popup_command $window "cp $valid_list $dir" "" 0
      add_popup_command $window "mv $valid_list $dir" "" 0
    }
  }
  if {[lsearch -exact $data_types "directory"]!=-1} {
    set seldir [file dirname [file join $window_dir $data_value]]
    if {$seldir!=$dir && $data_value!=$dir} {
      add_popup_command $window "cp -R [quote_file $data_value] $dir" "" 0
      add_popup_command $window "mv [quote_file $data_value] $dir" "" 0
    }
  }
  
  global popup_menu_entries
  fill_menu $window $window.popup $popup_menu_entries $termname
  $window.popup add separator
  $window.popup add command -label "Other..." -command \
    "open_type_editor {$data_types}"
  tk_popup $window.popup $x $y
}

proc cmdline_widget {window} {
  upvar #0 main_windows($window,cmdline_entry) cmdline_entry
  return $cmdline_entry
}

proc update_idletasks {} {
  update idletasks
}

proc update_geometry {window height_only} {
  global updating_geometry
  if {[info exists updating_geometry]} {
    return
  }
  set updating_geometry 1
  update_idletasks

  # Make sure the window is resizable before we try to update it's geometry.
  wm resizable $window 1 1

  if {$height_only} {
    set new_width [winfo width $window]
    set new_height [winfo reqheight $window]
  } else {
    set new_width [winfo reqwidth $window.outputarea]
    set new_height [winfo reqheight $window]
  }
  if {$new_width!=[winfo width $window] ||
      $new_height!=[winfo height $window]} {
    set geometry [wm geometry $window]
    scan $geometry "%dx%d+%d+%d" w h x y

    global options

    if {$options(fvwm_geometry_hack)} {
      # Under fvwm, if you set the geometry of a window that is already
      # mapped, you are setting the position of the window itself, instead
      # of the position of the window manager border.
      set x [winfo rootx $window]
      set y [winfo rooty $window]
    } else {
      global main_windows
      set position_changed 1
      if {[info exists main_windows($window,x)]} {
	if {$main_windows($window,x)==$x ||
	    $main_windows($window,y)==$y} {
	  set position_changed 0
	}
      }
	
      if {$position_changed} {
	incr x 1 
	incr y 1
      }
    }
    if {![winfo ismapped $window]} {
      # Don't position the window if it isn't mapped yet so the window
      # manager will place it.
      wm geometry $window ${new_width}x${new_height}
    } else {
      wm geometry $window ${new_width}x${new_height}+$x+$y
    }
    global main_windows
    set main_windows($window,x) $x
    set main_windows($window,y) $y
  } else {
    # If the new geometry is the same as the current gemetry, setting it
    # is really slow, so optimize this out.
  }
  update_idletasks

  global main_windows
  if {$main_windows($window,is_compact)} {
    wm resizable $window 1 0
  } else {
    wm resizable $window 1 1
  }

  unset updating_geometry
  global after_update_geometry
  if {[info exists after_update_geometry]} {
    after idle $after_update_geometry
    unset after_update_geometry
  }
}

proc focus_idletask {window} {
  # Make sure this is the last idletask
  update_idletasks

  # Geometry changes must complete before we can focus properly.
  global updating_geometry
  if {[info exists updating_geometry]} {
    global after_update_geometry
    set after_update_geometry [list focus_idletask $window]
    return
  }

  global main_windows
  set w $main_windows($window,focused_widget)
  if {![winfo exists $w]} {
    return
  }
  if {[wm state $window]=="iconic"} {
    wm deiconify $window
    update_idletasks
  }
  set current_window [focus]
  if {$current_window!=""} {
    while {1} {
      set parent [winfo parent $current_window]
      if {$parent!="."} {
	set current_window $parent
      } else {
	break
      }
    }
  }
  focus $w
  if {$current_window!="" && $current_window != $window} {
    focus [focus -lastfor $current_window]
  }
  assert {[focus -lastfor $window]==$w}
}

proc focus_widget {window w} {
  global main_windows
  set main_windows($window,focused_widget) $w
  after idle focus_idletask $window
}

proc focused_widget {window} {
  upvar #0 main_windows($window,focused_widget) focused_widget
  if {[info exists focused_widget]} {
    return $focused_widget
  } else {
    return ""
  }
}

proc cmdline_in_output {window} {
  upvar #0 main_windows($window,cmdline_entry) cmdline_entry
  if {[info exists cmdline_entry]} {
    return 0
  }
  global options
  assert {$options(cmdline_in_output)}
  return 1
}

proc focused_entry_term {window} {
  set widget $window.outputarea.text
  set tags [$widget tag names insert]

  if {[lsearch -exact $tags "focused_entry"]!=-1} {
    set termname [term_under_index $widget insert]
    assert {$termname!=""}
    return $termname
  }
  return ""
}

proc cmdline_focused {window} {
  global config
  if {[cmdline_in_output $window]} {
    upvar #0 main_windows($window,cmdline_term) cmdline_term
    if {[info exists cmdline_term]} {
      set focused_entry_term [focused_entry_term $window]
      if {$focused_entry_term==$cmdline_term} {
	return 1
      }
    }
    return 0;
  } else {
    upvar #0 main_windows($window,cmdline_entry) cmdline_entry
    upvar #0 main_windows($window,focused_widget) focused_widget
    if {[info exists focused_widget]} {
      return [expr {$cmdline_entry==$focused_widget}]
    } else {
      return 0
    }
  }
}

proc tag_ranges {widget tag start end} {
  set range [$widget tag prevrange $tag "$start +1c"]
  if {[llength $range]==0} {
    set range [$widget tag nextrange $tag $start]
  }
  set result {}
  while {[llength $range]==2} {
    set rs [lindex $range 0]
    set re [lindex $range 1]
    if {[$widget compare $rs >= $end]} {
      break
    }
    if {[$widget compare $re > $start]} {
      if {[$widget compare $rs < $start]} {
	set rs $start
      }
      if {[$widget compare $re > $end]} {
	set re $end
      }
      lappend result [list $rs $re]
    }
    set range [$widget tag nextrange $tag $re]
  }
  return $result
}

proc tag_range {widget tag start end} {
  set ranges [tag_ranges $widget $tag $start $end]
  assert {[llength $ranges]==1}
  return [lindex $ranges 0]
}

proc replace_tag {widget old new start end} {
  set range [tag_range $widget $old $start $end]
  assert {[llength $range]==2}
  set start [lindex $range 0]
  set end [lindex $range 1]
  $widget tag remove $old $start $end
  foreach tag $new {
    $widget tag add $tag $start $end
  }
}

proc replace_header_tag {termname old new} {
  assert {$old!=""}
  upvar #0 $termname term
  set term_widget [term_widget $termname]
  set header_index [$term_widget index ${termname}header]
  set top_index [$term_widget index ${termname}top]
  assert {[$term_widget compare $header_index < $top_index]}
  replace_tag $term_widget $old $new ${termname}header ${termname}top
  assert {[$term_widget compare ${termname}header < ${termname}top]}
  assert {[$term_widget compare ${termname}top <= ${termname}bottom]}
}

proc unfocus_term_entries {window} {
  set old_termname [focused_entry_term $window]
  if {$old_termname==""} {
    unfocus_cmdline $window
  } else {
    replace_header_tag $old_termname focused_entry entry
  }
}

proc init_vi_state {statevar} {
  upvar $statevar statelist
  set statelist {
    mode insert 
    cmd "i" 
    findcmd ""
    last_cmd "" 
    current_cmd ""
    inserted "" 
    undo_buffer "" 
    undo_insert_pos 0
    orig_line ""
    yanktype none
  }
}

proc window_init_vi_state {window} {
  upvar #0 main_windows($window,cmdline_state) cmdline_state
  init_vi_state cmdline_state
}

proc focus_term_entry {window termname} {
  upvar #0 $termname term
  window_init_vi_state $window
  set widget [term_widget $termname]
  unfocus_term_entries $window
  set header_start [$widget index "${termname}header"]
  set header_end [$widget index "${termname}top"]
  set entry_range [header_entry_range $window $termname]
  if {$entry_range==""} {
    puts "No entry found for term $termname"
    return
  }
  replace_header_tag $termname entry focused_entry
  term_unfocus_all $widget
  global theme
  $window.outputarea.text configure \
    -insertbackground $theme(focused_command_cursor_color)
  focus_widget $window $widget
  $widget mark set insert "[lindex $entry_range 1] -1c"
  $widget see insert
  window_set_default_title $window "gsh - $term(dir)"
}

proc add_cmdline_term {window} {
  global main_windows
  set dir $main_windows($window,dir)
  let termname [create_decor_term $window $dir "" entry] {
    #run_eof $window $termname
    global main_windows
    set main_windows($window,cmdline_term) $termname
    return $termname
  }
}

proc focus_cmdline {window} {
  upvar #0 main_windows($window,cmdline_entry) cmdline
  upvar #0 main_windows($window,dir) dir
  upvar #0 main_windows($window,cmdline_term) cmdline_term

  assert {[info exists cmdline] || [info exists cmdline_term]}
  term_unfocus_all $window.outputarea.text
  if {[info exists cmdline]} {
    unfocus_term_entries $window
    focus_widget $window $cmdline
    if {[lsearch -exact [$cmdline tag names insert] "focused_entry"]==-1} {
      replace_tag $cmdline entry focused_entry 1.0 end
    }
  } else {
    focus_term_entry $window $cmdline_term
  }
  assert {[cmdline_focused $window]}
  assert {[focused_term $window.outputarea.text]==""}
  window_set_default_title $window "gsh - $dir"
}

proc unfocus_cmdline {window} {
  global config
  upvar #0 main_windows($window,cmdline_entry) cmdline
  if {[info exists cmdline]} {
    assert {[winfo exists $cmdline]}
    if {[lsearch -exact [$cmdline tag names insert] "focused_entry"]!=-1} {
      replace_tag $cmdline focused_entry entry 1.0 end
    }
  }
}

proc term_focused_notify {window termname} {
  unfocus_term_entries $window
  upvar #0 $termname term
  window_set_default_title $window "gsh - $term(cmd)"
}


proc cmdline_placement {window} {
  upvar #0 main_windows($window,cmdline_placement) where
  if {[info exists where]} {
    return $where
  } else {
    return ""
  }
}

proc cmdline_get {widget} {
  set range [$widget tag ranges focused_entry]
  if {$range==""} {
    set range [$widget tag ranges entry]
  }
  assert {[llength $range]==2}
  return [$widget get [lindex $range 0] "[lindex $range 1] -1c"]
}

proc window_cmdline_term {window} {
  global main_windows
  return $main_windows($window,cmdline_term)
}

proc create_cmdline {window where} {
  global main_windows
  set main_windows($window,cmdline_placement) $where
  upvar #0 main_windows($window,cmdline_entry) cmdline_entry

  switch -- $where {
    bottom { 
      global options
      if {!$options(cmdline_in_output)} {
	set w $window.outputarea.cmdline 
      }
    }
    top { set w $window.topline.cmdline }
  }

  # If the old and new command lines don't exists, there is nothing to do
  if {![info exists cmdline_entry] && ![info exists w]} {
    return
  }
  
  # If the old a new command lines are the same widget, there is nothing
  # to do.
  if {[info exists cmdline_entry] && [info exists w] && $cmdline_entry==$w} {
    return
  }

  # Destroy the old command line

  if {[info exists cmdline_entry]} {
    set old_text [cmdline_get $cmdline_entry]
    destroy $cmdline_entry
  } else {
    global main_windows
    if {[info exists main_windows($window,cmdline_term)]} {
      set termname [window_cmdline_term $window]
      set old_text [term_entry_get $window $termname]
      unset main_windows($window,cmdline_term)
      remove_term $window $termname
    } else {
      set old_text ""
    }
  }

  if {[winfo exists $window.topline.history_arrow]} {
    destroy $window.topline.history_arrow
  }

  # Create the new command line
  if {![info exists w]} {
    unset cmdline_entry
    give_cmdline $window
    window_set_cmdline $window $old_text
    return
  }

  switch -- $where {
    bottom {
      text $w -height 1 -width 1 -wrap char
      grid $w -row 1 -column 0 -sticky ew
      global config
      if {$config(grab_handles)} {
	grid $w -columnspan 2
      } else {
	grid $w -columnspan 1
      }
      grid rowconfigure [winfo parent $w] 1 -weight 0
    }
    top {
      text $w -width 1 -height 1 -wrap char
      pack $w -side left -expand 1 -fill both
      arrow_button down $window.topline.history_arrow \
        -command [list toggle_history_window $window]
      pack $window.topline.history_arrow -side left -fill y
    }
  }

  init_cmdline $window $w
  update_prompt $window

  # If the command line is being created for a new window, some variables
  # that "window_set_cmdline" need may not be set yet. 
  # (i.e. main_windows($window,is_compact)).  The old text will be
  # empty in this case, and we don't need to set it anyway if it is empty.
  if {$old_text!=""} {
    window_set_cmdline $window $old_text
  }

  set cmdline_entry $w
}

proc layout_outputarea {window} {
  set outputarea $window.outputarea

  global options
  global config

  if {$config(grab_handles)} {
    grid columnconfigure $outputarea 0 -weight 0
    grid $outputarea.grab_frame -row 0 -column 0 -sticky ns -rowspan 2
    set output_column 1
    set scrollbar_column 2
  } else {
    set output_column 0
    set scrollbar_column 1
  }

  grid rowconfigure $outputarea 0 -weight 1
  grid columnconfigure $outputarea $output_column -weight 1
  grid $outputarea.text -row 0 -column $output_column -sticky nsew
  grid $outputarea.vscroll -row 0 -column $scrollbar_column -sticky ns

  if {$options(cmdline_in_output) || $options(cmdline_stays_at_top)} {
    grid $outputarea.text -rowspan 2
    grid $outputarea.vscroll -rowspan 2
    grid forget $outputarea.history_button
  } else {
    grid $outputarea.text -rowspan 1
    grid $outputarea.vscroll -rowspan 1
    grid $outputarea.history_button \
      -row 1 -column $scrollbar_column -sticky nsew
  }
}

proc window_layout_noncompact {window} {
  set outputarea $window.outputarea
  grid $outputarea -row 1 -column 0 -sticky nsew
  grid rowconfigure $window 1 -weight 1
  recreate_cmdline $window 0
}

proc layout_window {window compact} {
  set outputarea $window.outputarea

  # Always layout the outputarea, even in compact mode, because the
  # width of the window in compact mode is determined by the requested
  # width of the output area.
  layout_outputarea $window
  if {!$compact} {
    window_layout_noncompact $window
  } else {
    create_cmdline $window top
    grid forget $outputarea
  }

  global main_windows
  if {[info exists main_windows($window,is_compact)]} {
    set new_window 0
  } else {
    set new_window 1
  }
  if {$new_window} {
    set have_menus 0
  } else {
    if {[winfo exists $window.main_popup]} {
      set have_menus 0
    } else {
      set have_menus 1
    }
  }

  global main_windows
  set cmdline_placement $main_windows($window,cmdline_placement)
  global options
  if {$cmdline_placement=="top" && $options(no_menus_with_top_cmdline)} {
    set want_menus 0
  } else {
    set want_menus 1
  }

  set in_popup [expr {!$want_menus}]
  if {$new_window} {
    create_main_menu $window $in_popup
  } else {
    if {$have_menus!=$want_menus} {
      destroy_main_menu $window
      create_main_menu $window $in_popup
    }
  }

  global main_windows
  set main_windows($window,is_compact) $compact
  if {$compact} {
    set insert_handler [list term_insert_notify $window]
  } else {
    set insert_handler ""
  }
  term_set_insert_handler $window.outputarea.text $insert_handler
}

proc window_set_compact {window compact} {
  upvar #0 main_windows($window,is_compact) is_compact
  if {[info exists is_compact]} {
    if {$is_compact==$compact} {
      return
    }
    set is_new_window 0
  } else {
    set is_new_window 1
  }

  layout_window $window $compact

  if {!$is_new_window} {
    update_geometry $window 1
  } else {
    update_geometry $window 0
  }

  update idletasks

  set last_focus [focused_widget $window]
  if {$compact} {
    focus_cmdline $window\
  } else {
    if {$last_focus==$window} {
      focus_widget $window $window.outputarea.text
    } elseif {$last_focus!="$window.outputarea.text"} {
      focus_cmdline $window
    }
  }
}

# Called if there is new output to be seen.
proc term_insert_notify {window termname} {
  window_set_compact $window 0
}

proc entry_range {widget start end} {
  set range [tag_ranges $widget focused_entry $start $end]
  if {[llength $range]!=0} {
    assert {[llength $range]==1}
    set range [lindex $range 0]
    assert {[llength $range]==2}
  }
  if {$range=="" || [$widget compare [lindex $range 0] >= $end]} {
    set range [tag_ranges $widget entry $start $end]
    assert {[llength $range]==1}
    set range [lindex $range 0]
    assert {[llength $range]==2}
  }
  assert {[$widget compare [lindex $range 1] <= $end]}
  return $range
}

proc header_entry_range {window termname} {
  upvar #0 $termname term
  set widget [term_widget $termname]
  set start ${termname}header
  set end ${termname}top
  return [entry_range $widget $start $end]
}

proc cmdline_entry_range {cmdline} {
  set range [entry_range $cmdline 1.0 end]
  if {[llength $range]==0} {
    return $range
  }
  set start [lindex $range 0]
  set end [lindex $range 1]
  set end [$cmdline index "$end -1c"]
  return [list $start $end]
}

proc add_header {window termname index prompt cmd entry_tag} {
  upvar #0 $termname term
  set term_widget [term_widget $termname]
  set index [$term_widget index $index]

  # The bottom mark might be at the end mark, and if it is, we can't really
  # insert there.
  while {[$term_widget compare $index >= bottom]} {
    $term_widget insert bottom "\n"
  }
  $term_widget mark set ${termname}header $index
  $term_widget mark gravity ${termname}header left
  set line_count [llength [split "$prompt$cmd" "\n"]]
  while {$line_count>0} {
    if {[$term_widget \
         compare ${termname}header < "bottom -1 lines linestart"]} {
      $term_widget delete $index "$index +1 lines linestart"
    }
    incr line_count -1
  }
  $term_widget mark set ${termname}top ${termname}header
  $term_widget insert ${termname}top "$prompt" [list prompt ${termname}prompt]
  $term_widget insert ${termname}top "$cmd\n" $entry_tag

  assert {[$term_widget compare ${termname}header < ${termname}top]}
  $term_widget tag add header ${termname}header ${termname}top
}

proc recreate_cmdline {window compact} {
  global options
  if {!$compact} {
    if {$options(cmdline_stays_at_top)} {
      create_cmdline $window top
    } else {
      create_cmdline $window bottom
    }
  }
}

proc create_decor_term {window dir cmd entry_tag} {
  set term_widget $window.outputarea.text
  set last_termname [term_last $term_widget]
  if {$last_termname==""} {
    set index "bottom -1 lines linestart"
  } else {
    set last_term_top [$term_widget index ${last_termname}top]
    set last_term_bottom [$term_widget index ${last_termname}bottom]
    set index $last_term_bottom
    upvar #0 $last_termname last_term
    set last_term_insert [term_insert_mark $last_termname]
    # Temporarily set the gravity of the insert mark for the last term
    # to left so it won't move.
    $term_widget mark gravity $last_term_insert left
  }
  set termname [create_term $term_widget]
  global env
  set prompt [expand_prompt $window [variable_value "" PS1]]
  add_header $window $termname $index $prompt $cmd $entry_tag
  # If the last term has it's top or bottom mark at the end, then it would
  # get moved down.  Reset them here to work around that.
  if {$last_termname!=""} {
    $term_widget mark set ${last_termname}top $last_term_top
    $term_widget mark set ${last_termname}bottom $last_term_bottom
    $term_widget mark gravity $last_term_insert right
  }
  assert {[$term_widget compare ${termname}header < ${termname}top]}
  set rows [window_rows $window]
  set cols [window_cols $window]
  init_term $termname $term_widget ${termname}top "bottom" $rows $cols
  upvar #0 $termname term
  set term_widget [term_widget $termname]
  set header_index [$term_widget index "${termname}header"]
  set top_index [$term_widget index "${termname}top"]
  set bottom_index [$term_widget index "${termname}bottom"]
  assert {[$term_widget compare $header_index < $top_index]}
  assert {[$term_widget compare $top_index < $bottom_index]}
  set term(dir) $dir
  set term(cmd) $cmd
  return $termname
}

proc window_rows {window} {
  upvar #0 main_windows($window,rows) rows
  return $rows
}

proc window_cols {window} {
  upvar #0 main_windows($window,cols) cols
  return $cols
}

proc update_cmdline_height {window} {
  set cmdline_entry [cmdline_widget $window]
  set lines [full_text_height $cmdline_entry]

  set current_height [$cmdline_entry cget -height]
  if {$lines!=$current_height} {
    $cmdline_entry configure -height $lines

    upvar #0 main_windows($window,is_compact) is_compact
    if {[info exists is_compact]} {
      if {$is_compact} {
	update_geometry $window 1
      }
    }
  }
}

proc update_prompt {window} {
  if {![cmdline_in_output $window]} {
    set cmdline [cmdline_widget $window]
    set range [$cmdline tag ranges prompt]
    if {[llength $range]==2} {
      $cmdline delete [lindex $range 0] [lindex $range 1]
    }
    set prompt [expand_prompt $window [variable_value "" PS1]]
    $cmdline insert 1.0 $prompt prompt
    update_cmdline_height $window
  }
}

proc completion_window_is_open {window} {
  return [winfo exists $window.outputarea.completion_text]
}

proc completion_window_use_selection {window} {
  global completion_window_info

  set fieldref [focused_fieldref $window]
  set cmdline [fieldref_widget $fieldref]
  set old_text $completion_window_info($window,text)
  $cmdline delete "insert -[string length $old_text]c" "insert"
  set w $window.outputarea.completion_text
  set text [quote_file [$w get sel.first sel.last]]
  set completion_window_info($window,text) $text
  fieldref_insert $fieldref insert $text
}

proc completion_window_select_index {window index} {
  global completion_window_info

  set w $window.outputarea.completion_text
  set range [$w tag prevrange filename "$index +1c"]
  assert {$range!=""}
  set start [lindex $range 0]
  set end [lindex $range 1]
  $w tag remove sel 1.0 end
  $w tag add sel $start $end
  $w see sel.first
  $w see sel.last
}

proc completion_window_selected_row_col {window} {
  set w $window.outputarea.completion_text
  set index [$w index sel.first]
  upvar #0 completion_window_info($window,column_width) column_width
  scan $index "%d.%d" row col
  set row [expr {$row-1}]
  set col [expr {$col/$column_width}]
  return [list $row $col]
}

proc completion_window_row_col_pos {window row col} {
  upvar #0 completion_window_info($window,row_count) row_count
  return [expr {$col*$row_count+$row}]
}

proc completion_window_selected_pos {window} {
  foreach {row col} [completion_window_selected_row_col $window] { }
  return [completion_window_row_col_pos $window $row $col]
}

proc completion_window_select_row_col {window row col} {
  upvar #0 completion_window_info($window,column_width) column_width
  completion_window_select_index $window \
    "[expr {$row+1}].[expr {$col*$column_width}]"
}

proc completion_window_select_pos {window pos} {
  upvar #0 completion_window_info($window,row_count) row_count

  set col [expr {$pos / $row_count}]
  set row [expr {$pos % $row_count}]
  completion_window_select_row_col $window $row $col
}

proc completion_window_select_first {window} {
  completion_window_select_pos $window 0
}

proc completion_window_select_offset {window offset} {
  upvar #0 completion_window_info($window,file_count) file_count
  set pos [completion_window_selected_pos $window]
  incr pos $offset
  if {$pos>=$file_count} {
    incr pos [expr {-$file_count}]
  } elseif {$pos<0} {
    incr pos $file_count
  }
  completion_window_select_pos $window $pos
}

proc completion_window_select_offset_row_col {window rowoffset coloffset} {
  upvar #0 completion_window_info($window,file_count) file_count
  upvar #0 completion_window_info($window,row_count) row_count

  foreach {row col} [completion_window_selected_row_col $window] {}
  incr row $rowoffset
  incr col $coloffset
  if {$row>=0 && $row<$row_count} {
    set pos [completion_window_row_col_pos $window $row $col]
    if {$pos>=0 && $pos<$file_count} {
      completion_window_select_row_col $window $row $col
    }
  }
}

proc completion_window_button_press_1 {window x y} {
  set w $window.outputarea.completion_text
  set index [tkTextClosestGap $w $x $y]
  completion_window_select_index $window $index
}

proc variable_value {termname varname} {
  if {$varname==""} {
    return "\$"
  }
  set firstchar [string index $varname 0]
  if {$firstchar=="\{"} {
    regexp "^\{(.*)\}\$" $varname _match varname
  } elseif {$firstchar=="("} {
    regexp "^.(.*).$" $varname _match command
    return [exec $command]
  }

  upvar #0 $termname term

  global variables
  if {[info exists term(local_variables)] &&
      [klget term(local_variables) $varname value]} {
  } elseif {[info exists variables($varname)]} {
    if {$varname=="HISTORY"} {
      global history
      set value [join $history "\n"]
    } else {
      set value $variables($varname)
    }
  } elseif {[global env;info exists env($varname)]} {
    set value $env($varname)
  } else {
    set value ""
  }
  return $value
}

proc set_variable {termname varname value} {
  upvar #0 $termname term
  if {[info exists term(local_variables)] &&
      [klexists term(local_variables) $varname]} {
    klset term(local_variables) $varname $value
  } elseif {[global env;info exists env($varname)]} {
    set env($varname) $value
  } else {
    global variables
    set variables($varname) $value
  }
  assert {[variable_value $termname $varname]==$value}
}

proc varname_end {string index} {
  set firstchar [string index $string $index]
  if {$firstchar=="*"} {
    return $index
  } elseif {$firstchar=="\{"} {
    return [string_end $string "\[^\}]*\}" $index]
  } elseif {$firstchar=="("} {
    return [string_end $string {([^)]*)\)} $index]
  } else {
    return [string_end $string {[a-zA-Z0-9_]*} $index]
  }
}

proc cmd_arg_nextrange {cmd index} {
  set char [string index $cmd $index]
  while {1} {
    if {$char==" " || $char=="\t"} {
      set char [string index $cmd [incr index]]
    } elseif {$char=="#"} {
      set index [string_end $cmd "\[^\n]*" $index]
      set char [string index $cmd [incr index]]
      break
    } else {
      break
    }
  }
  if {$char==""} {
    return {}
  }
  if {$char=="\n" || $char==";" || $char=="(" || $char==")"} {
    return [list $index $index]
  }
  set start $index
  while {$char!="" && $char!=" " && $char!="\n" && $char!="\t" && $char!=";" 
         && $char!="(" && $char!=")"} {
    if {$char=="\""} {
      set end [getstring_index $cmd $char [incr index]]
      if {$end==-1} {
	return [list $start]
      }
      set index $end
      set char [string index $cmd [incr index]]
    } elseif {$char=="'"} {
      set end [string_end $cmd "\[^']*" [incr index]]
      incr end
      set char [string index $cmd $end]
      if {$char==""} {
	return [list $start]
      }
      assert {$char=="'"}
      set index $end
      set char [string index $cmd [incr index]]
    } elseif {$char=="\\"} {
      set char [string index $cmd [incr index]]
      if {$char==""} {
	error "Got '\\', but no character follows."
      }
      set char [string index $cmd [incr index]]
    } elseif {$char=="\$"} {
      set index [varname_end $cmd [incr index]]
      set char [string index $cmd [incr index]]
    } elseif {$char=="#"} {
      set index [string_end $cmd "\[^\n]*" $index]
      set char [string index $cmd [incr index]]
    } else {
      set index [string_end $cmd "\[^\"'\\\$# \t\n;()]*" $index]
      set char [string index $cmd [incr index]]
    }
  }
  return [list $start [expr {$index-1}]]
}

proc getvarname {stringname} {
  upvar $stringname string
  set end [varname_end $string 0]
  set varname [string range $string 0 $end]
  set string [string range $string [incr end] end]
  return $varname
}

proc term_entry_get {window termname} {
  set entry_range [header_entry_range $window $termname]
  set entry_start [lindex $entry_range 0]
  set entry_end "[lindex $entry_range 1] -1c"
  return [$window.outputarea.text get $entry_start $entry_end]
}

proc close_history_window {window} {
  upvar #0 main_windows($window,history_window) history_window
  if {[info exists history_window]} {
    upvar #0 main_windows($window,history_listbox) history_listbox
    unset history_listbox
    destroy $history_window
    unset history_window
  }
}

proc window_is_compact {window} {
  global main_windows
  return $main_windows($window,is_compact)
}

proc toggle_history_window {window} {
  upvar #0 main_windows($window,history_window) history_window
  upvar #0 main_windows($window,history_listbox) history_listbox
  if {[info exists history_frame]} {
    close_history_window $window
  } else {
    set cmdline_placement [cmdline_placement $window]
    if {$cmdline_placement=="top"} {
      set hw $window.history_window
      toplevel $hw
      wm overrideredirect $hw 1
      set frame $hw.history_frame
      frame $frame
      grid $frame -row 0 -column 0 -sticky nsew
      grid columnconfigure $hw 0 -weight 1
      grid rowconfigure $hw 0 -weight 1
      set history_window $hw
    } else {
      set parent $window.outputarea
      set frame $parent.history_frame
      frame $frame
      set history_window $frame
    }
    let w $frame {
      set listbox $w.listbox
      listbox $listbox -yscrollcommand "$w.scrollbar set" -selectmode extended
      set history_listbox $listbox
      grid $listbox -row 0 -column 0 -sticky nsew
      scrollbar $w.scrollbar -orient vertical -command "$listbox yview"
      grid $w.scrollbar -row 0 -column 1 -sticky ns
      grid columnconfigure $w 0 -weight 1
      grid columnconfigure $w 1 -weight 0

      global history
      foreach item $history {
	$listbox insert end $item
      }
      set history_pos [window_history_position $window]
      if {$history_pos<[llength $history]} {
        $listbox activate [window_history_position $window]
        $listbox selection set active
        $listbox see active
      } else {
	$listbox see end
      }
      bind $listbox <Control-ButtonRelease-1> \
        [list history_window_item_selected $window 1]
      bind $listbox <ButtonRelease-1> \
	[list history_window_item_selected $window 0]
      bind $listbox <Double-ButtonRelease-1> \
	[list after idle history_window_item_double_click $window]
      bind $w <ButtonRelease-1> [list close_history_window $window]
      grab $w
    }
    if {$cmdline_placement=="top"} {
      set cmdline [cmdline_widget $window]
      set width [winfo width $cmdline]
      incr width [winfo width $window.topline.history_arrow]
      set x [winfo rootx $cmdline]
      set y [winfo rooty $cmdline]
      incr y [winfo height $cmdline]
      set height [winfo reqheight $listbox]
      wm geometry $window.history_window ${width}x${height}+$x+$y
    } else {
      place $frame \
	-in $window.outputarea.text -relx 0 -rely 1 -anchor sw -relwidth 1
    }
  }
}

proc window_history_position {window} {
  global main_windows
  return $main_windows($window,history_position)
}

proc term_entry_is_cmdline {window termname} {
  if {[cmdline_in_output $window]} {
    upvar #0 main_windows($window,cmdline_term) cmdline_term
    if {[info exists cmdline_term]} {
      if {$cmdline_term==$termname} {
	return 1
      }
    }
  }
  return 0
}

proc window_set_term_cmdline {window termname cmd} {
  upvar #0 $termname term
  set entry_range [header_entry_range $window $termname]
  set term_widget [term_widget $termname]
  set start [lindex $entry_range 0]
  $term_widget delete $start "[lindex $entry_range 1] -1c"
  $term_widget insert $start $cmd [$term_widget tag names $start]
  set term(cmd) $cmd
}

proc window_set_cmdline {window str} {
  close_completion_window $window
  global config
  if {[cmdline_in_output $window]} {
    window_set_term_cmdline $window [window_cmdline_term $window] $str
  } else {
    set cmdline [cmdline_widget $window]
    cmdline_set $cmdline $str
    set range [cmdline_entry_range $cmdline]
    $cmdline mark set insert [lindex $range 1]
    set line_range [cmdline_line_range $cmdline insert]
    set linestart [lindex $line_range 0]
    set lineend [lindex $line_range 1]
    if {$lineend > $linestart} {
      if {[window_cmdline_mode $window]=="command"} {
	$cmdline mark set insert "insert -1c"
      }
    }
    update_cmdline_height $window
  }
}

proc window_set_history_position {window index} {
  upvar #0 main_windows($window,history_position) history_position
  set history_position $index
  global history
  window_set_cmdline $window [lindex $history $history_position]
}

proc history_window_item_selected {window ctrl} {
  upvar #0 main_windows($window,history_listbox) listbox
  set indices [$listbox curselection]
  if {[llength $indices]==1} {
    window_set_history_position $window $indices
    if {!$ctrl} {
      after 250 [list close_history_window $window]
    }
  } else {
    set lines {}
    global history
    foreach index $indices {
      lappend lines [lindex $history $index]
    }
    window_set_cmdline $window [join $lines "\n"]
  }
}

proc command_entered {window} {
  global config

  set termname [focused_entry_term $window]
  if {$termname!=""} {
    term_command_entered $window $termname
  } else {
    global main_windows

    set cmdline_entry [cmdline_widget $window]
    set cmd [cmdline_get $cmdline_entry]

    if {$cmd==""} {
      if {[window_is_compact $window]} {
	window_set_compact $window 0
	return
      } else {
	global options
	if {$options(enter_blank_toggles_compact)} {
	  window_set_compact $window 1
	  return
	}
      }
    }

    close_history_window $window
    if {[cmdline_at_bottom $cmdline_entry]} {
      set parsed_commands [parse_commands $window cmd]
      if {$parsed_commands!="'"} {
	window_set_cmdline $window ""
	add_to_history $window $cmd

	global options
	if {!$options(expand_only_for_output)} {
	  window_set_compact $window 0
	}
    
	global main_windows
	set dir $main_windows($window,dir)
	set termname [create_decor_term $window $dir $cmd disabled_entry]
	exec_commands $window $termname $parsed_commands
	return
      }
    }
    $cmdline_entry insert insert "\n"
    update_cmdline_height $window
  }
}

proc history_window_item_double_click {window} {
  command_entered $window
}

proc open_completion_window {window} {
  set termname ""
  set fieldref [focused_fieldref $window]
  set cmd [fieldref_get $fieldref]
  set cmdname [command_name $cmd]
  set insert_pos [fieldref_insert_pos $fieldref]
  set index 0
  set pattern "*"
  while {1} {
    set range [cmd_arg_nextrange $cmd $index]
    if {$range=={}} {
      set start ""
      set end ""
      break
    }
    foreach {start end} $range { }
    if {$insert_pos>=$start && $insert_pos<=[expr {$end+1}]} {
      set pattern "[string range $cmd $start $end]*"
      break
    }
    set index [expr {$end+1}]
  }

  global main_windows
  set dir $main_windows($window,dir)
  set files [files_matching $dir $pattern]
  if {$cmdname=="cd"} {
    set dirs {}
    foreach file $files {
      if {[file isdirectory $file]} {
	lappend dirs $file
      }
    }
    set files $dirs
  }
  if {[llength $files]==0} {
    bell
    return
  }
  if {$range!={}} {
    fieldref_set_insert_pos $fieldref [expr {${end}+1}]
  }
  set files [lsort $files]
  let w $window.outputarea.completion_text {
    upvar #0 completion_window_info($window,column_width) column_width
    upvar #0 completion_window_info($window,column_count) column_count
    upvar #0 completion_window_info($window,row_count) row_count
    upvar #0 completion_window_info($window,file_count) file_count
    upvar #0 completion_window_info($window,text) text

    if {$range!={}} {
      set text [string range $cmd $start $end]
    } else {
      set text ""
    }
    set file_count [llength $files]
    set max_len 0
    foreach file $files {
      set len [string length $file]
      if {$len>$max_len} {
	set max_len $len
      }
      file isdirectory $file
    }
    set column_width [expr {$max_len+2}]
    upvar #0 main_windows($window,cols) window_cols
    set column_count [expr {$window_cols/$column_width}]
    set row_count [expr {($file_count+$column_count-1)/$column_count}]

    global theme
    text $w -height $row_count -font $theme(cmdline_font)
    foreach event [bind Text] {
      bind $w $event {break}
    }
    bind $w <ButtonPress-1> \
      "completion_window_button_press_1 $window %x %y;break"
    bind $w <ButtonPress-3> \
      "completion_window_button_3 $window %x %y %X %Y;break"
    $w tag configure sel \
      -foreground [$w cget -background] \
      -background [$w cget -foreground] \
      -relief flat
    for {set row 1} {$row<=$row_count} {incr row} {
      $w insert end "\n" {}
    }
    set row 1
    foreach file $files {
      $w insert "$row.0 lineend" $file {filename}
      set padding [expr {$column_width-[string length $file]}]
      $w insert "$row.0 lineend" [format "%${padding}s" ""] {}
      incr row
      if {$row>$row_count} {
	set row 1
      }
    }
    $w delete "end -1c" end
    completion_window_select_first $window
    set w_height [winfo reqheight $w]
    set t_height [winfo reqheight $window.outputarea.text]
    if {$w_height>$t_height} {
      puts "making it $t_height instead of $w_height"
      set w_height $t_height
    }
    place $w \
      -in $window.outputarea.text \
      -relx 0 \
      -rely 1 \
      -relwidth 1 \
      -height $w_height \
      -anchor sw \
      -border ignore
  }
}

proc close_completion_window {window} {
  if {[completion_window_is_open $window]} {
    set w $window.outputarea.completion_text
    place forget $w
    destroy $w
  }
}

# Find the number of lines necessary to display all the text in the given
# text widget.
proc full_text_height {w} {
  set font [$w cget -font]
  set borderwidth [$w cget -borderwidth]
  set padx [$w cget -padx]
  set pady [$w cget -pady]
  set highlightthickness [$w cget -highlightthickness]
  set pixel_width [winfo width $w]
  set pixel_height [winfo height $w]
  incr pixel_width [expr {-2*($borderwidth+$padx+$highlightthickness)}]
  incr pixel_height [expr {-2*($borderwidth+$pady+$highlightthickness)}]
  set cwidth [font measure $font -displayof $w "W"]
  set cheight [font metrics $font -displayof $w -linespace]
  set index 1.0
  set line_count 0
  while {[$w compare $index < end]} {
    set line [$w get $index "$index lineend"]
    if {$line==""} {
      incr line_count 1
    } else {
      set linewidth [font measure $font -displayof $w $line]
      incr line_count [expr {($linewidth-1)/$pixel_width+1}]
    }
    set index [$w index "$index +1 lines"]
  }
  return $line_count
}

proc cmdline_modified {window} {
  history_position_end $window
  if {![cmdline_in_output $window]} {
    update_cmdline_height $window
  }
}

proc entry_delete_selected {widget entry_range} {
  set entry_start [lindex $entry_range 0]
  set entry_end [lindex $entry_range 1]
  set delranges [tag_ranges $widget sel $entry_start $entry_end]
  if {[llength $delranges]==0} {
    return 0
  }
  foreach delrange $delranges {
    set delstart [lindex $delrange 0]
    set delend [lindex $delrange 1]
    $widget mark set insert $delstart
    if {[$widget compare $delend == $entry_end]} {
      $widget delete $delstart "$delend -1c"
    } else {
      $widget delete $delstart $delend
    }
  }
  return 1
}

proc handle_vi_key {fieldref statevar key commands} {
  upvar $statevar statelist
  array set state $statelist
  set widget [fieldref_widget $fieldref]
  set entry_range [fieldref_range $fieldref]
  set entry_start [lindex $entry_range 0]
  set entry_end [lindex $entry_range 1]
  set orig_line [fieldref_get $fieldref]
  set orig_insert_pos [fieldref_insert_pos $fieldref]
  set change_made 0
  set line_range [fieldref_line_range $fieldref insert]
  assert {[llength $line_range]==2}
  set linestart [lindex $line_range 0]
  set lineend [lindex $line_range 1]
  switch -- $state(mode) {
    command {
      set command_complete 0
      set found 0
      set next_move_index ""
      set move_index ""
      set state(current_cmd) "$state(current_cmd)$key"
      if {!$found} {
	set found 1
	switch -- $state(cmd) {
	  r {
	    $widget delete insert
	    $widget insert insert $key [$widget tag names insert]
	    $widget mark set insert "insert -1c"
	    set state(cmd) ""
	  }
	  default { set found 0 }
	}
      }
      if {!$found} {
	set found 1
	switch -- $state(findcmd) {
	  t -
	  T -
	  f -
	  F {
	    set move_index [vi_char_search $widget $state(findcmd) $key]
	    if {$move_index!=""} {
	      switch -- $state(findcmd) {
		F -
		T { set next_move_index $move_index }
		t -
		f { set next_move_index [$widget index "$move_index +1c"]}
	      }
	      set state(last_charsearch) $state(findcmd)$key
	    }
	    set command_complete 1
	  }
	  default { set found 0 }
	}
      }
      set state(findcmd) ""
      if {!$found} {
	set found 1
	switch -- $key {
	  0 { set move_index $linestart }
	  $ { 
	    set move_index "$lineend -1c" 
	    set next_move_index $lineend
	  }
	  E -
	  e -
	  B -
	  b -
	  w -
	  W {
	    if {$key=="b" || $key=="e" || $key=="w"} {
	      set pattern {[0-9a-zA-Z_]}
	    } else {
	      set pattern {[^ ]}
	    }
	    if {$key=="B" || $key=="b"} {
	      set offset "-1c"
	    } else {
	      set offset "+1c"
	    }
	    if {$key=="w" || $key=="W"} {
	      set word 1
	    } else {
	      set word 0
	    }
	    set index [$widget index insert]
	    set end [$widget index $entry_end]
            # Always advance one character
	    if {[$widget compare "$index $offset" < $end]} {
	      set index [$widget index "$index $offset"]
	      if {[$widget compare $index < $entry_start]} {
		set index $entry_start
	      }
	    }
	    if {$word} {
	      # First advance to the end of the current word
	      set char [$widget get $index]
	      set norm [regexp $pattern $char]
	      while {[$widget compare $index > $entry_start] && 
		     [$widget compare "$index $offset" < $end]} {
		set char [$widget get $index]
		if {$char==" "} {
		  break
		}
		set newnorm [regexp $pattern $char]
		if {$newnorm!=$norm} {
		  break
		}
		set index [$widget index "$index $offset"]
	      }
	    }
	    while {[$widget compare $index > $entry_start] && 
	           [$widget compare "$index $offset" < $end]} {
	      set char [$widget get $index]
	      if {$char!=" "} { break }
	      set index [$widget index "$index $offset"]
	    }
	    if {!$word} {
	      set char [$widget get "$index"]
	      set norm [regexp $pattern $char]
	      while {[$widget compare $index > $entry_start] && 
		     [$widget compare "$index $offset" < $end]} {
		set char [$widget get "$index $offset"]
		if {$char==" "} { break }
		set newnorm [regexp $pattern $char]
		if {$newnorm!=$norm} { 
		  break 
		}
		set index [$widget index "$index $offset"]
	      }
	    }
	    set move_index $index
	    if {$key=="e" || $key=="E"} {
	      set next_move_index [$widget index "$index $offset"]
	    }
	  }
	  h { 
	    if {[$widget compare insert > $linestart]} {
	      set move_index "insert -1c" 
	    }
	  }
	  l { 
	    if {[$widget compare insert < "$lineend -1c"]} {
	      set move_index "insert +1c" 
	    }
	  }
	  ; { 
	    if {[info exists state(last_charsearch)]} {
	      puts "last_charsearch='$state(last_charsearch)'"
	      set move_index [
		vi_char_search $widget \
		  [string index $state(last_charsearch) 0] \
		  [string index $state(last_charsearch) 1]
	      ]
	    } else {
	      bell
	    }
	  }
	  f -
	  F -
	  t -
	  T { set state(findcmd) $key }
	  default { set found 0 }
	}
	if {$found && $state(findcmd)==""} {
	  set command_complete 1
	}
      }
      if {$state(cmd)!=""} {
	if {!$found} {
	  bell
	  set found 1
	  set command_complete 1
	} else {
	  if {$move_index!=""} {
	    switch -- $state(cmd) {
	      d -
	      c { 
		if {$next_move_index!=""} {
		  set move_index $next_move_index
		}
		if {[$widget compare $move_index < insert]} {
		  $widget delete $move_index insert
		} else {
		  $widget delete insert $move_index
		}
		if {$state(cmd)=="c"} {
		  set state(mode) insert
		}
	      }
	      default { error "Unknown state(cmd)=$state(cmd)" }
	    }
	  }
	}
      } else {
	if {$move_index!=""} {
	  $widget mark set insert $move_index 
	}
      }
      if {!$found} {
	set found 1
	switch -- $key {
	  k { switch -- up $commands }
	  j { switch -- down $commands }
	  default { set found 0 }
	}
	if {$found} {
	  set command_complete 1
	}
      }
      if {!$found} {
	set found 1
	switch -- $key {
	  Y { 
	    set state(yanked) [
	      $widget get $linestart "$lineend +1c"
	    ] 
	    set state(yanktype) line
	  }
	  default { set found 0 }
	}
      }
      if {!$found} {
	set found 1
	switch -- $key {
	  P {
	    switch -- $state(yanktype) {
	      line {
		set range \
		  [tag_range \
		    $widget [tag names insert] $linestart "insert"]
		$widget insert [lindex $range 0] $state(yanked)
	      }
	      char {
		$widget insert insert $state(yanked)
	      }
	      default {
		bell
	      }
	    }
	  }
	  p {
	    switch -- $state(yanktype) {
	      line {
		$widget insert "insert +1 lines linestart" $state(yanked)
		$widget mark set insert "insert +1 lines linestart"
	      }
	      char {
		$widget insert "insert +1c" $state(yanked)
		$widget mark set insert "insert +1c"
	      }
	      default {
		bell
	      }
	    }
	  }
	  D { $widget delete insert $lineend }
	  x {
	    if {[$widget compare insert < $lineend]} {
	      set state(yanked) [$widget get insert]
	      set state(yanktype) char
	      $widget delete insert 
	      if {[$widget compare insert == $lineend]} {
		if {[$widget compare insert > $linestart]} {
		  $widget mark set insert "insert -1c"
		}
	      }
	    }
	  }
	  X { 
	    if {[$widget compare insert > $entry_start]} {
	      $widget delete "insert -1c" 
	    }
	  }
	  default { set found 0 }
	}
	if {$found} {
	  set change_made 1
	  set state(last_cmd) $key
	  set command_complete 1
	}
      }
      if {!$found} {
	set found 1
	switch -- $key {
	  A {
	    $widget mark set insert $lineend
	    set state(mode) insert 
	  }
	  a {
	    if {[$widget compare insert < $lineend]} {
	      $widget mark set insert "insert +1c"
	    }
	    set state(mode) insert
	  }
	  i { set state(mode) insert }
	  I { 
	    $widget mark set insert $linestart
	    set state(mode) insert 
	  }
	  C { 
	    $widget delete insert $lineend
	    set state(mode) insert 
	  }
	  r { 
	    if {[$widget compare insert == "insert lineend"]} { 
	      bell
	      return 
	    }  
	  }
	  s {
	    $widget delete insert
	    set state(mode) insert
	  }
	  d { set state(cmd) d }
	  c { set state(cmd) c }
	  default {
	    set found 0
	  }
	}
	if {$found} {
	  set state(cmd) $key
	  set change_made 1
	}
      }
      if {!$found} {
	set found 1
	switch -- $key {
	  u {
	    if {[info exists state(undo_buffer)]} {
	      fieldref_set $fieldref $state(undo_buffer)
	      puts "undo_insert_pos=$state(undo_insert_pos)"
	      $widget mark set insert "$entry_start +$state(undo_insert_pos)c"
	      set change_made 1
	    } else {
	      bell
	    }
	  }
	  U {
	    if {[info exists state(undo_buffer)]} {
	      $widget delete $linestart $lineend
	      $widget insert $linestart $state(orig_line)
	      $widget mark set insert $linestart
	      set change_made 1
	    } else {
	      bell
	    }
	  }
	  . {
	    if {[info exists state(last_cmd)]} {
	      set last_cmd $state(last_cmd)
	      set len [string length $last_cmd]
	      for {set i 0} {$i<$len} {incr i} {
		handle_vi_key \
		  $fieldref statelist [string index $last_cmd $i] $commands
	      }
	    } else {
	      bell
	    }
	  }
	  default {
	    set found 0
	  }
	}
	if {$found} {
	  set command_complete 1
	  set state(current_cmd) ""
	  set change_made 1
	}
      }
      if {!$found} {
        if {$key=="\x3"} {
          switch -- abort $commands
        }
        return
      }
      if {!$found} {
	bell
      }
      if {$command_complete} {
	if {$state(current_cmd)!="" && $change_made} {
	  set state(last_cmd) $state(current_cmd)
	}
	set state(current_cmd) ""
	set state(cmd) ""
	set state(find_cmd) ""
      }
    }
    insert {
      switch -- $key {
	"\x1b" { 
	  if {[$widget compare insert > $linestart]} {
	    $widget mark set insert "insert -1c"
	  }
	  set state(mode) command 
	  set state(last_cmd) "$state(cmd)$state(inserted)$key"
	  set state(inserted) ""
	  set state(cmd) ""
	}
	"\x18" -
	"\x03" {
	  # Cut
	  set sel_ranges [tag_ranges $widget sel $entry_start $entry_end]
	  if {[llength $sel_ranges]==0 && $key=="\x03"} {
	    switch -- abort $commands
	  } else {
	    set cut_buffer ""
	    foreach sel_range $sel_ranges {
	      set start [lindex $sel_range 0]
	      set end [lindex $sel_range 1]
	      append cut_buffer [$widget get $start $end]
	      if {$key=="\x18"} {
		$widget delete $start $end
	      } else {
		$widget tag remove sel $start $end
	      }
	    }
	    clipboard clear
	    clipboard append $cut_buffer
	  }
	}
	"\x16" {
	  # Paste
	  if {![catch {selection get -selection CLIPBOARD} clipboard]} {
	    $widget insert insert $clipboard [$widget tag names insert]
	  }
	}
	"\b" { 
	  set sel_ranges [tag_ranges $widget sel $entry_start $entry_end]
	  if {[llength $sel_ranges]>0} {
	    foreach sel_range $sel_ranges {
	      $widget delete [lindex $sel_range 0] [lindex $sel_range 1]
	    }
	  } else {
	    set len [string length $state(inserted)]
	    if {[$widget compare insert > $entry_start]} {
	      $widget delete "insert -1c"
	    }
	    if {$len>0} {
	      incr len -2
	      set state(inserted) [string range $state(inserted) 0 $len]
	    }
	  }
	}
	"\x7f" {
	  if {[$widget tag nextrange sel 1.0 end] != ""} {
	    $widget delete sel.first sel.last
	  } else {
	    if {[$widget compare insert < "end -2c"]} {
	      $widget delete insert
	    }
	  }
	}
	default {
	  entry_delete_selected $widget $entry_range
	  $widget insert insert $key [$widget tag names insert]
	  set state(inserted) "$state(inserted)$key"
	}
      }
      switch -- modified $commands
    }
  }
  if {$change_made} {
    set state(undo_buffer) $orig_line
    set state(undo_insert_pos) $orig_insert_pos
    switch -- modified $commands
  }
  set statelist [array get state]
}

proc cmdline_fieldref {cmdline} {
  set range [entry_range $cmdline 1.0 "end"]
  set end [lindex $range 1]
  set end [$cmdline index "$end -1c"]
  set range [list [lindex $range 0] $end]
  return [fieldref $cmdline $range]
}

proc cmdline_line_range {cmdline index} {
  return [fieldref_line_range [cmdline_fieldref $cmdline] $index]
}

proc cmdline_index {cmdline start motion} {
  return [fieldref_index [cmdline_fieldref $cmdline] $start $motion]
}

proc term_entry_index {window termname start motion} {
  return \
    [fieldref_index [term_entry_fieldref $window $termname] $start $motion]
}

proc term_entry_fieldref {window termname} {
  set range [header_entry_range $window $termname]
  set widget $window.outputarea.text
  set end [lindex $range 1]
  set end [$widget index "$end -1c"]
  set range [list [lindex $range 0] $end]
  return [fieldref $widget $range]
}

proc find_history {window prefix direction} {
  global history
  upvar #0 main_windows($window,history_position) history_position

  set histlen [llength $history]
  set newpos [expr $history_position + $direction]
  if {$newpos>=$histlen} {
    return $histlen
  }
  set matchlen [string length $prefix]

  while {$newpos >= 0 && $newpos<$histlen} {
    set histent [lindex $history $newpos]

    set histfirst [string range $histent 0 [expr $matchlen - 1]]
    if {$histfirst==$prefix} {
      break;
    }
    incr newpos $direction
  }
  return $newpos
}

proc go_history {window fieldref pending_cmd_var direction} {
  global history
  upvar #0 main_windows($window,history_position) history_position

  set widget [fieldref_widget $fieldref]
  upvar #0 $pending_cmd_var pending_cmd

  if {$history_position==[llength $history]} {
    if {$direction>0} {
      return
    }
    set pending_cmd [fieldref_get $fieldref]
  }
  set matchstr $pending_cmd
  set matchlen [string length $matchstr]
  set histlen [llength $history]

  set newpos [find_history $window $matchstr $direction]

  if {$newpos>=0} {
    set history_position $newpos
    if {$newpos==$histlen} {
      fieldref_set $fieldref $matchstr
    } {
      set cmd [lindex $history $history_position]
      fieldref_set $fieldref $cmd
    }
    upvar #0 main_windows($window,history_listbox) history_listbox
    if {[info exists history_listbox]} {
      $history_listbox selection clear 0 end
      $history_listbox selection set $history_position
      $history_listbox see $history_position
    }
  }
}

#proc cmdline_go_history {window direction} {
#  upvar #0 main_windows($window,pending_cmd) pending_cmd
#  set cmdline [cmdline_widget $window]
#  set fieldref [cmdline_fieldref $cmdline]
#  go_history $window $fieldref pending_cmd $direction
#  update_cmdline_height $window
#}

proc fieldref_go_history {window fieldref direction} {
  upvar #0 main_windows($window,pending_cmd) pending_cmd
  go_history $window $fieldref pending_cmd $direction
  if {![cmdline_in_output $window]} {
    if {[fieldref_widget $fieldref]==[cmdline_widget $window]} {
      update_cmdline_height $window
    }
  }
}

proc term_entry_go_history {window termname direction} {
  upvar #0 $termname term
  set fieldref [term_entry_fieldref $window $termname]
  go_history $window $fieldref term(pending_cmd) $direction
}

#proc term_entry_up {window termname} {
#  set fieldref [term_entry_fieldref $window $termname]
#  if {[fieldref_at_top $fieldref]} {
#    term_entry_go_history $window $termname -1
#  } else {
#    fieldref_move_insert $fieldref "-1 lines"
#  }
#}

#proc term_entry_down {window termname} {
#  set fieldref [term_entry_fieldref $window $termname]
#  if {[fieldref_at_bottom $fieldref]} {
#    term_entry_go_history $window $termname 1
#  } else {
#    fieldref_move_insert $fieldref "+1 lines"
#  }
#}

proc cmdline_at_bottom {cmdline} {
  return [fieldref_at_bottom [cmdline_fieldref $cmdline]]
}

proc cmdline_at_top {cmdline} {
  return [fieldref_at_top [cmdline_fieldref $cmdline]]
}

proc fieldref_up {window fieldref} {
  if {[fieldref_at_top $fieldref]} {
    fieldref_go_history $window $fieldref -1
  } else {
    fieldref_move_insert $fieldref "-1 lines"
  }
}

proc fieldref_down {window fieldref} {
  if {[fieldref_at_bottom $fieldref]} {
    fieldref_go_history $window $fieldref 1
  } else {
    fieldref_move_insert $fieldref "+1 lines"
  }
}

proc focus_next_term_entry {window} {
  set termname [focused_entry_term $window]
  set next_term [term_next $termname 1]
  if {$next_term!=""} {
    focus_term_entry $window $next_term
  } else {
    global main_windows
    if {[info exists main_windows($window,cmdline_entry)]} {
      $window.outputarea.text see bottom
      focus_cmdline $window
    } else {
      set next_term [term_next $termname 0]
      if {$next_term!=""} {
	term_focus $next_term
      }
    }
  }
}

proc remove_term {window termname} {
  upvar #0 $termname term 
  set term_widget [term_widget $termname]
  set focused_term [focused_entry_term $window]
  if {$focused_term==$termname} {
    focus_next_term_entry $window
  }
  set prev_term [term_prev $termname 0]
  if {$prev_term!=""} {
    set prev_top_index [$term_widget index ${prev_term}top]
    set prev_bottom_index [$term_widget index ${prev_term}bottom]
  }
  if {[term_next $termname 0]==""} {
    set bottom bottom
  } else {
    set bottom "${termname}bottom"
  }
  $term_widget delete "${termname}header" $bottom
  $term_widget mark unset ${termname}header
  delete_term $termname
  if {$prev_term!=""} {
    $term_widget mark set ${prev_term}top $prev_top_index
    $term_widget mark set ${prev_term}bottom $prev_bottom_index
  }
}

proc remove_prev_term {window termname} {
  set prev_term [term_prev $termname 1]
  if {$prev_term==""} {
    bell
  } else {
    remove_term $window $prev_term
  }
}

proc remove_last_term {window} {
  set widget $window.outputarea.text
  set last_term [term_last $widget]
  if {$last_term!="" && ![term_is_active $last_term]} {
    $widget delete ${last_term}bottom bottom
    remove_term $window $last_term
  } else {
    bell
  }
}

proc window_close {window} {
  set term_list [term_list $window.outputarea.text]
  foreach termname $term_list {
    delete_term $termname
  }

  global main_window_list
  set index [lsearch -exact $main_window_list $window]
  set main_window_list [lreplace $main_window_list $index $index]
  if {[llength $main_window_list]==0} {
    exit
  }
  destroy $window
}

proc switch_compact {window} {
  window_set_compact $window [expr {![window_is_compact $window]}]
  return
}

proc fieldref_key {window termname fieldref key} {
  if {[completion_window_is_open $window]} {
    set other_key 0
    switch -- $key {
      Tab { completion_window_select_offset $window 1 }
      Shift-Tab { completion_window_select_offset $window -1 }
      Right { completion_window_select_offset_row_col $window 0 1 }
      Left { completion_window_select_offset_row_col $window 0 -1 }
      Up { completion_window_select_offset_row_col $window -1 0 }
      Down { completion_window_select_offset_row_col $window 1 0 }
      default { set other_key 1 }
    }
    if {!$other_key} {
      return
    }
  }
  if {[completion_window_is_open $window]} {
    if {$key=="\r" || $key=="\n" || $key==" " || $key=="/"} {
      completion_window_use_selection $window
    }
    close_completion_window $window
    if {$key=="\x1b"} {
      return
    }
  }
  global main_windows
#  set entry_range [cmdline_entry_range $cmdline]
#  assert {[llength $entry_range]==2}
#  set entry_start [lindex $entry_range 0]
#  set entry_end [lindex $entry_range 1]
  set widget [fieldref_widget $fieldref]
  set line_range [fieldref_line_range $fieldref insert]
  set linestart [lindex $line_range 0]
  set lineend [lindex $line_range 1]

  switch -glob -- $key {
    Ctrl-Return {help_command $window}
    Ctrl-BackSpace {
      if {$termname==""} {
	remove_last_term $window
      } else {
	remove_prev_term $window $termname
      }
    }
    Ctrl-End {
      focus_cmdline $window 
    }
    Ctrl-Delete { 
      if {![cmdline_focused $window]} {
	remove_term $window $termname 
      } else {
	bell
      }
    }
    Ctrl-Up {
      if {$termname==""} {
	set prev_term [term_last $window.outputarea.text]
      } else {
	set prev_term [term_prev $termname 1]
      }
      if {$prev_term!=""} {
	focus_term_entry $window $prev_term
      }
    }
    Ctrl-Down { focus_next_term_entry $window }
    F2 { switch_compact $window }
    Up { fieldref_up $window $fieldref }
    Shift-Up { 
      tkTextKeySelect $widget [fieldref_index $fieldref "insert" "-1 lines"]
    }
    Down { fieldref_down $window $fieldref }
    Shift-Down { 
      tkTextKeySelect $widget [fieldref_index $fieldref "insert" "+1 lines"]
    }
    Left { 
      fieldref_move_insert $fieldref "-1c"
    }
    Shift-Left { 
      tkTextKeySelect $widget [fieldref_index $fieldref "insert" "-1c"]
    }
    Right { 
      fieldref_move_insert $fieldref "+1c"
    }
    Shift-Right { 
      tkTextKeySelect $widget [fieldref_index $fieldref "insert" "+1c"]
    }
    End { $widget mark set insert $lineend}
    Shift-End { tkTextKeySelect $widget $lineend }
    Home { $widget mark set insert $linestart }
    Shift-Home { tkTextKeySelect $widget $linestart }
    "\x9" -
    Tab { 
      if {[completion_window_is_open $window]} {
	completion_window_select_offset $window 1
      } else {
	open_completion_window $window 
      }
    }
    Shift-Tab {
      if {![completion_window_is_open $window]} {
	open_completion_window $window
      }
      completion_window_select_offset $window -1
    }
    ? {
      switch -- $key {
	"\n" -
	"\r" {
	  command_entered $window
	  window_init_vi_state $window
	}
	"\x4" { 
          # ctrl-d
	  set cmd [fieldref_get $fieldref]
	  if {$cmd==""} {
	    window_close $window 
	  } else {
	    help_command $window
	  }
	}
	default { 
	  upvar #0 main_windows($window,cmdline_state) cmdline_state
	  handle_vi_key $fieldref cmdline_state $key \
            [list modified [list cmdline_modified $window] \
                  up       [list fieldref_up $window $fieldref] \
                  down     [list fieldref_down $window $fieldref] \
                  abort    [list abort_cmd $window]]
	}
      }
    }
    default {
    }
  }
}

proc cmdline_key {window key} {
  set cmdline [cmdline_widget $window]
  fieldref_key $window "" [cmdline_fieldref $cmdline] $key
}

proc entry_unselect {window termname} {
  upvar #0 $termname term
  set widget [term_widget $termname]
  set entry_range [header_entry_range $window $termname]
  set entry_start [lindex $entry_range 0]
  set entry_end [lindex $entry_range 1]
  $widget tag remove sel $entry_start $entry_end
}

proc cmdline_delete_selected {cmdline} {
  set entry_range [cmdline_entry_range $cmdline]
  set entry_start [lindex $entry_range 0]
  set entry_end [lindex $entry_range 1]
  entry_delete_selected $cmdline $entry_start $entry_end
}

proc term_entry_delete_selected {window termname} {
  upvar #0 $termname term

  set widget [term_widget $termname]
  set entry_range [header_entry_range $window $termname]
  entry_delete_selected $widget $entry_range
}

proc get_arg {cmd_line cmd_line_index_var} {
  upvar $cmd_line_index_var cmd_line_index
  assert {$cmd_line_index>=0}
  set range [cmd_arg_nextrange $cmd_line $cmd_line_index]
  if {[llength $range]==0} {
    return {}
  }
  if {[llength $range]==1} {
    return "'"
  }
  set start [lindex $range 0]
  set end [lindex $range 1]
  set arg [string range $cmd_line $start $end]
  set cmd_line_index $end
  incr cmd_line_index
  assert {$cmd_line_index>=0}
  return $arg
}

proc first_arg_range {cmd index} {
  while {1} {
    set range [cmd_arg_nextrange $cmd $index]
    if {[llength $range]<2} {
      return $range
    }
    set arg [string range $cmd [lindex $range 0] [lindex $range 1]]
    if {$arg!="\n"} {
      return $range
    }
    set index [lindex $range 1]
    incr index
  }
}

proc first_arg {cmd index} {
  set range [first_arg_range $cmd $index]
  switch -- [llength $range] {
    0 -
    1 { return "'" }
    2 { return [string range $cmd [lindex $range 0] [lindex $range 1]] }
  }
  assert {0}
}

# Find the first argument of the next command.  If there is not next command,
# consider it to be incomplete.
proc get_first_arg {cmd_line cmd_line_index_var} {
  upvar $cmd_line_index_var cmd_line_index
  set range [first_arg_range $cmd_line $cmd_line_index]
  switch -- [llength $range] {
    0 { set cmd_line_index [string length $cmd_line]; return "'" }
    1 { return "'" }
    2 { 
      set end [lindex $range 1]
      set arg [string range $cmd_line [lindex $range 0] $end]
      incr end
      set cmd_line_index $end
      return $arg
    }
  }
  assert {0}
}

proc get_command {cmd_line cmd_line_index_var} {
  upvar $cmd_line_index_var cmd_line_index

  if {[string index $cmd_line $cmd_line_index]=="@"} {
    set end [string_end $cmd_line "\[^\n]*" $cmd_line_index]
    if {$end<0} {
      return "'"
    }
    set command [string range $cmd_line $cmd_line_index $end]
    incr end
    set cmd_line_index $end
    return $command
  }
  set range [first_arg_range $cmd_line $cmd_line_index]
  switch -- [llength $range] {
    0 { set cmd_line_index [string length $cmd_line]; return "" }
    1 { return "'" }
  }
  set start [lindex $range 0]
  set end [lindex $range 1]
  set cmdname [string range $cmd_line $start $end]

  if {$cmdname=="\}" || $cmdname==")"} {
    return {}
  }
  incr end
  set cmd_line_index $end

  if {$cmdname=="("} {
    set command "("
    while {1} {
      set arg [get_arg $cmd_line cmd_line_index]
      if {$arg=="'" || $arg==""} {
	return "'"
      }
      append command " $arg"
      if {$arg==")"} {
	break
      }
    }
    while {1} {
      set arg [get_arg $cmd_line cmd_line_index]
      if {$arg==";" || $arg=="\n" || $arg==""} {
	break
      }
      append command " $arg"
    }
    return $command
  }

  global aliases
  set size_change 0
  if {[info exists aliases($cmdname)]} {
    set cmd_before [string range $cmd_line 0 [expr {$start-1}]]
    set cmd_after [string range $cmd_line $cmd_line_index end]
    set size_change [string length $cmd_line]
    set cmd_line "$cmd_before$aliases($cmdname) $cmd_after"
    set size_change [expr {$size_change-[string length $cmd_line]}]
    set cmd_line_index $start
    set range [first_arg_range $cmd_line $cmd_line_index]
    switch -- [llength $range] {
      0 { set cmd_line_index [string length $cmd_line]; return "" }
      1 { return "'" }
    }
    set start [lindex $range 0]
    set end [lindex $range 1]
    set cmdname [string range $cmd_line $start $end]
    incr end
    set cmd_line_index $end
  } 
  if {$cmdname=="function"} {
    set command [get_function $cmd_line cmd_line_index]
  } elseif {$cmdname=="if"} {
    set command {}
    while {1} {
      # Process condition command
      set test [get_command $cmd_line cmd_line_index]
      if {$test=="'"} { return $test }
      set arg [get_first_arg $cmd_line cmd_line_index]
      if {$arg=="'"} { return $arg }
      if {$arg!="then"} { return "@error {Expecting 'then', found '$arg'}" }
      # Process consequence commands
      set commands [get_commands "" cmd_line cmd_line_index]
      if {$commands=="'"} { return "'" }
      lappend command $cmdname $test then $commands
      set arg [get_arg $cmd_line cmd_line_index]
      if {$arg=="'" || $arg==""} { return "'" }
      if {$arg=="else" || $arg=="fi"} { break }
      if {$arg!="elif"} { return "@error {Expecting 'elif', but found '$arg'}" }
      set cmdname $arg
    }
    if {$arg=="else"} {
      set commands [get_commands "" cmd_line cmd_line_index]
      if {$commands=="'"} { return $commands }
      lappend command else $commands
      set arg [get_arg $cmd_line cmd_line_index]
      if {$arg=="'"} { return "'" }
    }
    if {$arg!="fi"} { return "@error {Expecting 'fi', but found '$arg'}" }
  } elseif {$cmdname=="for"} {
    set varname [get_arg $cmd_line cmd_line_index]
    if {$varname=="'"} { return "'" }
    if {$varname==";" || $varname=="\n"  || $varname==""} {
      return "@error {No control variable specified}"
    }
    set arg [get_arg $cmd_line cmd_line_index]
    if {$arg=="'"} { return $arg }
    if {$arg=="in"} {
      set list ""
      while {1} {
	set arg [get_arg $cmd_line cmd_line_index]
	if {$arg=="'" || $arg==""} { return "'" }
	if {$arg==";" || $arg=="\n"} { break }
	set list "$list $arg"
      }
      set arg [get_first_arg $cmd_line cmd_line_index]
      if {$arg=="'"} { return "'" }
      if {$arg!="do"} { return "@error {Expecting 'do', found '$arg'}" }
    } elseif {$arg=="do"} {
      return "@error {Use positional parameters: not implemented}"
    }
    set commands [get_commands "" cmd_line cmd_line_index]
    if {$commands=="'"} { return $commands }
    set arg [get_first_arg $cmd_line cmd_line_index]
    if {$arg=="'" || $arg==""} { return "'" }
    if {$arg!="done"} { return "@error {Expecting 'done', found '$arg'}" }
    set command [list for $varname $list $commands]
  } elseif {$cmdname=="while"} {
    set testcmd [get_command $cmd_line cmd_line_index]
    if {$testcmd=="'" || $testcmd==""} { return "'" }
    set arg [get_first_arg $cmd_line cmd_line_index]
    if {$arg=="'"} { return $arg }
    if {$arg!="do"} { return "@error {Expecting 'do', found '$arg'}" }
    set body [get_commands "" cmd_line cmd_line_index]
    if {$body=="'"} { return "'" }
    set arg [get_first_arg $cmd_line cmd_line_index]
    if {$arg=="'"} { return "'" }
    if {$arg!="done"} { return "@error {Expecting 'done', found '$arg'}" }
    set command [list while $testcmd $body]
  } else {
    set command ""
    set arg $cmdname
    while {1} {
      if {$arg==";" || $arg=="\n" || $arg==""} {
	break;
      }
      if {$command!=""} {
	append command " $arg"
      } else {
	append command $arg
      }
      set range [cmd_arg_nextrange $cmd_line $cmd_line_index]
      switch -- [llength $range] {
	0 { 
	  while {[string index $cmd_line $cmd_line_index]==" "} {
	    incr cmd_line_index
	  }
	  break 
	}
	1 { return "'" }
      }
      set start [lindex $range 0]
      set end [lindex $range 1]
      set arg [string range $cmd_line $start $end]
      if {$arg==")"} {
	break
      }
      set cmd_line_index [incr end]
    }
  }
  incr cmd_line_index $size_change
  return $command
}

proc get_commands {window cmd_line_var cmd_line_index_var} {
  upvar $cmd_line_var cmd_line
  upvar $cmd_line_index_var cmd_line_index
  set parsed_commands {}
  while {$cmd_line_index<[string length $cmd_line]} {
    set first_arg [first_arg $cmd_line $cmd_line_index]
    switch -- $first_arg {
      "\}" -
      ")" -
      "else" -
      "elif" -
      "fi" -
      "done" { break }
    }
    set cmd_start $cmd_line_index
    set command [get_command $cmd_line cmd_line_index]
    if {$command=="'"} {
      return $command
    }
    if {$window!=""} {
      if {$command=="!!"} {
	set findstr ""
      } elseif {[string index $command 0]=="!"} {
	set findstr [string range $command 1 end]
      } else {
	set command_index 0
	set command_arg0 [get_arg $command command_index]
	if {$command_arg0=="r"} {
	  set findstr [get_arg $command command_index]
	}
      }
      if {[info exists findstr]} {
	set histpos [find_history $window $findstr -1]
	if {$histpos>=0} {
	  global history
	  set cmd [lindex $history $histpos]
	  set cmd_line_before [string range $cmd_line 0 [expr {$cmd_start-1}]]
	  set cmd_line_after [string range $cmd_line $cmd_line_index end]
	  set cmd_line "$cmd_line_before$cmd$cmd_line_after"
	  set cmd_line_index $cmd_start
	  set command ""
	}
	unset findstr
      }
    }
    if {$command!=""} {
      lappend parsed_commands $command
    }
  }
  return $parsed_commands
}

proc get_function {cmd_line cmd_line_index_var} {
  upvar $cmd_line_index_var cmd_line_index
  set parsed_function {function}
  set arg [get_arg $cmd_line cmd_line_index]
  if {$arg=="function"} {
    set arg [get_arg $cmd_line cmd_line_index]
  }
  if {$cmd_line_index==[string length $cmd_line]} {
    return "@error {Missing function name}"
  }
  set funcname $arg
  lappend parsed_function $funcname
  set arg [get_arg $cmd_line cmd_line_index]
  if {$arg=="("} {
    set arg [get_arg $cmd_line cmd_line_index]
    if {$arg!=")"} {
      return "@error {Unmatched '('}"
    }
    set arg [get_arg $cmd_line cmd_line_index]
  }
  while {$arg=="\n"} {
    set arg [get_arg $cmd_line cmd_line_index]
  }
  if {$arg!="\{"} {
    return "@error \"Missing '\{' in function definition\""
  }
  set body [get_commands "" cmd_line cmd_line_index]
  if {$body=="'"} {
    return $body
  }
  while {1} {
    set arg [get_arg $cmd_line cmd_line_index]
    if {$arg!="\n"} {
      break
    }
  }
  if {$arg!="\}"} {
    return "'"
  }
  lappend parsed_function $body
  return $parsed_function
}

proc parse_commands {window cmd_line_var} {
  upvar $cmd_line_var cmd_line
  set cmd_line_index 0
  set commands [get_commands $window cmd_line cmd_line_index]
  return $commands
}

proc give_cmdline {window} {
  if {![winfo exists $window]} {
    return
  }
  upvar #0 main_windows($window,cmdline_entry) cmdline_entry
  if {[info exists cmdline_entry]} {
    assert {[winfo exists $cmdline_entry]}
    update_prompt $window
  } else {
    global options
    assert {$options(cmdline_in_output)}
    global main_windows
    if {![info exists main_windows($window,cmdline_term)]} {
      add_cmdline_term $window
    }
  }
}

proc exec_commands {window termname commands} {
  upvar #0 $termname term
  set cmd_line $term(cmd)
  set term_widget [term_widget $termname]

  if {[llength $commands]==0} {
    run_eof $window $termname
    give_cmdline $window
    focus_cmdline $window
    $term_widget see bottom
    return
  } 

  # if there is no focused widget, this is the initial command for the
  # window, so we should focus it.
  if {[cmdline_focused $window] || [focused_widget $window]==""} {
    if {[window_is_compact $window]} {
      # A widget that isn't visible won't focus, so we focus the toplevel
      # widget to take the focus off the command line.
      focus_widget $window $window
    } else {
      focus_widget $window $term_widget
    }
    term_focus $termname
  }
  term_open_pty $termname [list run_eof $window $termname]
  set term(command_queue) $commands
  exec_next_command $window $termname
}

proc exec_term {window termname} {
  if {[term_is_active $termname]} { 
    puts "term is active"
    return 0 
  }

  upvar #0 $termname term
  upvar #0 main_windows($window,cmdline_term) cmdline_term
  upvar #0 main_windows($window,rows) rows
  upvar #0 main_windows($window,cols) cols
  set term(cmd) [term_entry_get $window $termname]
  set parsed_commands [parse_commands $window term(cmd)]
  if {$parsed_commands=="'"} {
    return 0
  }
  if {[focused_entry_term $window]==$termname} {
    replace_header_tag $termname focused_entry disabled_entry
  } else {
    replace_header_tag $termname entry disabled_entry
  }
  if {[info exists cmdline_term] && $cmdline_term==$termname} {
    unset cmdline_term 
  }
  # If this is the last term, make it occupy the rest of the screen so that
  # any blank lines after the last time will be overwritten.
  set term_widget [term_widget $termname]
  if {[term_last $term_widget]==$termname} {
    $term_widget mark set ${termname}bottom bottom
  }
  term_clear $termname
  add_to_history $window $term(cmd)
  focus_widget $window $term_widget
  term_focus $termname
  exec_commands $window $termname $parsed_commands
  return 1
}

proc entry_backspace {widget entry_range} {
  if {[entry_delete_selected $widget $entry_range]} {
  } else {
    if {[$widget compare insert > [lindex $entry_range 0]]} {
      $widget delete insert-1c insert 
    }
  }
}

proc outputarea_key_press {window key} {
  if {$key=="Shift-Prior"} {
    page_up $window
  } elseif {$key=="Shift-Next"} {
    page_down $window
  } else {
    if {![cmdline_in_output $window] && [cmdline_focused $window]} {
      cmdline_key $window $key
    } else {
      set widget $window.outputarea.text
      set termname [focused_entry_term $window]
      if {$termname!=""} {
	fieldref_key \
	  $window $termname [term_entry_fieldref $window $termname] $key
      } else {
	send_focused_term_key $widget $key
      }
    }
  }
}

proc process_input {window} {
  global main_windows
  if {[info exists main_windows($window,input_suspended)]} {
    return
  }
  upvar #0 main_windows($window,input_buffer) input_buffer

  set len [string length $input_buffer]
  if {$len>0} {
    set focused_term [focused_term $window.outputarea.text]
    if {$focused_term!=""} {
      set string $input_buffer
      set input_buffer ""
      send_term $focused_term $string
    } else {
      set c [string index $input_buffer 0]
      set input_buffer [string range $input_buffer 1 end]
      outputarea_key_press $window $c
      if {$len>1} {
	after idle process_input $window
      }
    }
  }
}

proc add_input {window text} {
  upvar #0 main_windows($window,input_buffer) input_buffer
  append input_buffer $text
  after idle process_input $window
}

proc suspend_input {window} {
  global main_windows
  if {[info exists main_windows($window,input_suspended)]} {
    incr main_windows($window,input_suspended)
  } else {
    set main_windows($window,input_suspended) 1
    after cancel "process_input $window"
  }
}

proc resume_input {window} {
  global main_windows
  incr main_windows($window,input_suspended) -1
  if {$main_windows($window,input_suspended)==0} {
    unset main_windows($window,input_suspended)
    after idle process_input $window
  }
}

proc window_of_term_widget {term_widget} {
  global main_window_list main_windows
  foreach window $main_window_list {
    if {$term_widget=="$window.outputarea.text"} { 
      return $window
    }
  }
  return {}
}

proc no_focused_term_paste {term_widget text} {
  set window [window_of_term_widget $term_widget]
  focus_cmdline $window
  add_input $window $text
}

proc text_range {w start end} {
  set selection ""

  # Handle treating wrapped lines as continuous
  while {1} {
    if {[$w compare $end <= "$start lineend"]} {
      append selection [$w get $start $end]
      break
    } else {
      set rest [$w get $start "$start lineend"]
      set tags [$w tag names "$start lineend"]
      if {[lsearch -exact $tags "wrap"]!=-1} {
        append selection $rest
      } else {
        append selection "$rest\n"
      }
      set start [$w index "$start lineend +1c"]
    }
  }
  return $selection
}

proc scan_line {w start pattern} {
  set origstart $start
  set line [$w get $start "$start lineend"]
  set linelen [string length $line]
  while {1} {
    set tags [$w tag names "$start lineend"]
    if {[lsearch -exact $tags "wrap"]==-1} {
      break
    }
    set start [$w index "$start +1 lines linestart"]
    append line [$w get $start "$start lineend"]
  }
  if {[regexp -indices -- $pattern $line _match match1 match2]} {
    set startoff [lindex $match2 0]
    if {$startoff<$linelen} {
      set endoff [lindex $match2 1]
      set text [$w get "$origstart +$startoff c" "$origstart +$endoff c +1 c"]
      if {[string first "\n" $text]!=-1} {
	incr endoff
      }
      return [list $startoff $endoff]
    }
  }
  return ""
}

proc selection_handle {offset maxbytes} {
  set selection ""
  if {[catch {
    global main_window_list
    foreach window $main_window_list {
      set w $window.outputarea.text
      set selection_range [$w tag nextrange sel 1.0]
      if {[llength $selection_range]!=0} {
	set selection_start [lindex $selection_range 0]
	set selection_end [lindex $selection_range 1]
	set selection [text_range $w $selection_start $selection_end]

	set termname [term_at $w $selection_start]
	upvar #0 $termname term
	set dir ""
	catch {set dir $term(dir)}
	global main_windows
	if {$dir==$main_windows($window,dir)} {
	  set dir ""
	}

        # If this is an ls command of a different directory, look in that
        # directory for the filenames.
	set cmd ""
	catch {set cmd $term(cmd)}
	set args [cmd_arg_values $termname $cmd]
	if {[llength $args]==2} {
	  set arg0 [lindex $args 0]
	  set arg1 [lindex $args 1]
	  if {$arg0=="ls" && [string index $arg1 0]!="-"} {
	    set dir [file join $dir $arg1]
	  }
	}

	set files {}
	catch {
	  set files [file_list $dir $selection]
	}
	if {[llength $files]>0} {
	  set selection [quote_files $files]
	}
      }
    }
    if {$selection==""} {
      puts "No selection"
    }
  } message]} {
    puts "Error getting selection: $message"
  }
  return $selection
}

proc movement_index {cmdline key} {
}

proc vi_char_search {cmdline cmd key} {
  switch -- $cmd {
    f -
    t {
      set index \
	[$cmdline \
	  search -forwards -- $key "insert +1c" "insert lineend"]
    }
    F -
    T {
      set index \
	[$cmdline \
	  search -backwards -- $key "insert" "insert linestart"]
    }
    default {error "Invalid find char '$cmd'"}
  }
  if {$index==""} { 
    puts "Char not found"
    bell 
  } else {
    switch -- $cmd {
      t {set index "$index -1c"}
      T {set index "$index +1c"}
      f - 
      F { set index $index }
    }
  }
  return $index
}

#proc cmd_arg_list {cmd} {
#  set index 0
#  set args {}
#  while {1} {
#    set range [cmd_arg_nextrange $cmd $index]
#    if {[llength $range]==0} {
#      break
#    }
#    set start [lindex $range 0]
#    set end [lindex $range 1]
#    set arg [string range $cmd $start $end]
#    lappend args $arg
#    set index $end
#    incr index
#  }
#  return $args
#}

proc cmd_args {cmd} {
  set args {}
  set cmd_index 0
  while {1} {
    set cmd_index [string_end $cmd " *" $cmd_index]
    incr cmd_index
    if {$cmd_index==[string length $cmd]} {
      break
    }
    assert {$cmd_index>=0}
    lappend args [arg_value [get_arg $cmd cmd_index]]
  }
  return $args
}

proc cmd_arg_values {termname cmd} {
  set cmd [varsub_command $termname $cmd]
  return [cmd_args $cmd]
}

proc command_name {cmd} {
  set cmd [varsub_command "" $cmd]
  set cmd_index 0
  return [arg_value [get_arg $cmd cmd_index]]
}

proc window_cmdline_mode {window} {
  global main_windows
  array set state $main_windows($window,cmdline_state)
  return $state(mode)
}

proc cmdline_ctrl_key_press {window cmdline keysym key} {
  cmdline_key_press $window $cmdline "Ctrl-$keysym" [control_char $key]
}

proc page_up {window} {
  $window.outputarea.text yview scroll -1 page 
}

proc page_down {window} {
  $window.outputarea.text yview scroll 1 page 
}

proc cmdline_key_press {window cmdline keysym key} {
  if {$key=="" || $key=="\r" || $key=="\n" || $key=="\b" || $key=="\t" || 
      $keysym=="Ctrl-Delete"} {
    switch -- $keysym {
      Tab -
      Ctrl-Return -
      Ctrl-BackSpace -
      Ctrl-Up -
      Ctrl-Delete -
      Up -
      Down -
      Right -
      Left -
      Home -
      End -
      F2 -
      Shift-Up -
      Shift-Down -
      Shift-Right -
      Shift-Left -
      Shift-Home -
      Shift-End {
	cmdline_key $window $keysym
      }
      Shift-Prior -
      Prior { page_up $window}
      Shift-Next -
      Next { page_down $window}
      Shift-Return -
      Return { add_input $window "\r" }
      BackSpace { add_input $window "\b" }
      Shift-ISO_Left_Tab { cmdline_key $window "Shift-Tab" }
      default { 
	#puts "keysym='$keysym'" 
      }
    }
  } else {
    add_input $window $key
  }
}

proc cmdline_button_1 {window x y} {
  focus_cmdline $window
  set cmdline [cmdline_widget $window]
  set index [tkTextClosestGap $cmdline $x $y]
  $cmdline tag remove sel 1.0 end
  if {$index!=""} {
    $cmdline mark set anchor $index
    $cmdline mark set insert [cmdline_index $cmdline $index ""]

    global tkPriv
    set tkPriv(selectMode) char
    set tkPriv(mouseMoved) 0
    set tkPriv(pressX) $x
  }
}

proc cmdline_double_button_1 {window x y} {
  set w [cmdline_widget $window]

  $w mark set anchor [tkTextClosestGap $w $x $y]
  global tkPriv
  set tkPriv(selectMode) word
  term_text_select_to $w $x $y
}

proc init_cmdline {window w} {
  upvar #0 main_windows($window,cmdline_entry) cmdline_entry
  set cmdline_entry $w
  window_init_vi_state $window

  update_cmdline_colors $cmdline_entry
  bind $cmdline_entry <Shift-KeyPress> \
    "cmdline_key_press {$window} %W Shift-%K %A;break"
  bind $cmdline_entry <Control-KeyPress> \
    "cmdline_ctrl_key_press {$window} %W %K %A;break"
  bind $cmdline_entry <KeyPress> "cmdline_key_press {$window} %W %K %A;break"
  bind $cmdline_entry <ButtonPress-1> "cmdline_button_1 $window %x %y;break"
  bind $cmdline_entry <Double-Button-1> \
    "cmdline_double_button_1 $window %x %y;break"
  bind $cmdline_entry <ButtonRelease-2> \
    "focus_cmdline $window
     no_focused_term_paste $window.outputarea.text \[selection get]
     break"
  bind $cmdline_entry <ButtonPress-3> \
    [list cmdline_button_3 $window %W %x %y %X %Y]

  if {[cmdline_focused $window]} {
    $cmdline_entry tag add focused__entry "end -1c" end
  } else {
    $cmdline_entry tag add entry "end -1c" end
  }
}

proc cmdline_set {cmdline str} {
  set range [cmdline_entry_range $cmdline]
  set start [lindex $range 0]
  set end [lindex $range 1]
  set tags [$cmdline tag names $start]
  $cmdline delete $start $end
  $cmdline insert $start $str $tags
}

proc entry_button_3 {window lx ly gx gy} {
  set term_widget $window.outputarea.text
  set hit_index [tkTextClosestGap $term_widget $lx $ly]
  set termname [term_under_index $term_widget $hit_index]
  if {$termname!=""} {
    upvar #0 $termname term
    set dir ""
    set cmd ""
    catch {set dir $term(dir)}
    catch {set cmd $term(cmd)}
    popup_header_menu $window $termname $gx $gy
    #popup_output_menu $window $dir $cmd $gx $gy $termname
  }
}

proc outputarea_button_3 {window outputarea lx ly gx gy} {
  set index [tkTextClosestGap $outputarea $lx $ly]
  set termname [term_at $outputarea $index]
  if {$termname==""} {
    set termname [term_entry_at $window $lx $ly]
    popup_header_menu $window $termname $gx $gy
  } else {
    # Used to have this focus_cmdline here, but what if there is no
    # command line currently.
    #focus_cmdline $window

    set dir ""
    set cmd ""
    upvar #0 $termname term
    catch {set dir $term(dir)}
    catch {set cmd $term(cmd)}
    popup_output_menu $window $dir $cmd $gx $gy ""
  }
}

proc completion_window_button_3 {window lx ly gx gy} {
  upvar #0 main_windows($window,dir) dir
  popup_output_menu $window $dir "" $gx $gy ""
}

proc cmdline_button_3 {window widget lx ly gx gy} {
  upvar #0 main_windows($window,dir) dir
  global options
  if {[winfo exists $window.main_popup]} {
    tk_popup $window.main_popup $gx $gy
  } else {
    popup_output_menu $window $dir "" $gx $gy ""
  }
}

# If the key is a control character, return it unmodified.  Otherwise,
# convert it to a control character.
proc control_char {ascii} {
  if {[scan [string toupper $ascii] "%c" code]==1} {
    if {$code>=64} {
      incr code -64
    }
    return [format "%c" $code]
  } else {
    return $ascii
  }
}

proc outputarea_key_pressed {window keysym ascii} {
  if {$ascii != "" && 
      $keysym!="Ctrl-Delete" && 
      $keysym!="Ctrl-BackSpace" &&
      $keysym!="Ctrl-Return"} { 
    add_input $window $ascii
  } else { 
    if {[focused_widget $window]=="$window.outputarea.text"} {
      outputarea_key_press $window $keysym
    } else {
      cmdline_key_press $window [cmdline_widget $window] $keysym $ascii
    }
  }
}

# This is here to handle systems where the ascii value does not have the
# control key applied.  Under Linux, the ascii value is the control-modified 
# value.  Under Irix, the ascii value is just the unmodified key.
proc outputarea_ctrl_key_pressed {window keysym ascii} {
  outputarea_key_pressed $window "Ctrl-$keysym" [control_char $ascii]
}

proc update_cmdline_colors {cmdline} {
  global theme config

  $cmdline configure \
    -background $theme(inactive_command_background) \
    -foreground $theme(inactive_command_foreground) \
    -insertbackground $theme(focused_command_cursor_color) \
    -font $theme(cmdline_font) \
    -highlightthickness 0

  $cmdline tag configure prompt \
    -foreground $theme(prompt_foreground) \
    -background $theme(prompt_background)
  $cmdline tag configure entry \
    -foreground $theme(inactive_command_foreground) \
    -background $theme(inactive_command_background)
  $cmdline tag configure focused_entry \
    -foreground $theme(focused_command_foreground) \
    -background $theme(focused_command_background)
  $cmdline configure -insertbackground $theme(focused_command_cursor_color)
  $cmdline tag configure sel \
    -background $theme(selection_background) \
    -foreground $theme(selection_foreground) \
    -relief flat
  $cmdline tag raise sel
}

proc update_window_colors {window} {
  global theme
  set text $window.outputarea.text
  $window configure -background $theme(output_background)
  $window.outputarea configure -background $theme(output_background)
  $text configure -background $theme(output_background)
  $text configure -foreground $theme(output_foreground)
  $text configure -insertbackground $theme(focused_command_cursor_color)
  $text configure -highlightbackground $theme(output_background)
  $text tag configure sel -background $theme(selection_background)
  $text tag configure sel -foreground $theme(selection_foreground)
  $text tag configure prompt -background $theme(prompt_background)
  $text tag configure prompt -foreground $theme(prompt_foreground)
  $text tag configure disabled_entry \
    -foreground $theme(active_command_foreground) \
    -background $theme(active_command_background)
  $text tag configure entry \
    -foreground $theme(inactive_command_foreground) \
    -background $theme(inactive_command_background)
  $text tag configure focused_entry \
    -foreground $theme(focused_command_foreground) \
    -background $theme(focused_command_background)
  global config
  if {![cmdline_in_output $window]} {
    update_cmdline_colors [cmdline_widget $window]
  }
}

proc option_changed {option} {
  global main_window_list
  global main_windows
  global options

  switch -- $option {
    cmdline_in_output {
      foreach window $main_window_list {
	if {!$main_windows($window,is_compact)} {
	  layout_outputarea $window
	  window_layout_noncompact $window
	  update_geometry $window 1
	  focus_cmdline $window
	}
      }
    }
    cmdline_stays_at_top {
      foreach window $main_window_list {
	if {!$main_windows($window,is_compact)} {
	  if {([cmdline_placement $window]=="bottom") == $options($option)} {
	    window_layout_noncompact $window
	    update_geometry $window 1
	    focus_cmdline $window
	  }
	}
      }
    }
    no_menus_with_top_cmdline {
      foreach window $main_window_list {
	global options
	if {[cmdline_placement $window]=="top" && 
	    $options(no_menus_with_top_cmdline)} {
	  set menus_in_popup 1
	} else {
	  set menus_in_popup 0
	}
	if {$menus_in_popup!=[winfo exists $window.main_popup]} {
	  destroy_main_menu $window
	  create_main_menu $window $menus_in_popup
	}
      }
    }
    dir_menu_command {
      refill_dir_menus
    }
  }
}

proc theme_updated {} {
  global main_window_list
  foreach window $main_window_list {
    update_window_colors $window
  }
}

proc window_resized {window} {
  close_completion_window $window
  set w $window.outputarea.text
  set width [$w cget -width]
  set height [$w cget -height]
  upvar #0 main_windows($window,rows) window_rows
  upvar #0 main_windows($window,cols) window_cols
  set window_rows $height
  set window_cols $width
  set term_list [term_list $w]
  set last_term [lindex $term_list end]
  if {$last_term!=""} {
    if {[$w compare "bottom -$height lines" > ${last_term}bottom]} {
      $w delete ${last_term}bottom "bottom -$height lines"
    }
  }
  foreach termname [term_list $w] {
    term_set_screen_size $termname $height $width
  }

  set focused_term [focused_term $w]
  if {$focused_term==""} {
    $w see "bottom -1 lines"
  } else {
    upvar #0 $focused_term ft
    set row_offset [term_row_offset $focused_term]
    $w yview "${focused_term}top +$row_offset lines"
  }
}

proc outputarea_text_configure_notify {window} {
  # If the number of rows of the text area has been changed, and we are in the
  # middle of updating the geometry to reflect this, we don't want to
  # use the current height and end up resetting the number of rows back to
  # what it was.
  global updating_geometry
  if {[info exists updating_geometry]} {
    return
  }

  set w $window.outputarea.text
  set pixel_width [winfo width $w]
  set pixel_height [winfo height $w]
  set borderwidth [$w cget -borderwidth]
  set padx [$w cget -padx]
  set pady [$w cget -pady]
  set highlightthickness [$w cget -highlightthickness]
  incr pixel_width [expr {-2*($borderwidth+$padx+$highlightthickness)}]
  incr pixel_height [expr {-2*($borderwidth+$pady+$highlightthickness)}]
  set font [$w cget -font]
  set cwidth [font measure $font -displayof $w "0"]
  set width [expr {$pixel_width/$cwidth}]
  set cheight [font metrics $font -displayof $w -linespace]
  set height [expr {$pixel_height/$cheight}]
  set width_remainder [expr {$pixel_width-$width*$cwidth}]
  set height_remainder [expr {$pixel_height-$height*$cheight}]
  set current_width [$w cget -width]
  set current_height [$w cget -height]
  if {$width!=$current_width || $height!=$current_height} {
    $w configure -width $width -height $height
    window_resized $window
  }
}

proc bind_outputarea_buttons {window widget} {
  bind $widget <ButtonPress-1> {
    outputarea_button_1 %W %x %y
    break
  }
  bind $widget <Double-Button-1> "
    outputarea_double_button_1 $window %x %y;break"
  bind $widget <ButtonRelease-2> {
    outputarea_button_2 %W %x %y
    break
  }
}


proc compare_length {str1 str2} {
  set len1 [string length $str1]
  set len2 [string length $str2]
  if {$len1<$len2} { return -1 }
  if {$len1>$len2} { return 1 }
  return 0
}

proc regexp_escape {string} {
  regsub -all -- {\.} $string {\.} newstring
  regsub -all -- {\[} $newstring {\[} newstring
  regsub -all -- {\*} $newstring {\*} newstring
  regsub -all -- {\|} $newstring {\|} newstring
  regsub -all -- {\(} $newstring {\(} newstring
  regsub -all -- {\+} $newstring {\+} newstring
  return $newstring
}

proc scan_term_for_filenames {window termname} {
  upvar #0 $termname term
  set term_widget [term_widget $termname]
  $term_widget tag configure filename -foreground yellow
  set filenames \
    [concat [files_matching $term(dir) *] [files_matching $term(dir) .*]]
  set filenames [lsort -decreasing -command compare_length $filenames ]
  if {[llength $filenames]>0} {
    set filepattern [regexp_escape [lindex $filenames 0]]
    set filenames [lrange $filenames 1 end]
    foreach filename $filenames {
      set filename [regexp_escape $filename]
      append filepattern "|$filename"
    }
    set filepattern "( |\n|^)($filepattern)"
    set offset 0
    set start ${termname}top
    while {[$term_widget compare $start < ${termname}bottom]} {
      set range [scan_line $term_widget $start $filepattern]
      if {[llength $range]==0} {
	set start [$term_widget index "$start +1 lines linestart"]
      } else {
	set startoff [lindex $range 0]
	set endoff [lindex $range 1]
	incr endoff
	$term_widget tag add filename \
	  "$start +$startoff c" "$start +$endoff c"
	set start [$term_widget index "$start +$endoff c"]
      }
    }
  }
}

proc detach_term {window termname} {
  upvar #0 $termname old_term
  set old_term_widget [term_widget $termname]
  set term_is_active [term_is_active $termname]
  set term_is_focused [expr {[focused_term $old_term_widget]==$termname}]
  set new_width [term_screen_width $termname]
  if {$term_is_active} {
    set new_height [term_screen_height $termname]
  } else {
    set line_count [term_line_count $termname]
    set new_height [expr {$line_count+3}]
    if {$new_height > [term_screen_height $termname]} {
      set new_height [term_screen_height $termname]
    }
  }
  set new_window [create_window $old_term(dir) $new_width $new_height]
  set end "${termname}bottom"
  set output [$old_term_widget get ${termname}top "$end -1c"]
  set new_termname \
    [create_decor_term $new_window $old_term(dir) $old_term(cmd) disabled_entry]
  upvar #0 $new_termname new_term
  set new_term_widget [term_widget $new_termname]
  set new_term(dir) $old_term(dir)

  #Temporarily set the bottom mark to have right gravity so it will move
  #down as the text is inserted.
  $new_term_widget mark gravity ${new_termname}bottom right
  $new_term_widget insert "${new_termname}bottom" $output
  $new_term_widget mark gravity ${new_termname}bottom left
  if {[info exists old_term(command_queue)]} {
    set new_term(command_queue) $old_term(command_queue)
  }
  if {[info exists old_term(frame_stack)]} {
    set new_term(frame_stack) $old_term(frame_stack)
  }
  if {[info exists old_term(local_variables)]} {
    set new_term(local_variables) $old_term(local_variables)
  }
  if {[info exists old_term(finish_cmd)]} {
    set new_term(finish_cmd) $old_term(finish_cmd)
  }
  assume_term $new_termname $termname \
    "run_eof {$new_window} $new_termname" \
    "run_exit {$new_window} $new_termname"
  if {!$term_is_active} {
    run_eof $new_window $new_termname
  } else {
    term_make_active $new_termname
  }
  remove_term $window $termname
  if {$term_is_focused} {
    give_cmdline $window
    focus_cmdline $window
  }
  if {$term_is_active} {
    term_focus $new_termname
    focus_widget $new_window $new_term_widget
  } else {
    give_cmdline $new_window
    focus_cmdline $new_window
  }
}

proc fill_header_menu {window termname menu} {
  set isactive [term_is_active $termname]
  foreach {label command ifactive ifnotactive} {
    "Execute" "exec_term $window $termname" 0 1
    "Detach" "detach_term $window $termname" 1 1
    "Close" "close_term $window $termname" 0 1
    "---" "" 1 1
    "Stop" "stop_term $window $termname" 1 0
    "Continue" "cont_term $window $termname" 1 0
    "Kill" "kill_term $window $termname" 1 0
  } {
    if {$label=="---"} {
      $menu add separator
    } else {
      set state disabled
      if {$ifactive && $isactive || $ifnotactive && !$isactive} {
	set state normal
      }
      if {$label=="Close" || $label=="Detach"} {
        if {[term_entry_is_cmdline $window $termname]} {
	  set state disabled
	}
      }
      $menu add command \
	-label $label \
	-command [eval list $command] \
	-state $state
    }
  }
}

proc popup_header_menu {window termname x y} {
  set menu $window.popup 
  $menu delete 0 end
  fill_header_menu $window $termname $menu
  tk_popup $menu $x $y
}

# This routine is called when idle to handle any changes made to the cursor.
# It is only done when idle so that if the cursor changes then changes back
# it won't flicker.
proc update_cursor {widget} {
  global current_cursors
  $widget configure -cursor $current_cursors($widget)
}

# Set what the cursor should be and install an "after idle" procedure to
# actually do the change.  This prevents flicker if the cursor changes
# several times quickly.
proc set_cursor {widget name} {
  global current_cursors
  set current_cursors($widget) $name
  after idle "update_cursor $widget"
}

proc create_outputarea_text {window width height} {
  puts "create_outputarea_text:"
  set outputarea $window.outputarea
  set w $outputarea.text
  global theme
  create_terminal_widget $w \
    -width $width -height $height -borderwidth 0 -highlightthickness 0
  $w configure -background $theme(output_background)
  $w configure -foreground $theme(output_foreground)
  $w configure -font $theme(output_font)
  $w tag configure sel -relief flat
  $w tag configure prompt
  $w tag configure disabled_entry
  $w tag configure entry

  $w tag bind entry <ButtonPress-1> \
    "entry_button_1 $window %x %y;break"
  $w tag bind focused_entry <ButtonPress-1> \
    "entry_button_1 $window %x %y;break"
  $w tag bind entry <ButtonRelease-2> \
    "entry_button_2 $window %x %y;break"
  $w tag bind focused_entry <ButtonRelease-2> \
    "entry_button_2 $window %x %y;break"
  $w tag bind focused_entry <Double-Button-1> \
    "entry_double_button_1 %W %x %y;break"
  $w tag bind entry <Triple-Button-1> "
    entry_triple_button_1 $window %x %y;break"

  term_init_widget $w [list term_focused_notify $window]
  bind_outputarea_buttons $window $w
  bind $w <ButtonPress-3> \
    "outputarea_button_3 {$window} %W %x %y %X %Y;break"
  bind $w <Shift-KeyPress> \
    "outputarea_key_pressed {$window} Shift-%K %A;break"
  bind $w <Control-KeyPress> \
    "outputarea_ctrl_key_pressed {$window} %K %A;break"
  bind $w <KeyPress> \
    "outputarea_key_pressed {$window} %K %A;break"
  bind $w <Configure> \
    "after idle outputarea_text_configure_notify $window"
  selection handle $w selection_handle

  # this doesn't have much of an effect right now, but making our own bottom
  # tag gives flexibility to add something that always stays below the
  # output.
  $w mark set bottom end
  $w mark gravity bottom right
}

proc resize_window {window} {
  update_geometry $window 0
  window_resized $window
}

proc hexstring {s} {
  set result ""
  for {set i 0} {$i<[string length $s]} {incr i} {
    set c [string index $s $i]
    scan $c "%c" code
    set result "$result[format "%02x" $code]"
  }
  return $result
}

proc focus_last_term_entry {window} {
  set last_term [term_last_inactive $window.outputarea.text]
  if {$last_term!=""} {
    focus_term_entry $window $last_term
  }
}

proc close_term {window termname} {
  upvar #0 $termname term

  set term_widget [term_widget $termname]
  if {[term_is_active $termname]} { return }

  remove_term $window $termname
}

proc stop_term {window termname} {
  send_term $termname "\x1a"
}

proc cont_term {window termname} {
  term_kill $termname CONT
}

proc kill_term {window termname} {
  term_kill $termname QUIT
}

proc term_under_index {term_widget index} {
  while 1 {
    set index [$term_widget mark next $index]
    if {$index==""} { break }
    if {[regexp "^(term\[0-9]*)top$" $index match termname]} {
      return $termname
    }
  }
  return ""
}

proc term_under {term_widget x y} {
  set index [tkTextClosestGap $term_widget $x $y]
  return [term_under_index $term_widget $index]
}

proc fill_cmd_menu {window menu} {
  global cmd_menu_entries
  fill_menu $window $menu $cmd_menu_entries ""
}

set main_menu_specs {
  {cmd "Cmd"}
  {dir "Dir"}
  {edit "Edit"}
  {help "Help"}
}

proc destroy_main_menu {window} {
  global main_menu_specs

  global options
  if {[winfo exists $window.main_popup]} {
    destroy $window.main_popup
  } else {
    foreach menuspec $main_menu_specs {
      foreach {name label} $menuspec { }
      destroy $window.topline.$name
    }
  }
  global main_windows
  unset main_windows($window,cmd_menu)
}

proc create_main_menu {window menus_in_popup} {
  global options

  if {$menus_in_popup} {
    menu $window.main_popup
  }

  global main_menu_specs

  foreach menuspec $main_menu_specs {
    foreach {name label} $menuspec { }

    if {!$menus_in_popup} {
      let w $window.topline.$name {
	menubutton $w -text $label -pady 0
	set menu [menu $w.menu]
	$w configure -menu $menu
      }
    } else {
      let w $window.main_popup {
        set menu [menu $w.$name]
        $w add cascade -label $label -menu $menu
      }
    }

    switch -- $name {
      cmd {
	fill_cmd_menu $window $menu
	global main_windows
	set main_windows($window,cmd_menu) $menu
      }
      dir {
	fill_dir_menu $window $menu
	global main_windows
	set main_windows($window,dir_menu) $menu
      }
      edit {
	$menu add command \
	  -label "Colors..." -command {edit_theme}
	$menu add command  \
	  -label "Font..." -command "open_font_editor $window"
	$menu add command \
	  -label "Menus..." -command "edit_menus"
	$menu add command \
	  -label "Options..." -command [list edit_options $window]
	$menu add command \
	  -label "Save All Settings" -command [list save_all_settings $window]
      }
      help {
	$menu add command -label "Overview..."  \
	  -command [list open_help $window overview]
	$menu add command -label "Changes..."  \
	  -command [list open_help $window changes]
	$menu add separator
	$menu add command -label "About..."  \
	  -command [list open_about $window]
	set homepage "http://personal.atl.bellsouth.net/atl/v/c/vcato/gsh"
	$menu add command -label "Home Page" \
	  -command [list enter_command $window "browse $homepage"]
      }
      default { assert {0} }
    }

  }
  if {!$menus_in_popup} {
    let w $window.topline {
      pack $w.cmd -side left
      pack $w.dir -side left
      pack $w.edit -side left
      pack $w.help -side right
      if {[cmdline_placement $window]=="top"} {
	pack $w.cmdline -after $w.help
	pack $w.history_arrow -after $w.cmdline
      }
    }
  }
}

proc window_configure_notify {window} {
  global updating_geometry
  if {![info exists updating_geometry]} {
    global main_windows
    set main_windows($window,position_changed) 1
  }
}

proc set_icon {window} {
  if {$window=="."} {
    set icon_window ".icon"
  } else {
    set icon_window "$window.icon"
  }
  global icon_image
  if {![info exists icon_image]} {
    global gsh_lib
    set icon_image [image create photo -file $gsh_lib/icon.gif]
  }
  toplevel $icon_window
  label $icon_window.label -borderwidth 0 -image $icon_image
  pack $icon_window.label
  wm iconwindow $window $icon_window
}

proc create_window {dir width height} {
  chdir $dir

  global next_window_id
  if {![info exists next_window_id]} {
    set next_window_id 1
  }
  set window .window$next_window_id
  incr next_window_id
  toplevel $window
  wm group $window .
  wm withdraw $window

  wm protocol $window WM_DELETE_WINDOW "window_close $window"

  global main_windows
  set main_windows($window,widget) $window
  set main_windows($window,input_buffer) ""
  set main_windows($window,dir) $dir
  global history
  set main_windows($window,history_position) [llength $history]

  let w [frame $window.topline] {
    $w configure -borderwidth 1 -relief raised

    let w $w.compact_button {
      button $w -text "C" -padx 0 -pady 0 -command "switch_compact $window"
      pack $w -side right -anchor n
    }

    grid $window.topline -row 0 -column 0 -sticky ew
    grid rowconfigure $window 0 -weight 0

    menu $window.popup -tearoff 0
    bind $window.popup <2> \
      "menu_select $window $window.popup \$popup_menu_entries"
  }
  frame $window.outputarea
  set main_windows($window,cols) $width
  set main_windows($window,rows) $height
  let outputarea $window.outputarea {
    frame $outputarea.grab_frame -width 10
    create_outputarea_text $window \
      $main_windows($window,cols) $main_windows($window,rows)
    scrollbar $outputarea.vscroll -orient vertical \
      -command "$outputarea.text yview"
    $outputarea.text configure -yscrollcommand "$outputarea.vscroll set"
    arrow_button up $outputarea.history_button \
      -command [list toggle_history_window $window]
  }
  grid columnconfigure $window 0 -weight 1
  global options
  
  layout_window $window $options(start_compact)

  update_window_colors $window
  resize_window $window
  
  global main_window_list
  lappend main_window_list $window
  set_icon $window
  wm deiconify $window

  # Make sure the window is considered to have moved.  This is to work around
  # a problem where the window shifts one pixel
  global main_windows
  if {[info exists main_windows($window,x)]} {
    unset main_windows($window,x)
    unset main_windows($window,y)
  }

  return $window
}

proc new_window {dir width height} {
  set window [create_window $dir $width $height]

  global main_window_list
  set window_count [llength $main_window_list]

  if {$window_count==1} {
    global env
    if {[info exists env(ENV)]} {
      set first_command "source '$env(ENV)'"
    }
  }

  tkwait visibility $window
  if {[info exists first_command]} {
    do_command [lindex $main_window_list 0] $first_command
  } else {
    give_cmdline $window
    focus_cmdline $window
  }

  return $window
}

proc term_entry_hit_index {window x y} {
  set term_widget $window.outputarea.text
  set hit_index [$term_widget index @$x,$y]
  if {[$term_widget get $hit_index]!="\n"} {
    set hit_index [tkTextClosestGap $term_widget $x $y]
  }
  return $hit_index
}

proc term_entry_at {window x y} {
  set term_widget $window.outputarea.text
  set hit_index [term_entry_hit_index $window $x $y]
  set termname [term_under_index $term_widget $hit_index]
}

proc click_entry {window x y} {
  set term_widget $window.outputarea.text
  set hit_index [term_entry_hit_index $window $x $y]
  set termname [term_under_index $term_widget $hit_index]
  focus_term_entry $window $termname
  $term_widget mark set insert $hit_index
  upvar #0 $termname term
  focus_widget $window $term_widget
}

proc entry_button_1 {window x y} {
  click_entry $window $x $y
}

proc entry_button_2 {window x y} {
  puts "entry_button_2:"
  click_entry $window $x $y

  # The selection will be added to the input by the normal outputarea_button_2
  # proc.
  #add_input $window [selection get]
}

proc entry_triple_button_1 {window x y} {
  set term_widget $window.outputarea.text
  $term_widget tag remove sel sel.first sel.last
  set termname [term_under $term_widget $x $y]
  if {$termname!=""} {
    set entry_range [header_entry_range $window $termname]
    $term_widget tag add sel [lindex $entry_range 0] [lindex $entry_range 1]
  }
}

proc text_index_at {widget x y} {
  set index [$widget index @$x,$y]
  set bbox [$widget bbox $index]
  foreach {bx by bw bh} $bbox {
    if {$x>=$bx && $y>=$by && $x<=($bx+$bw) && $y<=($by+$bh)} {
      return $index
    }
  }
  return ""
}

proc click_window {window x y} {
  set w $window.outputarea.text
  $w mark set anchor [tkTextClosestGap $w $x $y]
  set term [term_at $w anchor]

  global tkPriv
  set tkPriv(selectMode) char
  set tkPriv(mouseMoved) 0
  set tkPriv(pressX) $x

  if {$term!=""} {
    if [term_is_active $term] {
      focus_widget $window $w
      term_focus $term
    } else {
      if {[term_entry_is_cmdline $window $term]} {
	focus_cmdline $window
      }
    }
  }
}

proc outputarea_button_1 {w x y} {
  set focused_term [focused_term $w]
  click_window [window_of_term_widget $w] $x $y
  if {[focused_term $w]==$focused_term} {
    $w tag remove sel 0.0 end
  }
}

proc outputarea_button_2 {w x y} {
  set window [window_of_term_widget $w]
  click_window $window $x $y
  add_input $window [selection get]
}

proc entry_double_button_1 {w x y} {
  $w mark set anchor [tkTextClosestGap $w $x $y]
  set term [term_at $w anchor]
  global tkPriv
  set tkPriv(selectMode) word
  term_text_select_to $w $x $y
}

proc outputarea_double_button_1 {window x y} {
  set w $window.outputarea.text
  $w mark set anchor [tkTextClosestGap $w $x $y]
  if {[lsearch [$w tag names anchor] prompt]!=-1} {
    global options
    if {$options(double_click_prompt)} {
      remove_term $window [term_under_index $w anchor]
      return
    }
  }
  set term [term_at $w anchor]
  if {$term!=""} {
    global tkPriv
    set tkPriv(selectMode) word
    term_text_select_to $w $x $y
  }
}

proc show_output {window output_term output} {
  set lines [split $output "\n"]
  foreach line $lines {
    term_insert $output_term $line
    term_newline $output_term
  }
}

proc exec_next_command {window termname} {
  upvar #0 $termname term
  if {[llength $term(command_queue)]==0} {
    if {[info exists term(finish_cmd)]} {
      set finish_cmd $term(finish_cmd)
      while {1} {
	assert {[llength $finish_cmd]>0}
	set cmd [lindex $finish_cmd 0]
	if {$cmd=="then" || $cmd=="else" || $cmd=="elif"} {
	  set then_commands [lindex $finish_cmd 1]
	  set finish_cmd [lrange $finish_cmd 2 end]
	  set term(finish_cmd) $finish_cmd
	  if {$term(exit_code)==0} {
	    if {$cmd=="then"} {
	      set term(command_queue) $then_commands
	      unset term(finish_cmd)
	      break
	    }
	  } else {
	    if {$cmd=="else"} {
	      set term(command_queue) $then_commands
	      unset term(finish_cmd)
	      break
	    } elseif {$cmd=="elif"} {
	      set term(command_queue) [list $then_commands]
	      break
	    }
	  }
	} elseif {$cmd=="for"} {
	  set varname [lindex $finish_cmd 1]
	  set value_list [lindex $finish_cmd 2]
	  if {[llength $value_list]==0} {
	    unset term(finish_cmd)
	  } else {
	    set_variable $termname $varname [arg_value  [lindex $value_list 0]]
	    set value_list [lrange $value_list 1 end]
	    set commands [lindex $finish_cmd 3]
	    set term(command_queue) $commands
	    set term(finish_cmd) [list for $varname $value_list $commands]
	  }
	  break
	} elseif {$cmd=="whiletest"} {
	  assert {[llength $finish_cmd]==3}
	  if {$term(exit_code)!=0} {
	    unset term(finish_cmd)
	  } else {
	    set test [lindex $finish_cmd 1]
	    set commands [lindex $finish_cmd 2]
	    set term(command_queue) $commands
	    set term(finish_cmd) [list while $test $commands]
	  }
	  break
	} elseif {$cmd=="while"} {
	  assert {[llength $finish_cmd]==3}
	  set test [lindex $finish_cmd 1]
	  set commands [lindex $finish_cmd 2]
	  set term(command_queue) $test
	  set term(finish_cmd) [list whiletest $test $commands]
	  break
	}
	if {[llength $finish_cmd]==0} {
	  unset term(finish_cmd)
	  break
	}
      }
      exec_next_command $window $termname
      return
    }
    if {[info exists term(frame_stack)]} {
      pop_frame $termname
      exec_next_command $window $termname
      return
    }
    set focused_term [focused_term [term_widget $termname]]
    set unread [term_close_pty $termname]

    # Temporarily remove any input until the command line can be focused.
    # If we focused the command line before we had called update idletasks,
    # then any pending eof would not be processed until after we had added
    # the command line, which would cause an annoying screen update.

    if {[winfo exists $window]} {
      suspend_input $window

      # Have to do an update instead of update idletasks here to make
      # sure the eof is processed.  If the command line is placed in the
      # output and we don't process the eof before the command line is added, 
      # it will cause a distracting text jump, because the command line will 
      # be added, and then the unused lines will be removed.
      update

      if {$focused_term==$termname} {
	global main_windows
	set main_windows($window,dir) $term(dir)
	give_cmdline $window
	focus_cmdline $window

	# Our new input buffer will have the unread characters at the
	# beginning, and the current input buffer at the end.  This will
	# make it behave as if we had typed the entire input buffer into
	# the terminal, so we don't get any characters out of sequence.
	global main_windows
	append unread $main_windows($window,input_buffer)
	set main_windows($window,input_buffer) ""
	add_input $window $unread
      }
      resume_input $window
      global main_windows
      if {[info exists main_windows($window,spring_expand)]} {
	unset main_windows($window,spring_expand)
	puts "Springing back to compact mode"
	window_set_compact $window 1
      }
    }
    unset term(command_queue)
    return 
  }
  set command [lindex $term(command_queue) 0]
  set term(command_queue) [lrange $term(command_queue) 1 end]
  if {[catch {exec_command $window $termname $command} message]} {
    puts "Error: $message"
    global errorInfo
    puts "$errorInfo"
    show_output $window $termname $message
    set term(command_queue) {}
    run_exit $window $termname
  }
}

proc run_exit {window termname} {
  if {![winfo exists $window]} {
    return
  }
  upvar #0 $termname term
  set term_widget [term_widget $termname]

  assert {$term_widget=="$window.outputarea.text"}
  if {$term(exit_code)=="INT"} {
    catch { unset term(frame_stack) }
    catch { unset term(finish_cmd) }
    set term(command_queue) ""
  }
  exec_next_command $window $termname
}

proc run_eof {window termname} {
  upvar #0 $termname term
  set term_widget [term_widget $termname]

  term_remove_unused_lines $termname
  set next_term [term_next $termname 1]
  if {$next_term!=""} {
    set last_line_index [$term_widget index "${next_term}header -1 lines"]
    $term_widget delete ${termname}bottom ${next_term}header
    # Make sure that there are enough lines in the buffer so that if the
    # screen was cleared, the clear command will be off the top.
    while {[$term_widget compare $last_line_index >= end]} {
      $term_widget insert end "\n" {}
    }
  }
  replace_header_tag $termname disabled_entry entry
  $term_widget tag raise sel
}

proc local_command {window output_term cmd} {
  run_term $output_term $cmd "run_exit {$window} $output_term"
  return true
}

# Execute a command without adding it to the history
proc do_command {window cmd} {
  set parsed_commands [parse_commands "" cmd]
  if {$parsed_commands=="'"} {
    return 0
  }
  upvar #0 main_windows($window,dir) dir
  set termname [create_decor_term $window $dir $cmd disabled_entry]
  exec_commands $window $termname $parsed_commands
  return 1
}

proc focused_fieldref {window} {
  global config
  set termname [focused_entry_term $window]
  if {$termname!=""} {
    return [term_entry_fieldref $window $termname]
  }
  return [cmdline_fieldref [cmdline_widget $window]]
}

proc help_command {window} {
  set fieldref [focused_fieldref $window]
  set cmd [fieldref_get $fieldref]
  set cmdlen [fieldref_insert_pos $fieldref]
  switch [llength $cmd] {
    0 { do_command $window ls }
    1 {
      if {[string index $cmd [expr $cmdlen - 1]]!=" "} {
	fieldref_set $fieldref "$cmd "
      }
      switch [lindex $cmd 0] {
	mpg123 {
	  do_command $window "ls -l *.mp3"
	}
	cd {
	  do_command $window "lsdir"
	}
	rpm {
	  do_command $window "ls -l *.rpm"
	}
      }
    }
    default {
      if {[string index $cmd [expr {[string length $cmd]-1}]]!="/"} {
        do_command $window "ls -d [lindex $cmd end]*" 
      } else {
        do_command $window "ls [lindex $cmd end]" 
      }
    }
  }
}

proc history_position_end {window} {
  global history main_windows
  set main_windows($window,history_position) [llength $history]
  if {[info exists main_windows($window,pending_cmd)]} {
    unset main_windows($window,pending_cmd)
  }
  close_history_window $window
}

proc add_to_history {window cmd} {
  if {$cmd==""} {
    return
  }
  global history
  set index [lsearch -exact $history $cmd]
  if {$index >= 0} {
    set history [lreplace $history $index $index]
  }
  lappend history $cmd
  history_position_end $window
  if {![file exists [rc_dir]]} {
    file mkdir [rc_dir]
  }
  if {[catch {set histfile [open [rc_dir]/history "a"]}]} {
    set histfile [open [rc_dir]/history "w"]
  }
  puts $histfile [list command $cmd]
  close $histfile
}

proc enter_command_term {window termname cmd} {
  window_set_term_cmdline $window $termname $cmd
  return [exec_term $window $termname]
}

# Execute a command as if it was entered from the command line.
# i.e. add it to the history as well as executing it.
proc enter_command {window cmd} {
  global history

  if {![do_command $window $cmd]} {
    return 0
  }
  add_to_history $window $cmd
  return 1
}

proc command_selected {window termname entry} {
  set desc [lindex $entry 1]
  set cmd [lindex $entry 2]
  set verify [lindex $entry 3]
  set spring_expand [lindex $entry 4]

  if {[completion_window_is_open $window]} {
    close_completion_window $window
    window_set_cmdline $window ""
  }
  if {$verify} {
    set text "Are you sure you want to $desc?"
    set title "Verify execute"
    set result [message_dialog $window $title $text {"Yes" "No"} 1]
    if {$result!=0} {
      return
    }
  }
  if {$termname==""} {
    upvar #0 main_windows($window,cmdline_term) cmdline_term
    if {[info exists cmdline_term]} {
      set termname [window_cmdline_term $window]
    }
  }
  if {$termname==""} {
    if {$spring_expand=="1"} {
      if {[window_is_compact $window]} {
	global main_windows
	set main_windows($window,spring_expand) 1
      }
    }
    set result [enter_command $window $cmd]
  } else {
    set result [enter_command_term $window $termname $cmd]
  }
  if {!$result} {
    error "Command is incomplete"
  }
}

proc term_command_entered {window termname} {
  upvar #0 $termname term
  if {![exec_term $window $termname]} {
    term_entry_delete_selected $window $termname
    set widget [term_widget $termname]
    $widget insert insert "\n" [$widget tag names insert]
    $widget see insert
  }
}

proc abort_cmd {window} {
  global config
  if {[cmdline_focused $window]} {
    if {[cmdline_in_output $window]} {
      upvar #0 main_windows($window,cmdline_term) cmdline_term
      term_remove_unused_lines $cmdline_term
      unset cmdline_term
      give_cmdline $window
    } else {
      upvar #0 main_windows($window,dir) dir
      set cmdline_entry [cmdline_widget $window]

      # add command to output, but don't execute it
      set cmd [cmdline_get $cmdline_entry]
      window_set_cmdline $window ""
      set termname [create_decor_term $window $dir $cmd $entry]
      upvar #0 $termname term
      set term(exit_code) 0
      #run_eof $window $termname
    } 
  }
  focus_cmdline $window
}

proc repeat_cmd {window args} {
  global history

  set histpos [find_history $window $args -1]
  if {$histpos>=0} {
    set cmd [lindex $history $histpos]
    do_command $window $cmd
  }
}

proc read_dir_history {} {
  global dir_history
  if {[catch {set file [open "[rc_dir]/dir_history" "r"]}]} {
    set dir_history {}
    return
  }
  set dir_history [split [read -nonewline $file] "\n"]
  close $file
}

proc write_dir_history {} {
  set file [create_file "[rc_dir]/dir_history"]
  global dir_history
  puts -nonewline $file [join $dir_history "\n"]
  close $file
}

proc create_file {filename} {
  create_dir [file dirname $filename]
  return [open $filename "w"]
}

proc refill_dir_menus {} {
  global main_window_list
  foreach window $main_window_list {
    global main_windows
    if {[info exists main_windows($window,dir_menu)]} {
      fill_dir_menu $window $main_windows($window,dir_menu)
    }
  }
}

proc add_directory {dir} {
  read_dir_history

  global dir_history

  set index [lsearch -exact $dir_history $dir]
  if {$index >= 0} {
    set dir_history [lreplace $dir_history $index $index]
  }
  set dir_history [concat [list $dir] $dir_history]
  if {[llength $dir_history] > 20} {
    set dir_history [lreplace $dir_history 20 end]
  }
  write_dir_history

  refill_dir_menus
}

proc handle_dir_changed {window} {
  global main_windows
  set dir $main_windows($window,dir)
  add_directory $dir
  if 0 {
    if {[llength [files_matching $dir *]] >= 20} {
      enter_command $window "ls -lt | head -20"
    } else {
      enter_command $window "ls"
    }
  }
}

proc matchassign {str varnamename valuename} {
  upvar $varnamename varname
  upvar $valuename value
  return [regexp {^([a-zA-Z_][a-zA-Z_0-9]*)=(.*)$} $str match varname value]
}

proc save_aliases {} {
  if {[catch {set file [open [rc_dir]/aliases "w"]} message]} {
    puts "Couldn't save aliases: $message"
    return
  }
  if {[catch {
    global default_aliases
    global aliases
    foreach {name value} [array get default_aliases] {
      if {![info exists aliases($name)]} {
	puts $file [list unalias $name]
      }
    }
    foreach {name value} [array get aliases] {
      puts $file [list alias $name $value]
    }
  } message]} {
    puts "save_aliases: $message"
  }
  close $file
}

proc save_function {name} {
  set funcdir [rc_dir]/functions
  file mkdir $funcdir
  global functions
  if {![info exists functions($name)]} {
    file delete $funcdir/$name
  } else {
    if {[catch {set file [open $funcdir/$name "w"]} message]} {
      puts "Couldn't save function $name: $message"
      return
    }
    if {[catch {
      global functions
      puts $file [list function $name $functions($name)]
    } message]} {
      puts "save_function: $message"
    }
    close $file
  }
}

proc define_function {funcdef} {
  set funcname [lindex $funcdef 1]
  set body [lindex $funcdef 2]
  global functions
  set functions($funcname) $body
  return $funcname
}

proc source_file {window termname filename args} {
  puts "Sourcing $filename"
  set file [open $filename "r"]
  if {[catch {
    set contents [read -nonewline $file]
  } message]} {
    close $file
    error "Error reading $filename: $message"
  }
  close $file
  set result [parse_commands "" contents]
  if {$result=="'"} {
    error "File $filename has an incomplete command"
  }
  upvar #0 $termname term
  push_frame $termname
  set term(command_queue) $result
}

proc push_frame {termname} {
  upvar #0 $termname term

  klnew frame
  klset frame command_queue $term(command_queue)
  unset term(command_queue)
  if {[info exists term(local_variables)]} {
    klset frame local_variables $term(local_variables)
    unset term(local_variables)
  }
  if {[info exists term(finish_cmd)]} {
    klset frame finish_cmd $term(finish_cmd)
    unset term(finish_cmd)
  }
  if {![info exists term(frame_stack)]} {
    set term(frame_stack) {}
  }
  push term(frame_stack) $frame
}

proc pop_frame {termname} {
  upvar #0 $termname term

  set frame [pop term(frame_stack)]
  if {[llength $term(frame_stack)]==0} {
    unset term(frame_stack)
  }
  set term(command_queue) [klget frame command_queue]
  if {[info exists term(local_variables)]} {
    unset term(local_variables)
  }
  klget frame local_variables term(local_variables)
  assert {![info exists term(finish_cmd)]}
  klget frame finish_cmd term(finish_cmd)
}

proc exec_internal_command {window termname cmd cmdname} {
  upvar #0 $termname term
  if {$cmdname==":"} {
    set term(exit_code) 0
    return 1
  }
  if {$cmdname=="set"} {
    # Need to implement this
    return 1
  }
  if {$cmdname=="function"} {
    set funcname [define_function $cmd]
    save_function $funcname
    return 1
  }
  if {$cmdname=="if"} {
    push_frame $termname
    set term(command_queue) [list [lindex $cmd 1]]
    set term(finish_cmd) [lrange $cmd 2 end]
    return 1
  }
  if {$cmdname=="for"} {
    set varname [lindex $cmd 1]
    set valuelist [varsub_command $termname [lindex $cmd 2]]
    set commands [lindex $cmd 3]
    if {[llength $valuelist]>0} {
      push_frame $termname
      set_variable $termname $varname [arg_value [lindex $valuelist 0]]
      set term(command_queue) $commands
      set term(finish_cmd) \
        [list for $varname [lrange $valuelist 1 end] $commands]
    }
    return 1
  }
  if {$cmdname=="while"} {
    assert {[llength $cmd]==3}
    set test [lindex $cmd 1]
    set commands [lindex $cmd 2]
    push_frame $termname
    set term(command_queue) $test
    set term(finish_cmd) [list whiletest $test $commands]
    return 1
  }
  if {$cmdname=="alias" || $cmdname=="export"} {
    set cmd_index 0
    get_arg $cmd cmd_index
    set arg [get_arg $cmd cmd_index]
    if {$arg==""} {
      if {$cmdname=="export"} {
	global env
	foreach {name value} [array get env] {
	  show_output $window $termname "$name='$value'"
	}
      }
      if {$cmdname=="alias"} {
	global aliases
	foreach {name value} [array get aliases] {
	  show_output $window $termname "$name='$value'"
	}
      }
    } else {
      while {$arg!=""} {
	if {[matchassign $arg varname value]} {
	  set value [varsub_command $termname $value]
	  set value [arg_value $value]
	  if {$cmdname=="export"} {
	    global env
	    set env($varname) $value
	  } elseif {$cmdname=="alias"} {
	    global aliases
	    set aliases($varname) $value
	    save_aliases
	  }
	} else {
	  set varname $arg
	  if {$cmdname=="export"} {
	    global env variables
	    if {[info exists variables($varname)]} {
	      set env($varname) $variables($varname)
	      unset variables($varname)
	    } elseif {![info exists env($varname)]} {
	      set env($varname) ""
	    }
	  } elseif {$cmdname=="alias"} {
	    global aliases
	    if {[info exists aliases($arg)]} {
	      show_output $window $termname "$arg='$aliases($arg)'"
	    } else {
	      show_output $window $termname "No such alias"
	    }
	  }
	}
	set arg [get_arg $cmd cmd_index]
      }
    } 
    return 1
  }
  return 0
}

proc window_set_size {window new_width new_height} {
  set current_width [$window.outputarea.text cget -width]
  set current_height [$window.outputarea.text cget -height]
  
  # If the window is compact, we don't really know what our
  # current width is since the output area is not visible.  Always
  # update the geometry in this case.
  if {$current_width!=$new_width || $current_height!=$new_height ||
      [window_is_compact $window]} {
    $window.outputarea.text configure \
      -width $new_width \
      -height $new_height
    resize_window $window
  }
}

proc window_update_title {window} {
  upvar #0 main_windows($window,title) title
  upvar #0 main_windows($window,default_title) default_title
  if {[info exists title]} {
    wm title $window $title
  } else {
    wm title $window $default_title
  }
}

proc window_set_default_title {window title} {
  global main_windows
  set main_windows($window,default_title) $title
  window_update_title $window
}

proc window_set_title {window title} {
  global main_windows
  set main_windows($window,title) $title
  window_update_title $window
}

proc window_untitle {window} {
  upvar #0 main_windows($window,title) title
  if {[info exists title]} {
    unset title
    window_update_title $window
  }
}

proc exec_builtin_command {window termname cmd} {
  upvar #0 $termname term
  set args [cmd_arg_values $termname $cmd]
  set cmdname [lindex $args 0]
  set args [lreplace $args 0 0]
  global functions
  if {[info exists functions($cmdname)]} {
    push_frame $termname
    klnew term(local_variables) 
    set argnum 1
    set allargs ""
    foreach arg $args {
      klset term(local_variables) $argnum $arg
      append allargs " $arg"
      incr argnum
    }
    klset term(local_variables) * $allargs
    set term(command_queue) $functions($cmdname)
  } else {
    switch -- $cmdname {
      source -
      . {
	if {[llength $args]==0} {
	  error "no filename specified"
	}
	set filename [lindex $args 0]
	set args [lreplace $args 0 0]
	set paths [variable_value $termname "PATH"]
	set paths [split $paths ":"]
	foreach path $paths {
	  set full_path [file join $path $filename]
	  if {[file exists $full_path]} {
	    source_file $window $termname $full_path $args
	    break
	  }
	}
      }
      unalias {
	global aliases
	foreach arg $args {
	  catch {unset aliases($arg)}
	}
	save_aliases
      }
      unset {
	global variables
	global env
	if {[lindex $args 0]=="-f"} {
	  set args [lreplace $args 0 0]
	  global functions
	  foreach arg $args {
	    if {![catch {unset functions($arg)}]} {
	      save_function $arg
	    }
	  }
	} else {
	  foreach arg $args {
	    if {[catch {unset variables($arg)}] &&
		[catch {unset env($arg)}]} {
	      # Assume they are trying to unset a function if no variable
	      # is found.
	      if {![catch {unset functions($arg)}]} {
		save_function $arg
	      }
	    }
	  }
	}
      }
      setenv {
	global env
	if {[llength $args]==0} {
	  foreach {name value} [array get env] {
	    show_output $window $termname "$name=$value"
	  }
	} elseif {[llength $args]==1} {
	  set varname [lindex $args 0]
	  set env($varname) ""
	} elseif {[llength $args]==2} {
	  set varname [lindex $args 0]
	  set value [lindex $args 1]
	  set env($varname) $value
	} else {
	  show_output $window $termname "setenv: Too many arguments."
	}
      }
      exit {
	window_close $window
      }
      gshcmd {
	set unknown 0
	if {[llength $args]==0} {
	  set msg ""
	  append msg "Usage: gshcmd new_window [<directory>]\n"
	  append msg "       gshcmd size <width> <height>\n"
	  append msg "       gshcmd title <window title>\n"
	  append msg "       gshcmd untitle"
	  error $msg
	}
	set cmd [lindex $args 0]
	set args [lreplace $args 0 0]
	switch -- $cmd {
	  size {
	    if {[llength $args]!=2} {
	      error "gshcmd size: wrong number of parameters"
	    }
	    window_set_size $window [lindex $args 0] [lindex $args 1]
	  }
	  title {
	    if {[llength $args]!=1} {
	      error "gshcmd title: wrong number of parameters"
	    }
	    window_set_title $window [lindex $args 0]
	  }
	  untitle {
	    if {[llength $args]!=0} {
	      error "gshcmd untitle: wrong number of parameters"
	    }
	    window_untitle $window
	  }
	  new_window {
	    if {[llength $args]==0} {
	      global main_windows
	      set dir $main_windows($window,dir)
	    } else {
	      set dir [lindex $args 0]
	    }
	    global main_windows
	    set width $main_windows($window,cols)
	    set height $main_windows($window,rows)
	    if {[catch {
		   set new_window [new_window $dir $width $height]
		 } message]} {
	      term_insert $termname $message
	    } else {
	      handle_dir_changed $new_window
	    }
	  }
	  default {
	    error "gshcmd: unknown command '$cmd'"
	  }
	}
      }
      clone {
	global clone_cmd
	if {[llength $args]==0} {
	  eval $clone_cmd
	} else {
	  set old_dir [pwd]
	  chdir [lindex $args 0]
	  eval $clone_cmd
	  chdir $old_dir
	}
      }
      lsdir {
	set directories {}
	foreach file [files_matching $dir *] {
	  if [file isdirectory $file] {
	    lappend directories $file
	  }
	}
	show_output $window $termname [join [columnize [lsort $directories]] "\n"]
      }
      cd {
	if {[llength $args]==0} {
	  set dirname "~"
	} else {
	  if {[llength $args]>1} {
	    error "Only one directory can be specified"
	  } {
	    set dirname [lindex $args 0]
	    if {$dirname=="-"} {
	      global main_windows
	      upvar #0 main_windows($window,prev_dir) prev_dir
	      if {[info exists prev_dir]} {
		set dirname $prev_dir
		show_output $window $termname $dirname
	      } else {
		show_output $window $termname "No previous directory."
		unset dirname
	      }
	    } else {
	      set files {}
	      catch {set files [glob $dirname]}
	      if {[llength $files]>1} {
		error "Only one directory can be specified"
	      } elseif {[llength $files]==1 && 
			 [file isdirectory [lindex $files 0]]} {
		# If the name contains spaces, the lindex command must be used
		# to prevent braces from being placed around the name.
		set dirname [lindex $files 0]
	      } else {
		global dir_history
		set destdir $dirname
		foreach hdir $dir_history {
		  if {[file tail $hdir]==$dirname} {
		    show_output $window $termname $hdir
		    set destdir $hdir
		    break
		  }
		}
		set dirname $destdir
	      }
	    }
	  }
	}
	if {[info exists dirname]} {
	  chdir $dirname
	  set dirname [pwd]
	  global main_windows
	  set main_windows($window,prev_dir) $main_windows($window,dir)
	  set term(dir) $dirname
	  global env
	  set env(PWD) $dirname
	  add_directory $dirname
	}
      }
      default {
	return 0
      }
    }
  }
  return 1
}

proc exec_command {window termname cmd} {
  upvar #0 $termname term
  set internal_command 1
  global main_windows
  set dir $term(dir)
  global env
  set env(PWD) $dir
  set term(exit_code) 0

  if {[regexp {^ *@(.*)$} $cmd _match tcl_command]} {
    set result [uplevel #0 $tcl_command]
    show_output $window $termname $result
    run_exit $window $termname
    return 1
  }
  if {[string index $cmd 0]=="("} {
    local_command $window $termname $cmd
    return
  }
  while {1} {
    set range [cmd_arg_nextrange $cmd 0]
    if {[llength $range]==0} {
      run_exit $window $termname
      return
    }
    set start [lindex $range 0]
    set end [lindex $range 1]
    set cmdname [string range $cmd $start $end]
    if {![matchassign $cmdname varname value]} {
      break
    }
    set value [varsub_command $termname $value]
    set value [arg_value $value]
    set_variable $termname $varname $value
    set cmd [string range $cmd [expr {$end+1}] end]
  }
  if {[exec_internal_command $window $termname $cmd $cmdname]} {
    run_exit $window $termname
    return
  }
  if {[catch {chdir $dir}]} {
    set cmdname [command_name $cmd]
    if {$cmdname!="cd" && $cmdname!="exit"} {
      error "Couldn't change directory to $dir.\nPlease use 'cd' to change to an appropriate directory."
    }
  }
  if {[exec_builtin_command $window $termname $cmd]} {
    run_exit $window $termname
    return
  }
  local_command $window $termname [varsub_command $termname $cmd]
}

proc push {listname value} {
  upvar $listname list
  lappend list $value
}

proc pop {listname} {
  upvar $listname list
  set len [llength $list]
  set last [expr {$len-1}]
  set value [lindex $list $last]
  set list [lreplace $list $last $last]
  return $value
}

proc top {listname} {
  upvar $listname list
  set len [llength $list]
  set last [expr {$len-1}]
  set value [lindex $list $last]
  return $value
}

proc fill_history_menu {window menu max_entries} {
  global history
  set count [llength $history]
  set start [expr {$count-$max_entries}]
  if {$start<0} {
    set start 0
  }
  while {$start<$count} {
    set item [lindex $history $start]
    lappend entries [list command $item $item 0]
    incr start
  }
  fill_menu $window $menu $entries ""
}

proc fill_dir_menu {window menu} {
  read_dir_history
  global dir_history
  global options
  set dir_cmd $options(dir_menu_command)
  $menu delete 0 end
  set entries {}
  klnew vars

  foreach dir $dir_history {
    klset vars dir [quote_file $dir]
    lappend entries [list command $dir [expand_command $dir_cmd $vars] 0]
  }
  fill_menu $window $menu $entries ""
}

proc window_get_cmdline {window} {
  return [cmdline_get [cmdline_widget $window]]
}

proc menu_select {window menu entries termname} {
  set index [$menu index active]
  if {[$menu cget -tearoff]} {
    set index_offset -1
  } else {
    set index_offset 0
  }
  if {$index!="none"} {
    set entry [lindex $entries [expr {$index + $index_offset}]]
    if {[lindex $entry 0]=="command"} {
      set cmd [lindex $entry 2]
      if {$termname==""} {
	window_set_cmdline $window $cmd
      } else {
	window_set_term_cmdline $window $termname $cmd
      }
    }
  }
}

proc menu_select_history {window menu} {
  set index [$menu index active]
  if {$index!="none"} {
    window_set_cmdline $window "[$menu entrycget $index -label]"
  }
}

proc update_cmd_menus {} {
  global main_window_list

  foreach window $main_window_list {
    global main_windows
    fill_cmd_menu $window $main_windows($window,cmd_menu)
  }
}

proc add_cmd_entry {path entry} {
  global cmd_menu_entries

  set cmd_menu_entries [set_menu_entries_entry $cmd_menu_entries $path $entry]
}

proc add_cmd {path label command} {
  set verify 0
  set spring_expand 0
  set entry [list command $label $command $verify $spring_expand]
  add_cmd_entry [concat $path [list $label]] $entry
  update_cmd_menus
}

proc edit_theme {} {
  open_theme_dialog
}

proc open_help {window name} {
  catch {destroy .help}
  global gsh_lib
  set filename $gsh_lib/help/$name
  if {![file exists $filename]} {
    error_dialog $window "$filename is missing."
    return
  }
  if {[catch {
    set file [open $filename "r"]
    set text [read $file]
    close $file
  }]} {
    error_dialog $window "Error reading $filename"
    catch {close $file}
    return
  }

  let w .help {
    toplevel $w
    wm title $w "Help $name"
    let w $w.text {
      set text_widget $w
      text $w 
      $w insert end $text
      pack $w -side left -expand true -fill both
    }
    let w $w.scrollbar {
      scrollbar $w -orient vertical
      pack $w -side left -fill y
    }
    $w.scrollbar configure -command [list $w.text yview]
    $w.text configure -yscrollcommand [list $w.scrollbar set]
    $w.text configure -state disabled
  }
  center_window .help $window
}

proc open_about {window} {
  if {[winfo exists .about_image]} {
    return
  }
  global gsh_lib
  if {[lsearch -exact [image names] about] == -1} {
    image create photo about -file $gsh_lib/splash.gif
  }
  toplevel .about_image -background black
  wm overrideredirect .about_image 1
  label .about_image.label -image about -borderwidth 0
  pack .about_image.label
  toplevel .about_info -background black
  wm overrideredirect .about_info 1
  label .about_info.label \
    -text "Version 0.1.2  (C) 1999\nemail: vcato@bellsouth.net" \
    -foreground darkblue -background darkgrey \
    -borderwidth 2 -relief raised
  pack .about_info.label
  foreach label {.about_image.label .about_info.label} {
    bind $label <ButtonRelease-1> { destroy .about_image .about_info }
  }
  center_windows {.about_image .about_info} $window
}

proc change_window_fonts {window output_font cmdline_font} {
  set current_font [$window.outputarea.text cget -font]
  if {$current_font!=$output_font} {
    $window.outputarea.text configure -font $output_font
    if {![cmdline_in_output $window]} {
      [cmdline_widget $window] configure -font $cmdline_font
    }
    resize_window $window
  } 
}

proc change_window_font {window font} {
  change_window_fonts $window $font $font
}

proc open_font_editor {window} {
  edit_font $window [$window.outputarea.text cget -font] \
    [list change_window_font $window]
}

proc read_history {filename} {
  if {[file exists $filename]} {
    puts "Reading history from $filename"
    global history
    process_file $filename line {
      lappend history [lindex $line 1]
    }
    puts "Done reading history"
  }
  global history max_history
  set histlen [llength $history]
  if {$histlen>$max_history} {
    puts "Reducing history file to $max_history lines"
    set histfile [open $filename "w"]
    foreach entry [lrange $history [expr {$histlen-$max_history}] end] {
      puts $histfile [list command $entry]
    }
    close $histfile
  }
}


proc read_theme {filename varname} {
  if {![file exists $filename]} {
    puts "No theme file"
    return
  }
  upvar $varname theme
  process_file $filename line {
    switch -- [lindex $line 0] {
      theme_item {
	set item [lindex $line 1]
	set value [lindex $line 2]
	set theme($item) $value
      }
    }
  }
}

proc read_options {filename} {
  if {![file exists $filename]} {
    puts "No options file"
    return
  }
  process_file $filename line {
    switch -- [lindex $line 0] {
      option {
	set name [lindex $line 1]
	set value [lindex $line 2]
	global options
	set options($name) $value
      }
    }
  }
}

proc read_aliases {filename aliasvar} {
  if {![file exists $filename]} {
    puts "No alias file"
    return
  }

  upvar $aliasvar aliases
  process_file $filename line {
    switch -- [lindex $line 0] {
      "alias" {
	set name [lindex $line 1]
	set value [lindex $line 2]
	set aliases($name) $value
      }
      "unalias" {
	set name [lindex $line 1]
	if {[info exists aliases($name)]} {
	  unset aliases($name)
	}
      }
    }
  }
}

proc read_function {funcname} {
  process_file $funcname line {
    set type [lindex $line 0]
    if {$type!="function"} {
      error "Unknown type '$type' specified in $funcname"
    }
    set name [lindex $line 1]
    if {$name==""} {
      error "No name specified for function in $funcname"
    }
    set body [lindex $line 2]
    global functions
    set functions($name) $body
  }
}

proc read_functions {funcdir} {
  set funcnames [glob -nocomplain -- $funcdir/*]
  foreach funcname $funcnames {
    read_function $funcname
  }
}

proc write_theme {filename} {
  global theme

  set file [open $filename "w"]
  foreach {name value} [array get theme] {
    puts $file [list "theme_item" $name $value]
  }
  close $file
}

proc write_options {filename} {
  global options

  set file [open $filename "w"]
  foreach {name value} [array get options] {
    puts $file [list option $name $value]
  }
  close $file
}


proc save_theme {} {
  write_theme [rc_dir]/theme
}

proc save_options {} {
  write_options [rc_dir]/options
}

proc save_all_settings {window} {
  save_theme
  global options
  set options(initial_width) [window_cols $window]
  set options(initial_height) [window_rows $window]
  save_options
}

proc write_menu {file label entries indent} {
  puts $file "${indent}[list menu -label $label -entries] {"
  set old_indent $indent
  set indent "  $indent"
  foreach entry $entries {
    assert {$entry!={}}
    switch [lindex $entry 0] {
      separator { puts $file "${indent}separator" }
      history_menu { 
	puts $file "${indent}history_menu -label [list [lindex $entry 1]]"
      }
      menu {
	write_menu $file [lindex $entry 1] [lindex $entry 2] $indent
      }
      command {
	set label [lindex $entry 1]
	set value [lindex $entry 2]
	set verify [lindex $entry 3]
	set spring_expand [lindex $entry 4]
        set string "${indent}command [list -label $label -command $value]"
	if {$verify} {
	  append string " -verify"
	}
	if {$spring_expand} {
	  append string " -spring_expand"
	}
        puts $file $string
      }
    }
  }
  puts $file "${old_indent}}"
}

proc process_menu_line {line} {
  set type [getword line]
  switch -- $type {
    "menu" {
      while {[llength $line]>0} {
	switch -- [getword line] {
	  -label { 
	    set label [getword line] 
	  }
	  -entries { 
	    set lines [split [getword line] "\n"]
	    set entries {}
	    process_lines $lines entry_line {
	      set entry [process_menu_line $entry_line]
	      lappend entries $entry
	    }
	  }
	}
      }
      set entry [list menu $label $entries]
    }
    "history_menu" {
      set max_entries 20
      while {[llength $line]>0} {
	switch -- [getword line] {
	  -label { set label [getword line] }
	  -max_entries { set max_entries [getword line] }
	}
      }
      set entry [list history_menu $label $max_entries]
    }
    "separator" {
      set entry {separator}
    }
    "command" {
      set verify 0
      set spring_expand 0
      while {[llength $line]>0} {
	switch -- [getword line] {
	  -label   { set label [getword line] }
	  -command { set command [getword line] }
	  -verify  { set verify 1 }
	  -spring_expand { set spring_expand 1 }
	}
      }
      set entry [list command $label $command $verify $spring_expand]
    }
  }
  if {![info exists entry]} {
    error "Unknown menu type $type"
  } else {
    return $entry
  }
}

proc read_menu {filename} {
  if {[catch {process_file $filename line {
    set entry [process_menu_line $line]
    if {[lindex $entry 0]=="menu"} {
      set label [lindex $entry 1]
      if {$label=="Cmd"} {
	global cmd_menu_entries
	set cmd_menu_entries [lindex $entry 2]
      } else {
	puts "Unknown menu $label"
      }
    }
  }}]} {
    puts "$filename does not exist"
  }
}

proc save_cmd_menu {} {
  global cmd_menu_entries
  set filename [rc_dir]/cmd_menu
  set file [open $filename "w"]
  if {[catch {write_menu $file "Cmd" $cmd_menu_entries ""} message]} {
    puts "Error writing $filename: $message"
    global errorInfo
    puts "  $errorInfo"
  }
  close $file
}

proc read_config {} {
  set config_dir [rc_dir]
  set defaults_dir [defaults_dir]
  read_options "$config_dir/options"
  read_types "$config_dir/types"
  read_popup_commands "$config_dir/popup_commands"
  read_history "$config_dir/history"
  global theme
  read_theme "$config_dir/theme" theme
  global default_aliases
  read_aliases "$defaults_dir/aliases" default_aliases
  global aliases
  array set aliases [array get default_aliases]
  read_aliases "$config_dir/aliases" aliases
  read_functions "$config_dir/functions"
  read_menu "$config_dir/cmd_menu"
}

set cmd_menu_entries {}

# Install a fallback command menu.  This would only be used if the user
# didn't have a custom menu and there was no default.
add_cmd_entry {"History"} [list history_menu "History" 20]
add_cmd {} "Restart" {clone;exit}
add_cmd {} "New Window" {new_window}
add_cmd {"Size Window"} "80x25" {size_window 80 25}
add_cmd {"Size Window"} "80x40" {size_window 80 40}
add_cmd {} "Edit init" "(cd [rc_dir];vi init)"
add_cmd {} "Shutdown" {shutdown}
add_cmd_entry {"----"} [list separator]
add_cmd {} "Exit" {exit}

read_config
wm withdraw .
wm command . [concat $argv0 $argv]
wm group . .

set_icon .

new_window [pwd] $options(initial_width) $options(initial_height)

let window [lindex $main_window_list 0] {
  handle_dir_changed $window
}
puts "Entering main loop"

# Comment out to leave in asserts
#proc assert {x} {}
