#********************************************************/
#* Gsh Copyright (C) 1999 Vaughn Cato                   */
#* See the file "LICENSE", included with this software, */
#* for license information.                             */
#********************************************************/
proc menu_entry_name {entry} {
  switch -- [lindex $entry 0] {
    separator {
      return "separator"
    }
    history_menu -
    menu -
    command {
      return [lindex $entry 1]
    }
    default { assert {0} }
  }
}

proc menu_entry_label {entry} {
  switch -- [lindex $entry 0] {
    separator {
      set label "----------"
    }
    history_menu -
    menu {
      set label "[lindex $entry 1]/"
    }
    command {
      set label [lindex $entry 1]
    }
    default { assert {0} }
  }
  return $label
}

proc menu_editor_close {} {
  destroy .menu_editor
}

proc menu_editor_fill_menu_items_listbox {} {
  global menu_editor

  menu_editor_set_notebook_frame menu_frame

  set entries [menu_editor_menu_entries]
  set listbox $menu_editor(menu_items_listbox)
  set index [$listbox curselection]

  $listbox delete 0 end

  foreach entry $entries {
    set label [menu_entry_label $entry]
    $listbox insert end $label
  }
  $listbox insert end ""
  if {$index==""} {
    set index 0
  }
  $listbox selection set $index
  $listbox activate $index
  menu_editor_menu_item_selected
}

proc focus_entry {entry} {
  focus $entry
  $entry selection range 0 end
  $entry icursor end
}

proc menu_editor_fill_command {} {
  global menu_editor
  menu_editor_set_notebook_frame command_frame
  set entry [menu_editor_entry]
  set menu_editor(command_string) [lindex $entry 2]
  set menu_editor(command_verify) [lindex $entry 3]
  set menu_editor(command_spring_expand) [lindex $entry 4]
  focus_entry $menu_editor(command_entry)
}

proc menu_editor_fill_history_menu {} {
  global menu_editor
  menu_editor_set_notebook_frame history_menu_frame
  set max_entries 20
  set entry [menu_editor_entry]
  if {[llength $entry]>2} {
    set max_entries [lindex $entry 2]
  }
  set menu_editor(history_max_entries) $max_entries
  focus_entry $menu_editor(history_max_entries_entry)
}

proc menu_editor_fill_separator {} {
  global menu_editor
  menu_editor_set_notebook_frame separator_frame
}

proc menu_editor_fill_item_label_menu {} {
  global menu_editor
  set menu $menu_editor(item_label_menu)
  $menu delete 0 end
  set path {}
  foreach label $menu_editor(item_path) {
    lappend path $label
    set entry [menu_entry [menu_editor_top_entry] $path]
    set label [menu_entry_label $entry]
    $menu add command  \
      -label $label \
      -command [list menu_editor_set_item_path $path]
  }
}


proc menu_entry {top_entry path} {
  set name [lindex $path 0]
  assert {[menu_entry_name $top_entry]==$name}
  set remaining_path [lreplace $path 0 0]
  if {[llength $remaining_path]==0} {
    return $top_entry
  }
  switch -- [lindex $top_entry 0] {
    menu {
      set entries [lindex $top_entry 2]
      set next_name [lindex $remaining_path 0]
      foreach entry $entries {
	if {[menu_entry_name $entry]==$next_name} {
	  return [menu_entry $entry $remaining_path]
	}
      }
    }
  }
  assert {0}
}

proc menu_editor_entry {} {
  global menu_editor
  return [menu_entry [menu_editor_top_entry] $menu_editor(item_path)]
}

proc menu_editor_menu_entries {} {
  set entry [menu_editor_entry]
  assert {[lindex $entry 0]=="menu"}
  return [lindex $entry 2]
}

proc set_menu_entries_entry {entries path value} {
  set found 0
  set entry_count [llength $entries]
  assert {[llength $path]>0}
  set name [lindex $path 0]
  for {set index 0} {$index<$entry_count} {incr index} {
    set entry [lindex $entries $index]
    set entry_name [menu_entry_name $entry]
    if {$entry_name==$name} {
      set remaining_path [lreplace $path 0 0]
      set new_entry [set_menu_entry $entry $remaining_path $value]
      set entries [lreplace $entries $index $index $new_entry]
      set found 1
      break
    }
  }
  if {!$found} {
    set remaining_path [lreplace $path 0 0]
    lappend entries [set_menu_entry [list menu $name {}] $remaining_path $value]
  }
  return $entries
}

proc set_menu_entry {top_entry path value} {
  if {[llength $path]==0} {
    return $value
  }
  assert {[lindex $top_entry 0]=="menu"}
  set entries [lindex $top_entry 2]
  set entries [set_menu_entries_entry $entries $path $value]
  return [list menu [menu_entry_name $top_entry] $entries]
}

proc menu_editor_set_entry {entry} {
  global menu_editor
  global cmd_menu_entries

  set path $menu_editor(item_path)
  set path [lreplace $path 0 0]
  set new_entry [set_menu_entry [menu_editor_top_entry] $path $entry]
  set cmd_menu_entries [lindex $new_entry 2]
  assert {[menu_editor_entry]==$entry}
  save_cmd_menu
  update_cmd_menus
}

proc menu_editor_set_menu_entries {entries} {
  set entry [menu_editor_entry]
  menu_editor_set_entry [lreplace $entry 2 2 $entries]
}

proc menu_editor_top_entry {} {
  global cmd_menu_entries
  return [list menu "Cmd" $cmd_menu_entries]
}

proc menu_editor_set_item_path {path} {
  global menu_editor
  set menu_editor(item_path) $path
  set entry [menu_entry [menu_editor_top_entry] $path]
  set menu_editor(item_label) [menu_entry_label $entry]
  switch -- [lindex $entry 0] {
    command { set menu_editor(item_type) Command }
    history_menu { set menu_editor(item_type) "History menu" }
    menu { set menu_editor(item_type) "Menu" }
    separator { set menu_editor(item_type) "Separator" }
    default { assert {0} }
  }
  menu_editor_fill_item_label_menu
  switch -- $menu_editor(item_type) {
    "Menu" { menu_editor_fill_menu_items_listbox }
    "Command" { menu_editor_fill_command }
    "History menu" { menu_editor_fill_history_menu }
    "Separator" { menu_editor_fill_separator }
  }
}


proc menu_editor_edit_menu_item {} {
  global menu_editor

  set listbox $menu_editor(menu_items_listbox)
  set index [$listbox curselection]

  # Clear the selection so when the listbox is filled again, the selection
  # will be set to the first item
  $listbox selection clear 0 end

  set entry [lindex [menu_editor_menu_entries] $index]
  menu_editor_set_item_path \
    [concat $menu_editor(item_path) [list [menu_entry_name $entry]]]
}


proc menu_editor_cut_menu_item {} {
  if {[catch {
    global menu_editor

    set listbox $menu_editor(menu_items_listbox)
    set index [$listbox curselection]
    set entries [menu_editor_menu_entries]
    if {$index<[llength $entries]} {
      set menu_editor(cut_entry) [lindex $entries $index]
      set entries [lreplace $entries $index $index]
      menu_editor_set_menu_entries $entries
      menu_editor_fill_menu_items_listbox
    } else {
      bell
    }
  } message]} {
    puts $message
    global errorInfo
    puts $errorInfo
  }
}


proc menu_editor_paste_menu_item {} {
  global menu_editor

  if {[info exists menu_editor(cut_entry)]} {
    menu_editor_insert_menu_item $menu_editor(cut_entry)
  }
}

proc menu_editor_insert_menu_item {entry} {
  global menu_editor

  set listbox $menu_editor(menu_items_listbox)
  set index [$listbox curselection]
  set entries [menu_editor_menu_entries]
  if {$index<[llength $entries]} {
    set entries \
      [lreplace $entries $index $index $entry [lindex $entries $index]]
  } else {
    lappend entries $entry
  }
  menu_editor_set_menu_entries $entries
  menu_editor_fill_menu_items_listbox
}

proc menu_editor_menu_item_selected {} {
  global menu_editor

  set listbox $menu_editor(menu_items_listbox)
  set index [$listbox curselection]
  if {$index==""} {
    return
  }
  set item_value [menu_editor_menu_entries]
  set entry [lindex $item_value $index]
  set name ""
  if {$entry!=""} {
    set name [menu_entry_name $entry]
  }
  $menu_editor(menu_item_label_entry) delete 0 end
  $menu_editor(menu_item_label_entry) insert end $name
  focus_entry $menu_editor(menu_item_label_entry)
}

proc menu_editor_set_notebook_frame {selected_frame} {
  global menu_editor
  set notebook $menu_editor(item_edit_notebook)

  update idletasks
  set max_width 0
  set max_height 0
  set frames {menu_frame command_frame history_menu_frame separator_frame}
  foreach frame $frames {
    set width [winfo reqwidth $notebook.$frame]
    set height [winfo reqheight $notebook.$frame]
    if {$width>$max_width} {
      set max_width $width
    }
    if {$height>$max_height} {
      set max_height $height
    }
  }
  foreach frame $frames {
    if {$frame==$selected_frame} {
      place $notebook.$frame -width $max_width -height $max_height
    } else {
      place forget $notebook.$frame
    }
  }
  $notebook configure -width $max_width -height $max_height
}

proc menu_editor_menu_item_label_changed {name1 name2 op} {
  if {[catch {
    global menu_editor
    set label_entry $menu_editor(menu_item_label_entry)
    set text [$label_entry get]
    set listbox $menu_editor(menu_items_listbox)
    set index [$listbox curselection]
    set value [menu_editor_menu_entries]
    if {$index>=[llength $value]} {
      set entry [list command $text {} 0 0]
      lappend value $entry
    } else {
      set entry [lindex $value $index]
      set entry [lreplace $entry 1 1 $text]
      set value [lreplace $value $index $index $entry]
    }
    menu_editor_set_menu_entries $value
    menu_editor_fill_menu_items_listbox
  } message]} {
    puts "message=$message"
    global errorInfo
    puts "errorInfo=$errorInfo"
  }
}

proc menu_editor_command_changed {name1 name2 op} {
  global menu_editor

  set entry [menu_editor_entry]
  set string $menu_editor(command_string)
  set verify $menu_editor(command_verify)
  set spring_expand $menu_editor(command_spring_expand)
  set entry [lreplace $entry 2 4 $string $verify $spring_expand]
  menu_editor_set_entry $entry
}

proc menu_editor_history_max_entries_changed {name1 name2 op} {
  global menu_editor

  set entry [menu_editor_entry]
  set max_entries 20
  catch {set max_entries [expr {$menu_editor(history_max_entries)+0}]}
  set entry [concat [lrange $entry 0 1] [list $max_entries]]
  menu_editor_set_entry $entry
}

proc menu_editor_move_selection {dir} {
  global menu_editor

  set listbox $menu_editor(menu_items_listbox)
  set index [$listbox curselection]
  incr index $dir
  if {$index<0} { 
    return 
  }
  if {$index>=[$listbox index end]} {
    return
  }
  $listbox selection clear 0 end
  $listbox selection set $index
  menu_editor_menu_item_selected
}

proc open_menu_dialog {} {
  let w [toplevel .menu_editor -class Dialog] {
    wm title $w "Menu Editor"
    let w [frame $w.item_frame] {
      let w [label $w.item_label] {
	$w configure -text "Item:"
	pack $w -side left
      }
      let w [optionmenu $w.item_optionmenu] {
	$w configure -textvariable menu_editor(item_label)
	let w [menu $w.menu -tearoff 0] {
	  global menu_editor
	  set menu_editor(item_label_menu) $w
	}
	$w configure -menu $w.menu
	pack $w -side left
      }
      pack $w -side top -anchor w
    }
    let w [frame $w.type_frame] {
      let w [label $w.type_label] {
	$w configure -text "Item type:"
	pack $w -side left
      }
      let w [label $w.type_value_label] {
	$w configure -textvariable menu_editor(item_type)
	global menu_editor
	set menu_editor(item_type) "Menu"
	pack $w -side left
      }
      pack $w -side top -anchor w
    }
    let w [frame $w.separator] {
      $w configure -borderwidth 1 -height 2 -relief sunken
      pack $w -side top -fill x
    }
    let w [frame $w.item_edit_notebook] {
      global menu_editor
      set menu_editor(item_edit_notebook) $w
      let w [frame $w.menu_frame] {
	let w [label $w.label] {
	  $w configure -text "Items:"
	  pack $w -side top -anchor w
	}
	let w [frame $w.list_frame] {
	  let w [listbox $w.listbox] {
	    $w configure -exportselection no
	    global menu_editor
	    set menu_editor(menu_items_listbox) $w
	    bind $w <Double-ButtonRelease-1> menu_editor_edit_menu_item
	    bind $w <ButtonPress> "after idle menu_editor_menu_item_selected"
	    bind $w <B1-Motion> "after idle menu_editor_menu_item_selected"
	    bind $w <ButtonRelease> "menu_editor_menu_item_selected"
	    bind $w <KeyPress> "after idle menu_editor_menu_item_selected"
	    pack $w -side left -expand true -fill both
	  }
	  let w [scrollbar $w.scrollbar] {
	    $w configure -orient vertical
	    pack $w -side left -fill y
	  }
	  $w.scrollbar configure -command "$w.listbox yview"
	  $w.listbox configure -yscrollcommand "$w.scrollbar set"
	  pack $w -side top -expand true -fill both
	}
	let w [entry $w.label_entry] {
	  set menu_editor(menu_item_label_entry) $w
	  $w configure -textvariable menu_editor(menu_item_label)
	  global menu_editor
	  bind $w <Return> {menu_editor_menu_item_label_changed blah blah blah}
	  bind $w <Up> {menu_editor_move_selection -1}
	  bind $w <Down> {menu_editor_move_selection 1}
	  pack $w -side top -expand true -fill x
	}
	let w [frame $w.button_frame] {
	  let w [button $w.edit_button] {
	    $w configure -text "Edit" -command "menu_editor_edit_menu_item"
	    pack $w -side left
	  }
	  let w [button $w.cut_button] {
	    $w configure -text "Cut" -command "menu_editor_cut_menu_item"
	    pack $w -side left
	  }
	  let w [button $w.paste_button] {
	    $w configure -text "Paste" -command "menu_editor_paste_menu_item"
	    pack $w -side left
	  }
	  let w [menubutton $w.insert_button] {
	    $w configure -text "Insert" -relief raised
	    let w [menu $w.menu] {
	      $w configure -tearoff 0
	      $w add command -label "Command" \
	        -command \
		  [list menu_editor_insert_menu_item [list command "" "" 0 0]]
	      $w add command -label "Sub-menu" \
	        -command [list menu_editor_insert_menu_item [list menu "" {}]]
	      $w add command -label "History menu" \
	        -command \
		  [list \
		     menu_editor_insert_menu_item [list history_menu "History"]]
	      $w add command -label "Separator" \
	        -command [list menu_editor_insert_menu_item separator]
	    }
	    $w configure -menu $w.menu
	    pack $w -side left
	  }
	  pack $w -side top
	}
      }
      let w [frame $w.command_frame] {
	let w [label $w.label] {
	  $w configure -text "Command:"
	  pack $w -side top -anchor w
	}
	let w [entry $w.entry] {
	  global menu_editor
	  set menu_editor(command_entry) $w
	  $w configure -textvariable menu_editor(command_string)
	  trace variable menu_editor(command_string) "w" \
	    menu_editor_command_changed
	  pack $w -side top -fill x -anchor n
	}
	let w [checkbutton $w.verify_toggle] {
	  $w configure \
	    -text "Verify before executing" \
	    -variable menu_editor(command_verify)
	  trace variable menu_editor(command_verify) "w" \
	    menu_editor_command_changed
	  pack $w -side top -anchor w
	}

	let w [checkbutton $w.spring_expand_toggle] {
	  $w configure \
	    -text "Spring back to compact" \
	    -variable menu_editor(command_spring_expand)
	  trace variable menu_editor(command_spring_expand) "w" \
	    menu_editor_command_changed
	  pack $w -side top -anchor w
	}
      }
      let w [frame $w.history_menu_frame] {
	let w [frame $w.frame] {
	  let w [label $w.label] {
	    $w configure -text "Maximum entries:"
	    pack $w -side left
	  }
	  let w [entry $w.entry] {
	    global menu_editor
	    set menu_editor(history_max_entries_entry) $w
	    $w configure -width 4 -textvariable menu_editor(history_max_entries)
	    trace variable menu_editor(history_max_entries) "w" \
	      menu_editor_history_max_entries_changed
	    pack $w -side left
	  }
	  pack $w -side top -anchor w
	}
      }
      pack $w -side top
      let w [frame $w.separator_frame] {
      }
    }
    let w [frame $w.bottom_separator] {
      $w configure -borderwidth 1 -height 2 -relief sunken
      pack $w -side top -expand true -fill x
    }
    let w [frame $w.bottom_button_frame] {
      let w [button $w.done_button] {
	$w configure -text "Close" -command "menu_editor_close"
	pack $w -side top
      }
      pack $w -side top
    }
    wm withdraw $w
  }
}

proc edit_menus {} {
  global menu_editor

  open_menu_dialog
  menu_editor_set_item_path {"Cmd"}
  update idletasks
  wm deiconify .menu_editor
}
