#********************************************************/
#* Gsh Copyright (C) 1999 Vaughn Cato                   */
#* See the file "LICENSE", included with this software, */
#* for license information.                             */
#********************************************************/
proc choose_theme_color {name button} {
  global theme
  set new_color [tk_chooseColor -initialcolor $theme($name)]
  if {$new_color!=""} {
    set theme($name) $new_color
    $button configure -background $theme($name) -activebackground $theme($name)
    theme_updated
  }
}

proc close_theme_dialog {} {
  destroy .theme_editor
}

proc open_theme_dialog {} {
  global theme
  let w .theme_editor {
    if {[winfo exists $w]} {
      destroy $w
    }
    toplevel $w -class Dialog
    wm title $w "Theme Editor"
    let w [frame $w.upper_frame] {
      set items {
	{"Output" output}
	{"Prompt" prompt}
	{"Active Command" active_command}
	{"Inactive Command" inactive_command}
	{"Focused Command" focused_command}
	{"Selection" selection}
      }
      set attributes {
	{"Foreground" foreground}
	{"Background" background}
	{"Cursor" cursor_color}
      }
      set col 1
      foreach attr $attributes {
	set label [lindex $attr 0]
	set name [lindex $attr 1]
	let w [label $w.${name}_label -text $label] {
	  grid $w -row 0 -column $col
	}
	incr col
      }
      set row 1
      foreach item $items {
	set label [lindex $item 0]
	set name [lindex $item 1]
	label $w.$name -text $label
	grid $w.$name -row $row -column 0
	set col 1
	foreach attr $attributes {
	  set attr_name [lindex $attr 1]
	  set themename ${name}_${attr_name}
	  set varname theme($themename)
	  if {[info exists $varname]} {
	    set color [set $varname]
	    button $w.$themename \
	      -background $color \
	      -activebackground $color \
	      -command "choose_theme_color $themename $w.$themename"
	    grid $w.$themename -row $row -column $col
	  }
	  incr col
	}
	incr row
      }
      pack $w -side top
    }
    let w [frame $w.separator -relief sunken -borderwidth 1] {
      $w configure -height 2
      pack $w -side top -fill x -expand true
    }
    let w [frame $w.bottom_frame] {
      let w [button $w.save_button] {
	$w configure -text "Save" -command "save_theme"
	pack $w -side left -expand true
      }
      let w [button $w.close_button] {
	$w configure -text "Close" -command "close_theme_dialog"
	pack $w -side left -expand true
      }
      pack $w -side top -expand true -fill x
    }
  }
}
