#********************************************************/
#* Gsh Copyright (C) 1999 Vaughn Cato                   */
#* See the file "LICENSE", included with this software, */
#* for license information.                             */
#********************************************************/
source keylist.tcl

proc handle_types_line {line} {
  set linelen [llength $line]
  set index 0
  set spec [next_element $line index]
  switch -- $spec {
    mime_types_file { read_mime_types [next_element $line index] }
    type {
      while {$index < $linelen} {
	set word [next_element $line index]
	switch -- $word {
	  -name { set name [next_element $line index] }
	  -desc { set desc [next_element $line index] }
	  -exts { set exts [next_element $line index] }
	  -base { set base [next_element $line index] }
	  -recognizer { set recognizer [next_element $line index] }
	  default { error "Unknown option ($word) for type specification" }
	}
      }
      global type_info
      if {[info exists desc]} {
	klset type_info($name) desc $desc
      }
      if {[info exists exts]} {
	klset type_info($name) exts $exts
      }
      if {[info exists recognizer]} {
	klset type_info($name) recognizer $recognizer
      }
      if {[info exists base]} {
	klset type_info($name) base $base
	if {![info exists type_info($base)]} {
	  set type_info($base) {}
	}
      }
    }
    default { error "Unknown specification ($spec) in types file" }
  }
}

proc read_types {filename} {
  process_file $filename line { handle_types_line $line }
}

proc handle_mime_types_line {line} {
  set exts ""
  regexp {type=([^ ]*)} $line match type
  regexp {desc="([^"]*)"} $line match desc
  regexp {exts="([^"]*)"} $line match exts

  # need to do something here
}

proc read_mime_types {filename} {
  process_file $filename line { handle_mime_types_line $line }
}

proc handle_popup_commands_line {line} {
  set index 0
  set linelen [llength $line]
  switch -- [next_element $line index] {
    popup_command {
      set command(verify) 0
      while {$index < $linelen} {
	switch -- [next_element $line index] {
	  -type { set type [next_element $line index] }
	  -command { set command(string) [next_element $line index] }
	  -verify { set command(verify) [next_element $line index] }
	  -desc { set command(desc) [next_element $line index] }
	}
      }
      global type_info
      klappend type_info($type) commands [klfromarray command]
      unset command
    }
  }
}

proc read_popup_commands {filename} {
  process_file $filename line { handle_popup_commands_line $line }
}

if 0 {
  proc evaluate_recognizer {recognizer varsname} {
    upvar $varsname vars
    set body "$recognizer\n"
    set body "$recognizer
    proc evaluate_recognizer_helper [lindex $vars 0] \
      "$recognizer\n
    set result [eval [concat [list evaluate_recognizer_helper] [lindex $vars 0]]]
    return $result
    foreach var [klkeys vars] {
      klget vars $var $var
    }

    set result [eval $recognizer]

    klnew vars
    foreach var [info vars] {
      klset vars $var set $var
    }
    puts "vars final = $vars"
    return $result
  }
}

proc unquoted_filename {filename} {
  if {[regexp {^"(.*)"$} $filename match inside]} {
    set filename $inside
  }
  if {[regexp {^'(.*)'$} $filename match inside]} {
    set filename $inside
  }
  return $filename
}

proc data_types2 {} {
  set types {}
  upvar sel selection
  upvar dir dir
  upvar cmd cmd
  global type_info
  set value $selection

  if {$selection == ""} {
    set cmdname [command_name $cmd]
    set type [list $cmdname "command"]
    klset type_info($type) \
      recognizer "expr {\$sel==\"\" && \[command_name \$cmd]==\"$cmdname\"}"
    lappend types $type
  } else {
    set filenames [cmd_args $selection]
    set type_list [type_list]
    if {[llength $filenames]==1} {
      set filename [lindex $filenames 0]
      if {[file exists $filename]} {
        set value $filename
        foreach type $type_list {
	  if {[klget type_info($type) exts exts]} {
	    foreach ext $exts {
	      if {[string match *$ext $filename]} {
	        lappend types $type
		puts "Added type $type"
	      }
	    }
	  }
        }
        while {[regexp {^.[^.]*(\..*)$} $filename match ext]} {
	  lappend types $ext
	  set filename $ext
        }
      }
    } elseif {[llength $filenames]>1} {
      set files [file_list $dir $selection]
      if {$files!=""} {
	lappend types "multiple_files"
	set value $files
      }
    }

    # If there is no base type, or the base type is valid or unrestricted
    #   If the type has a recognizer
    #     If the recognizer succeeds, the type is valid
    #     If the recognizer fails, the type is invalid
    #   Else, the type is unrestricted
    # Else, the type is invalid

    set unresolved_types $type_list
    set unrestricted_types {}
    set invalid_types {}
    foreach type $types {
      set unresolved_types [lremove $unresolved_types $type]
    }
    while {[llength $unresolved_types]>0} {
      foreach type $unresolved_types {
	if {[klget type_info($type) base base]} {
	  if {[lsearch $invalid_types $base]!=-1} {
	    set invalid_base 1
	    set valid_base 0
	  } elseif {[lsearch $unrestricted_types $base]!=-1} {
	    set valid_base 1
	    set invalid_base 0
	  } elseif {[lsearch $types $base]!=-1} {
	    set valid_base 1
	    set invalid_base 0
	  } else {
	    set valid_base 0
	    set invalid_base 0
	  }
	} else {
	  set invalid_base 0
	  set valid_base 1
	}
	if {$invalid_base || $valid_base} {
	  if {$invalid_base} {
	    set type_type invalid
	  } else {
	    if {[klget type_info($type) recognizer recognizer]} {
	      set recognizer [klget type_info($type) recognizer]
	      set code [catch {uplevel $recognizer} result]
	      if {$code!=2 && $code!=0} {
		return -code $code \
		  "Error executing recognizer for type $type: $result"
	      }
	      if {$result} {
		set type_type valid
	      } else {
		set type_type invalid
	      }
	    } else {
	      set type_type unrestricted
	    }
	  }
	  switch -- $type_type {
	    valid { lappend types $type }
	    invalid { lappend invalid_types $type }
	    unrestricted { lappend unrestricted_types $type }
	    default { assert 0 }
	  }
	  set unresolved_types [lremove $unresolved_types $type]
	  unset type_type
	}
      }
    }
    foreach type $types {
      if {[info exists type_info($type)] && [klget type_info($type) base base]} {
	if {[lsearch $types $base]==-1} {
	  lappend types $base
	}
      }
    }
  }
  uplevel [list set _types $types]
  uplevel [list set _value $value]
}

# This proc provides a scope for the recognizers.  The recognizers will be
# executed in this scope so that any variables they set will persist until
# all the data types have been found, and so that the number of variables
# visible to the recognizer will be limited.
proc data_types { sel dir cmd _user_var_keylist_name} {
  data_types2
  upvar $_user_var_keylist_name _user_var_keylist
  foreach _var [info vars] {
    if {![string match _* $_var]} {
      klset _user_var_keylist $_var [set $_var]
    }
  }
  return [list $_types $_value]
}

proc type_editor_new_proc {} {
  type_editor_set_selection end
}

proc type_editor_fill_command_list {} {
  global type_editor
  set listbox $type_editor(command_listbox)
  set sel_index [$listbox curselection]
  global type_info
  upvar #0 type_editor(type) type
  $listbox delete 0 end
  if {![info exists type_info($type)]} {
    klnew type_info($type)
  }
  if {[klget type_info($type) commands commands]} {
    foreach command $commands {
      set string [klget command string]
      $listbox insert end $string
    }
  }
  $listbox insert end ""
  if {$sel_index!=""} {
    $listbox selection clear 0 end
    $listbox selection set $sel_index
  }
}

proc type_editor_remove_proc {} {
  global type_editor
  global type_info
  set listbox $type_editor(command_listbox)
  set index [$listbox curselection]
  set commands [klget type_info($type_editor(type)) commands]
  if {$index>=[llength $commands]} {
    return
  }
  set commands [lreplace $commands $index $index]
  klset type_info($type_editor(type)) commands $commands
  type_editor_fill_command_list
  type_editor_set_selection $index
}

proc type_editor_command_selected {w} {
  global type_info
  global type_editor

  set sel_index [$w curselection]
  set entry $type_editor(cmd_entry)
  set verify 0
  set command_string ""
  if {[klget type_info($type_editor(type)) commands commands]} {
    set command [lindex $commands $sel_index]
    klget command verify verify
    klget command string command_string
    klget command desc desc
  }
  set type_editor(command_verify) $verify
  if {[info exists desc]} {
    $type_editor(desc_entry) configure -state normal
    set type_editor(desc_enabled) 1
    $type_editor(desc_entry) delete 0 end
    $type_editor(desc_entry) insert end $desc
  } else {
    $type_editor(desc_entry) configure -state normal
    set type_editor(desc_enabled) 0
    $type_editor(desc_entry) delete 0 end
    $type_editor(desc_entry) insert end $command_string
    $type_editor(desc_entry) configure -state disabled
  }
  $entry delete 0 end
  $entry insert end $command_string
}

proc type_editor_set_selection {new_index} {
  global type_editor
  set listbox $type_editor(command_listbox)
  set old_index [$listbox curselection]
  if {[llength $old_index]>0} {
    $listbox selection clear $old_index
  }
  $listbox selection set $new_index
  $listbox activate $new_index
  type_editor_command_selected $listbox
}

proc type_editor_entry_proc {w} {
  global type_editor

  set listbox $type_editor(command_listbox)
  set index [$listbox curselection]
  set command [type_editor_get_command]
  klget command string command_string
  puts "command_string=$command_string"
  if {$command_string==""} {
    type_editor_remove_proc
  } else {
    global type_info
    set commands ""
    klget type_info($type_editor(type)) commands commands
    if {$index>=[llength $commands]} {
      lappend commands $command
    } else {
      set commands [lreplace $commands $index $index $command]
    }
    klset type_info($type_editor(type)) commands $commands
    type_editor_fill_command_list
  }
  type_editor_set_selection $index
}

proc type_editor_move_selection {dir} {
  global type_editor
  set listbox $type_editor(command_listbox)
  set index [$listbox curselection]
  set new_index [expr {$index+$dir}]
  set end_index [$listbox index end]
  if {$new_index>=0 && $new_index<$end_index} {
    type_editor_set_selection $new_index
  }
}

proc type_editor_type_changed {type} {
  set desc ""
  global type_info
  global type_editor
  if {[info exists type_info($type)]} {
    klget type_info($type) desc desc
  }
  set type_editor(type_desc) $desc
  type_editor_update_base
  type_editor_fill_command_list
  set listbox $type_editor(command_listbox)
  type_editor_set_selection end
}

proc type_editor_set_type {type} {
  global type_editor
  set type_editor(type) $type
  type_editor_type_changed $type
}

proc type_editor_set_base {type} {
  global type_editor
  set type_editor(base) $type
  global type_info
  klset type_info($type_editor(type)) base $type
  puts "set base of $type_editor(type) to $type"
}

proc sort_types {type_list} {
  global type_info
  while {[llength $type_list]>0} {
    set index 0
    set type [lindex $type_list $index]
    while {[klget type_info($type) base base]} {
      set base_index [lsearch $type_list $base]
      assert {$base_index!=0}
      if {$base_index>0} {
	set type $base
	set index $base_index
      } else {
	break
      }
    }
    set type_list [lreplace $type_list $index $index]
    lappend result $type
  }
  return $result
}

proc type_list {} {
  global type_info
  return [array names type_info]
}

proc rc_dir {} {
  global env
  return "$env(HOME)/.gsh"
}

proc save_types {filename} {
  set types [type_list]
  set types [sort_types $types]
  set file [open "[rc_dir]/types" "w"]
  global type_info
  foreach type $types {
    set args {}
    foreach member [klkeys type_info($type)] {
      if {$member!="commands"} {
	set value [klget type_info($type) $member]
	if {$member!="desc" || $value!=""} {
	  lappend args -$member $value
	}
      }
    }
    if {$args!={}} {
      puts $file [concat [list type -name $type] $args]
    }
  }
  close $file
}

proc save_popup_commands {filename} {
  set file [open "[rc_dir]/popup_commands" "w"]
  global type_info
  set types [type_list]
  foreach type $types {
    if {[klget type_info($type) commands commands]} {
      foreach command $commands {
	set command_string ""
	set verify 0
	catch {unset desc}
	klget command string command_string
	klget command verify verify
	klget command desc desc
	set line [list popup_command -type $type]
	lappend line -command $command_string
	if {$verify} {
	  lappend line -verify $verify
	}
	if {[info exists desc] && $desc!=""} {
	  lappend line -desc $desc
	}
	puts $file $line
      }
    }
  }
  close $file
}

proc close_type_editor {} {
  destroy .type_editor
  save_types "types"
  save_popup_commands "popup_commands"
}

proc type_editor_type_desc_changed {name1 name2 op} {
  global type_editor
  global type_info
  klset type_info($type_editor(type)) desc $type_editor(type_desc)
}

# Get the state of the widgets and return them in a keylist
proc type_editor_get_command {} {
  global type_editor

  klnew command
  klset command verify $type_editor(command_verify)
  klset command string [$type_editor(cmd_entry) get]
  if {$type_editor(desc_enabled)} {
    klset command desc [$type_editor(desc_entry) get]
  }
  return $command
}

proc type_editor_update_command {} {
  global type_info
  global type_editor

  klget type_info($type_editor(type)) commands commands
  set sel_index [$type_editor(command_listbox) curselection]
  if {$sel_index<[llength $commands]} {
    set command [type_editor_get_command]
    set commands [lreplace $commands $sel_index $sel_index $command]
    klset type_info($type_editor(type)) commands $commands
    $type_editor(command_listbox) delete $sel_index
    klget command string command_string
    $type_editor(command_listbox) insert $sel_index $command_string
    $type_editor(command_listbox) selection set $sel_index
  }
}

proc type_editor_desc_toggled {} {
  global type_editor
  if {$type_editor(desc_enabled)} {
    $type_editor(desc_entry) configure -state normal
    focus $type_editor(desc_entry)
  } else {
    $type_editor(desc_entry) delete 0 end
    $type_editor(desc_entry) insert end [$type_editor(cmd_entry) get]
    $type_editor(desc_entry) configure -state disabled
  }
  type_editor_update_command
}

proc type_editor_verify_toggled {} {
  type_editor_update_command
}

proc type_editor_cmd_desc_changed {name1 name2 op} {
  type_editor_update_command
}

proc type_editor_command_changed {name1 name2 op} {
  global type_editor
  if {!$type_editor(desc_enabled)} {
    $type_editor(desc_entry) configure -state normal
    $type_editor(desc_entry) delete 0 end
    $type_editor(desc_entry) insert end [$type_editor(cmd_entry) get]
    $type_editor(desc_entry) configure -state disabled
  }
  type_editor_update_command
}

proc type_editor_update_base {} {
  set base "Any"
  global type_info
  global type_editor
  if {[info exists type_info($type_editor(type))]} {
    klget type_info($type_editor(type)) base base
  }
  global type_editor
  set type_editor(base) $base
}

proc open_type_editor {types} {
  catch {destroy .type_editor}
  let w [toplevel .type_editor -class Dialog] {
    wm title $w "Type Editor"
    wm protocol $w WM_DELETE_WINDOW "close_type_editor"
    let w [frame $w.type_frame] {
      let w [label $w.type_label -text "Type:"] {
	pack $w -side left
      }
      let w [menubutton $w.type_optionmenu] {
	global type_editor
	set type_editor(type) [lindex $types [expr {[llength $types]-1}]]
	$w configure \
	      -textvariable type_editor(type) -indicatoron 1 -menu $w.menu \
	      -relief raised -bd 2 -highlightthickness 2 -anchor c \
	      -direction flush
	let w [menu $w.menu -tearoff 0] {
	  foreach type $types {
	    $w add command -label $type \
	    -command [list type_editor_set_type $type]
	  }
	}
	$w configure -menu $w.menu
	pack $w -side left
      }
      pack $w -side top
    }
    let w [frame $w.base_type_frame] {
      let w [label $w.label -text "Base Type:"] {
	pack $w -side left
      }
      let w [menubutton $w.optionmenu] {
	set type_editor(base) ""
	$w configure \
	  -textvariable type_editor(base) -indicatoron 1 -menu $w.menu \
	  -relief raised -bd 2 -highlightthickness 2 -anchor c \
	  -direction flush
	let w [menu $w.menu -tearoff 0] {
	  foreach type [type_list] {
	    $w add command -label $type \
	      -command "type_editor_set_base $type"
	  }
	}
	pack $w -side top
      }
      pack $w -side top -anchor w
    }
    let w [frame $w.desc_frame] {
      let w [label $w.label -text "Description:"] {
	pack $w -side left
      }
      let w [entry $w.entry -textvariable type_editor(type_desc)] {
	global type_editor
	trace variable type_editor(type_desc) "w" \
	  {type_editor_type_desc_changed}
	pack $w -side left
      }
      pack $w -side top
      global type_info
      global type_editor
    }
    let w [label $w.commands_label] {
      $w configure -text "Commands:"
      pack $w -side top -anchor w
    }
    let w [frame $w.command_frame -relief groove -borderwidth 2] {
      let w [listbox $w.command_list] {
	$w configure -exportselection no
	pack $w -side top -expand 1 -fill both -padx 10 -pady 10
	bind $w <ButtonPress> "after idle type_editor_command_selected $w"
	bind $w <ButtonRelease> "after idle type_editor_command_selected $w"
	bind $w <B1-Motion> "after idle type_editor_command_selected $w"
	bind $w <KeyPress> "after idle type_editor_command_selected $w"
	global type_editor
	set type_editor(command_listbox) $w
      }
      let w [checkbutton $w.desc_toggle] {
	$w configure -text "Description:"
	$w configure -variable type_editor(desc_enabled)
	$w configure -command "type_editor_desc_toggled"
	pack $w -side top -anchor w
      }
      let w [entry $w.desc_entry] {
	$w configure -textvariable type_editor(cmd_desc)
	trace variable type_editor(cmd_desc) w {type_editor_cmd_desc_changed}
	set type_editor(desc_entry) $w
	$w configure -state disabled
	pack $w -side top -expand true -fill x -padx 10
	bind $w <Return> "type_editor_entry_proc $w"
	bind $w <Down> "type_editor_move_selection 1"
	bind $w <Up> "type_editor_move_selection -1"
      }
      let w [label $w.command_label] {
	$w configure -text "Command:"
	pack $w -side top -anchor w
      }
      let w [entry $w.cmd_entry] {
	$w configure -textvariable type_editor(command)
	trace variable type_editor(command) w {type_editor_command_changed}
	pack $w -side top -expand 1 -fill x -padx 10
	bind $w <Return> "type_editor_entry_proc $w"
	bind $w <Down> "type_editor_move_selection 1"
	bind $w <Up> "type_editor_move_selection -1"
	focus $w
	set type_editor(cmd_entry) $w
      }
      let w [frame $w.verify_frame] {
	let w [checkbutton $w.checkbutton] {
	  $w configure -text "Verify before executing"
	  $w configure -command "type_editor_verify_toggled"
	  $w configure -variable type_editor(command_verify)
	  pack $w -side left
	}
	pack $w -side top
      }
      pack $w -side top -expand 1 -fill both -padx 10
    }
    let w [frame $w.separator] {
      $w configure -height 2 -borderwidth 1 -relief raised
      pack $w -side top -fill x -expand true
    }
    let w [frame $w.command_buttons] {
      let w [button $w.new_button -text "New"] {
	$w configure -command "type_editor_new_proc"
	pack $w -side left
      }
      let w [button $w.remove_button -text "Remove"] {
	$w configure -command "type_editor_remove_proc"
	pack $w -side left
      }
      pack $w -side bottom
    }
  }
  type_editor_type_changed [lindex $types 0]
}
