/********************************************************/
/* Gsh Copyright (C) 1999 Vaughn Cato                   */
/* See the file "LICENSE", included with this software, */
/* for license information.                             */
/********************************************************/
#include <stdio.h>
#include <assert.h>
#include "termstate.h"

static void EvalProc(void *data,TermOp_t op)
{
  switch (op) {
    case TERM_HOME_CURSOR:
      fprintf(stderr,"home_cursor");
      break;
    case TERM_CLEAR_BELOW:
      fprintf(stderr,"clear_below");
      break;
    case TERM_CLEAR_TO_RIGHT:
      fprintf(stderr,"clear_to_right");
      break;
    case TERM_NORM_ATTR:
      fprintf(stderr,"norm_attr");
      break;
    case TERM_SAVE_CURSOR:
      fprintf(stderr,"save_cursor");
      break;
    case TERM_RESTORE_CURSOR:
      fprintf(stderr,"restore_cursor");
      break;
    case TERM_BELL:
      fprintf(stderr,"bell");
      break;
    case TERM_BACKSPACE:
      fprintf(stderr,"backspace");
      break;
    case TERM_CARRIAGE_RETURN:
      fprintf(stderr,"carriage_return");
      break;
    default:
      assert(0);
  }
}

static void EvalProcInt(void *data,TermOp_t op,int arg1)
{
  switch (op) {
    case TERM_APP_CURSOR_KEYS:
      fprintf(stderr,"app_cursor_keys %d\n",arg1);
      break;
    case TERM_SET_ROW:
      fprintf(stderr,"set_row %d\n",arg1);
      break;
    case TERM_INSERT_LINE:
      fprintf(stderr,"insert_line %d\n",arg1);
      break;
    case TERM_DELETE_LINE:
      fprintf(stderr,"delete_line %d\n",arg1);
      break;
    case TERM_CURSOR_RIGHT:
      fprintf(stderr,"cursor_right %d\n",arg1);
      break;
    case TERM_CURSOR_DOWN:
      fprintf(stderr,"cursor_down %d\n",arg1);
      break;
    case TERM_APP_KEYPAD:
      fprintf(stderr,"app_keypad %d\n",arg1);
      break;
    case TERM_UP_LINE:
      fprintf(stderr,"up_line %d\n",arg1);
      break;
    case TERM_TEXT:
      fprintf(stderr,"text '%c'\n",arg1);
      break;
    case TERM_UNEXPECTED:
      fprintf(stderr,"unexpected %d\n",arg1);
      assert(0);
      break;
    case TERM_ALT_CHARSET:
      fprintf(stderr,"alt_charset %d\n",arg1);
      break;
    default:
      assert(0);
  }
}

static void EvalProc2Int(void *data,TermOp_t op,int arg1,int arg2)
{
  switch (op) {
    case TERM_SET_CURSOR:
      fprintf(stderr,"set_cursor %d %d\n",arg1,arg2);
      break;
    case TERM_SET_SCROLL_REGION:
      fprintf(stderr,"set_scroll_region %d %d\n",arg1,arg2);
      break;
    default:
      assert(0);
  }
}

int main(int argc,char **argv)
{
  TermState state;
  TermState_Init(&state,0,EvalProc,EvalProcInt,EvalProc2Int);

  {
    char *string = "test\x1b[25;1H\x1b[25;1Hagaintest\x1b[25;1H";
    {
      char *p = string;
      for (;*p;++p) {
        TermState_AddChar(&state,*p);
      }
    }
  }
  return 0;
}
