#ifndef TERMTCL_H_
#define TERMTCL_H_

#include <tcl.h>

typedef enum { 
  NO_STYLE,
  USASCII_STYLE,
  DEC_SPECIAL_STYLE,
  UK_STYLE,
  MULTINATIONAL_STYLE,
  FINNISH_STYLE,
  GERMAN_STYLE,
  END_STYLE
} FontStyle_t;

typedef struct Term_s Term;

struct Term_s {
  int id;
  char *termname;
  char *widget_name;
  char *tags;
  char *fgcolor;
  char *bgcolor;
  char *font_tags;
  char *insert_mark;
  int charset;
  int insert_mode;
  int inverse;
  int cur_row;
  int cur_col;
  int row_offset;
  int scroll_top;
  int scroll_bottom;
  int screen_height;
  int screen_width;
  int update_pending;
  Tcl_Interp *interp;
  ClientData widget_clientdata;
  Tcl_CmdProc *widget_proc;
  FontStyle_t font_styles[2];
};

extern int TermTcl_Init(Tcl_Interp *interp);
extern void TermTcl_PrintProfile(void);
extern void TermTcl_Insert(Term *term,const char *text,int len);
extern void TermTcl_SetFgColor(Term *term,int fgcolor);
extern void TermTcl_SetBgColor(Term *term,int bgcolor);
extern void TermTcl_SetCharset(Term *term,int charset);
extern void TermTcl_ResetTags(Term *term);
extern void TermTcl_SetCursor(Term *term,int row,int col);
extern void TermTcl_Tab(Term *term);
extern void TermTcl_MoveDown(Term *term);
extern void TermTcl_CR(Term *term);
extern void TermTcl_Backspace(Term *term);
extern void TermTcl_HomeCursor(Term *term);
extern void TermTcl_MoveUp(Term *term);
extern void TermTcl_SetInverse(Term *term,int value);
extern void TermTcl_SetScrollRegion(Term *term,int top,int bottom);
extern void TermTcl_InsertLine(Term *term,int count);
extern void TermTcl_ClearScreen(Term *term);
extern void TermTcl_SetInsertMode(Term *term,int value);
extern void TermTcl_SaveCursor(Term *term);
extern void TermTcl_RestoreCursor(Term *term);
extern void TermTcl_ClearBelow(Term *term);
extern void TermTcl_SetTab(Term *term,int value);
extern void TermTcl_ClearTabs(Term *term);
extern void TermTcl_Unexpected(Term *term,int value);
extern void TermTcl_SetFontTag(Term *term,char *tagname,int value);
extern void TermTcl_SetRow(Term *term,int value);
extern void TermTcl_SetCol(Term *temr,int value);
extern void TermTcl_SetFontStyle(Term *term,int charset,FontStyle_t style);
extern void TermTcl_DeleteLine(Term *term,int count);
extern void TermTcl_ClearToRight(Term *term);
extern void TermTcl_CursorUp(Term *term,int count);
extern void TermTcl_CursorDown(Term *term,int count);
extern void TermTcl_CursorLeft(Term *term,int count);
extern void TermTcl_CursorRight(Term *term,int count);
extern void TermTcl_InsertChar(Term *term,int count);
extern void TermTcl_DeleteChar(Term *term,int count);
extern Term *TermPtr(Tcl_Interp *interp,char *termname);

#endif /* TERMTCL_H_ */
